/*
 *                        s a c e s
 *      Simple Artificial Chemistry Experiment System
 *
 * A semester project Summer 2005 and a diploma project 2006
 *           Berne University of Applied Sciences
 *   HTI School of Engineering and Information Technology
 *             Anthony Aguillon, Daniel Noelpp
 *
 */
package saces.app;


import saces.Util;
import saces.app.gui.ApplicationButtonPanel;
import static saces.app.gui.ApplicationButtonPanel.VIEWDATA;
import static saces.app.gui.ApplicationButtonPanel.SETTINGS;
import static saces.app.gui.ApplicationButtonPanel.SNAPSHOT;
import static saces.app.gui.RuntimeButtonPanel.*;
import saces.app.gui.RuntimeButtonPanel;
import saces.app.gui.SettingsFrame;
import saces.app.gui.ViewDataFrame;
import saces.app.gui.ViewDataModel;
import saces.exp.Experiment;
import saces.file.BadExperimentFileException;
import saces.sim.Mediator;
import saces.sim.SimulationPNPException;

import javax.swing.JFrame;
import javax.swing.JPanel;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.Reader;

/**
 * The Application class handles user input, delegating tasks to the
 * objects responsible for processing. It maintains instances of gui
 * components such as toolbars and dialogs. All actions made on toolbars
 * are routed to the Application object which implements and registers
 * an ActionListener interface.
 *
 */
public class Application implements ActionListener {

// Read data from the binary log file
private ViewDataModel viewDataModel;

private JFrame mainFrame = new JFrame();
private SettingsFrame settingsFrame = new SettingsFrame(this);
private ViewDataFrame viewDataFrame;
private JPanel buttonPanel = new JPanel(); // for the horizontal toolbar
private JPanel viewControlPanel = new JPanel(); // for the vertical toolbar
private ApplicationButtonPanel applicationButtons =
    new ApplicationButtonPanel(this); // the horizontal toolbar
private RuntimeButtonPanel runtimeButtons =
    new RuntimeButtonPanel(this); // the vertical toolbar

private Component glComponent;

// Three basic application states
private static final int BUTTON_STATE_INITIAL = 0;
private static final int BUTTON_STATE_EXPERIMENT_LOAD = 1;
private static final int BUTTON_STATE_EXPERIMENT_RUN = 2;

private Mediator mediator;

/** Create an Application instance with mediator. */
public Application(Mediator mediator) {

    this.mediator = mediator;

    mainFrame.setBounds(20, 20, 450, 400);
    mainFrame.getContentPane().setBackground(Color.black);
    mainFrame.setBackground(Color.black);

    mainFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    mainFrame.getContentPane().add(buttonPanel, BorderLayout.SOUTH);
    buttonPanel.add(applicationButtons);
    mainFrame.getContentPane().add(viewControlPanel, BorderLayout.EAST);
    viewControlPanel.add(runtimeButtons);

    setButtonState(BUTTON_STATE_INITIAL);

    glComponent = mediator.getGLComponent();
    glComponent.requestFocus();
    mainFrame.add(glComponent, BorderLayout.CENTER);
}

/** Show the main frame window. */
public void show() {
    mainFrame.setVisible(true);
}

/** Load an experiment (in XML) from Reader. */
public void loadExperiment(Reader r) throws BadExperimentFileException {
    if (r == null) return;
    Experiment experiment = new Experiment(r);
    setExperiment(experiment);
    if (experiment.getValidationProblem() != null) {
        Util.showErrorMessage("Problem validating experiment",
            experiment.getValidationProblem());
    }
}

/** Set the application's experiment. */
public void setExperiment(Experiment experiment) {
    mediator.setExperiment(experiment);

    if (viewDataModel != null) viewDataModel.setExperiment(experiment);

    this.setButtonState(BUTTON_STATE_EXPERIMENT_LOAD);
    mainFrame.setTitle("s A c E s :: " + experiment.getDescription());
    mainFrame.repaint();
}

/** Get the current experiment the application is using. */
public Experiment getExperiment() {
    return mediator.getSimulation().getExperiment();
}

/** Start or halt the simulation. */
public void setRunning(boolean running) {

    mediator.setRunning(running);

    applicationButtons.setState(RUNSTOP, !running);
    //applicationButtons.setEnabled(SETTINGS, !running);
    runtimeButtons.setState(RuntimeButtonPanel.RUNSTOP, !running);

    if (running) settingsFrame.setVisible(false);
}

/** Handle button clicks from the toolbars. */
public void actionPerformed(ActionEvent e) {
    clicked(e.getActionCommand());
    if(mediator != null) {
      //mediator.setCanvasFocus();
    }
}

/** Show the data viewer window. */
public void showViewData() {
    boolean isRunning = mediator.isRunning();
    boolean yes = false;
    if (isRunning && viewDataModel == null)
        yes = Util.showYesNo("Open view", ""
            + "Do you really want to open a view?\n\n"
            + "This makes a running simulation slower and jerky.\n"
            + "It is recommended to put the binary log file on a\n"
            + "network share and run the viewer on a different\n"
            + "work station.");

    if (yes || !isRunning) {
        if (viewDataModel == null)
            viewDataModel = new ViewDataModel(getExperiment());
        viewDataFrame = new ViewDataFrame(viewDataModel);

        // If started with data view only, must exit the program on closing.
        viewDataFrame.setDefaultCloseOperation(mainFrame.isVisible()
            ? JFrame.HIDE_ON_CLOSE : JFrame.EXIT_ON_CLOSE);
    }

    viewDataFrame.setVisible(true);
}

// Handle all button clicks
private void clicked(String cmd) {
    if (cmd.equals(RUNSTOP)) {

        setButtonState(BUTTON_STATE_EXPERIMENT_RUN);
        setRunning(applicationButtons.getState(RUNSTOP));

    }

    else if (cmd.equals(SETTINGS)) {

        setRunning(false);
        settingsFrame.setExperiment(mediator.getSimulation().getExperiment());
        settingsFrame.setVisible(true);

    }

    else if (cmd.equals(SNAPSHOT)) {

        setRunning(false);
        mediator.getSimulation().saveSnapshot();

    }

    else if (cmd.equals(VIEWDATA)) {
        showViewData();
    }

    else if (cmd.equals(DEFAULT_PERSPECTIVE)) {

        mediator.setViewMode(Mediator.DEFAULT_PERSPECTIVE);
        runtimeButtons.setEnabled(DEFAULT_PERSPECTIVE, false);
        runtimeButtons.setEnabled(PARTICLE_VIEW, true);
        runtimeButtons.setEnabled(ORTHOGRAPHIC_VIEW, true);
    }

    else if (cmd.equals(ORTHOGRAPHIC_VIEW)) {
        mediator.setViewMode(Mediator.ORTHOGRAPHIC);
        runtimeButtons.setEnabled(DEFAULT_PERSPECTIVE, true);
        runtimeButtons.setEnabled(PARTICLE_VIEW, true);
        runtimeButtons.setEnabled(ORTHOGRAPHIC_VIEW, false);
    }

    else if (cmd.equals(PARTICLE_VIEW)) {
        mediator.setViewMode(Mediator.PARTICLE_PERSPECTIVE);
        runtimeButtons.setEnabled(DEFAULT_PERSPECTIVE, true);
        runtimeButtons.setEnabled(PARTICLE_VIEW, true);
        runtimeButtons.setEnabled(ORTHOGRAPHIC_VIEW, true);
    }

    else if (cmd.equals(DETAIL_LOW)) {
        mediator.setDetailLevel(Mediator.DETAIL_LEVEL_LOW);
        validateDetailSettings();
    }

    else if (cmd.equals(DETAIL_MEDIUM)) {
        mediator.setDetailLevel(Mediator.DETAIL_LEVEL_MEDIUM);
        validateDetailSettings();
    }

    else if (cmd.equals(DETAIL_HIGH)) {
        mediator.setDetailLevel(Mediator.DETAIL_LEVEL_HIGH);
        validateDetailSettings();
    }

    else if (cmd.equals(FASTER)) {

        if (mediator.increaseSpeed() >=
            Mediator.SIMULATION_MAX_SLOWING_FACTOR) {
            runtimeButtons.setEnabled(FASTER, false);
            runtimeButtons.setEnabled(SLOWER, true);
        }
        else {
            runtimeButtons.setEnabled(FASTER, true);
            runtimeButtons.setEnabled(SLOWER, true);
        }
    }

    else if (cmd.equals(SLOWER)) {
        if (mediator.decreaseSpeed() <=
            Mediator.SIMULATION_MIN_SLOWING_FACTOR) {
            runtimeButtons.setEnabled(FASTER, true);
            runtimeButtons.setEnabled(SLOWER, false);
        }
        else {
            runtimeButtons.setEnabled(FASTER, true);
            runtimeButtons.setEnabled(SLOWER, true);
        }
    }

    else if (cmd.equals(RESET)) {
        boolean wasRunning = mediator.isRunning();
        Experiment experiment = getExperiment();

        if (!Util.showYesNo("Reset simulation?",
            "This will reset the current simulation\n"
                + "AND the binary log.\n\n"
                + "Do you really want to reset the simulation\n"
                + "to the beginning and delete the binary log?")) return;

        try {
            setExperiment(experiment);
            setRunning(wasRunning);
        }
        catch (SimulationPNPException spnpe) {
            // the user error message has been shown already
            spnpe.printStackTrace();
            setExperiment(Experiment.NULL);
        }
        catch (NumberFormatException nfe) {
            Util.showErrorMessage("Bad number format",
                "Bad number format " + nfe.getMessage());
        }
    }

    else {
        assert false; // Should not happen
    }

    glComponent.requestFocusInWindow();
}

/**
 * Makes sure that the toolbar buttons for adjusting the simulation speed,
 * are disabled when lower and upper speed limits are reached
 *
 */
private void validateSpeedSettings() {
    float slowingFactor = mediator.getSpeedFactor();

    if (slowingFactor >= Mediator.SIMULATION_MAX_SLOWING_FACTOR) {
        runtimeButtons.setEnabled(FASTER, false);
        runtimeButtons.setEnabled(SLOWER, true);
    }
    else if (slowingFactor <= Mediator.SIMULATION_MIN_SLOWING_FACTOR) {
        runtimeButtons.setEnabled(FASTER, true);
        runtimeButtons.setEnabled(SLOWER, false);
    }
    else {
        runtimeButtons.setEnabled(FASTER, true);
        runtimeButtons.setEnabled(SLOWER, true);
    }
}

/**
 * Makes sure toolbar buttons concerning perspective change.
 * are enabled or disabled according to the simulation and application
 * state.
 *
 */
private void validateViewSettings() {

    switch (mediator.getViewMode()) {

    case Mediator.DEFAULT_PERSPECTIVE :

        runtimeButtons.setEnabled(DEFAULT_PERSPECTIVE, false);
        runtimeButtons.setEnabled(PARTICLE_VIEW, true);
        runtimeButtons.setEnabled(ORTHOGRAPHIC_VIEW, true);

        break;

    case Mediator.PARTICLE_PERSPECTIVE :

        runtimeButtons.setEnabled(DEFAULT_PERSPECTIVE, true);
        runtimeButtons.setEnabled(PARTICLE_VIEW, true);
        runtimeButtons.setEnabled(ORTHOGRAPHIC_VIEW, true);

        break;

    case Mediator.ORTHOGRAPHIC :

        runtimeButtons.setEnabled(DEFAULT_PERSPECTIVE, true);
        runtimeButtons.setEnabled(PARTICLE_VIEW, true);
        runtimeButtons.setEnabled(ORTHOGRAPHIC_VIEW, false);

        break;

    default:

        break;
    }

}

/**
 * Makes sure toolbar buttons concerning detail level change.
 * are enabled or disabled according to the simulation and application
 * state.
 *
 */
private void validateDetailSettings() {

    int detailLevel = mediator.getDetailLevel();

    if (detailLevel == Mediator.DETAIL_LEVEL_LOW) {
        runtimeButtons.setEnabled(DETAIL_LOW, false);
        runtimeButtons.setEnabled(DETAIL_MEDIUM, true);
        runtimeButtons.setEnabled(DETAIL_HIGH, true);
    }
    else if (detailLevel == Mediator.DETAIL_LEVEL_MEDIUM) {
        runtimeButtons.setEnabled(DETAIL_LOW, true);
        runtimeButtons.setEnabled(DETAIL_MEDIUM, false);
        runtimeButtons.setEnabled(DETAIL_HIGH, true);
    }
    else if (detailLevel == Mediator.DETAIL_LEVEL_HIGH) {
        runtimeButtons.setEnabled(DETAIL_LOW, true);
        runtimeButtons.setEnabled(DETAIL_MEDIUM, true);
        runtimeButtons.setEnabled(DETAIL_HIGH, false);
    }
}

/**
 * Sets the toolbar button state (enabled or disabled) according to
 * the application and simulation state.
 *
 */
private void setButtonState(int buttonState) {

    switch (buttonState) {

    case BUTTON_STATE_INITIAL:

        validateViewSettings();
        applicationButtons.setEnabled(SNAPSHOT, true);

        break;

    case BUTTON_STATE_EXPERIMENT_LOAD:

        validateViewSettings();
        validateSpeedSettings();
        validateDetailSettings();

        runtimeButtons.setEnabled(RESET, false);
        applicationButtons.setEnabled(RESET, false);

        applicationButtons.setEnabled(RUNSTOP, true);
        // applicationButtons.setEnabled(SETTINGS, true);
        applicationButtons.setEnabled(SNAPSHOT, true);
        applicationButtons.setEnabled(VIEWDATA, true);

        break;

    case BUTTON_STATE_EXPERIMENT_RUN:

        validateViewSettings();
        validateSpeedSettings();
        validateDetailSettings();

        runtimeButtons.setEnabled(RESET, true);
        applicationButtons.setEnabled(RESET, true);
        applicationButtons.setEnabled(RUNSTOP, true);
        //applicationButtons.setEnabled(SETTINGS, true);
        applicationButtons.setEnabled(VIEWDATA, true);
        applicationButtons.setEnabled(SNAPSHOT, true);

        break;
    }
}

}


