/*
 *                        s a c e s
 *      Simple Artificial Chemistry Experiment System
 *
 * A semester project Summer 2005 and a diploma project 2006
 *           Berne University of Applied Sciences
 *   HTI School of Engineering and Information Technology
 *             Anthony Aguillon, Daniel Noelpp
 *
 */
package saces.app;

import saces.Util;
import saces.app.gui.SplashScreen;
import saces.exp.Experiment;
import saces.exp.ExperimentInvalidException;
import saces.file.BadExperimentFileException;
import saces.file.BinaryLogRetriever;
import saces.sim.Mediator;
import saces.sim.SimulationPNPException;

import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import java.awt.EventQueue;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.Reader;
import java.io.File;
import java.net.URL;
import java.nio.charset.Charset;


/**
 * The Main class contains the application's main entry point. It initializes
 * the application, loads the default experiment, processes command line
 * options, extracts jogl.dll from the jar, and loads a splash screen while the
 * Java vm is loading.
 */
public class Main {

// Splash screen while Java VM is loading
private static SplashScreen splashScreen;

/** Loads a look-and-feel */
private static boolean loadLAF(String lafClassName) {
    Throwable e;
    try {
        UIManager.setLookAndFeel(lafClassName);
        return true;
    }
    catch (ClassNotFoundException cnfe) { e = cnfe; }
    catch (IllegalAccessException iae) { e = iae; }
    catch (UnsupportedLookAndFeelException ulaf) { e = ulaf; }
    catch (InstantiationException ie) { e = ie; }

    System.err.println("loadLAF() failed: " + e);
    return false;
}

/** The application's main entry point */
public static void main(String argv[]) {
    if (System.getProperty("dontcatch") == null)
        Thread.setDefaultUncaughtExceptionHandler(
            new UncaughtExceptionHandler());

    int n = argv.length;

    // Check parameter syntax. If bad, print usage and exit (no GUI).
    if (n != 0
        && !(argv[0].equals("-nosplash") && n == 1)
        && !(argv[0].equals("-load") && n == 2)
        && !(argv[0].equals("-tail") && (n == 2 || n == 3))
        && !(argv[0].equals("-view") && (n == 2 || n == 3))
        ) {
        System.out.println(String.format(""
            + "Usage with arguments:"
            + "%n-load experiment.xml"
            + "%n\tLoad experiment instead of the default one"
            + "%n-view experiment.xml [saces.binlog]"
            + "%n\tView binary log of the experiment graphically"
            + "%n-tail experiment.xml [saces.binlog]"
            + "%n\tTail dump binary log (like the Unix tail command)"
        ));
        return;
    }

    // The -tail option doesn't need GUI, splash screen and library.
    String logFileName = n == 3 ? argv[2] : "saces.binlog";
    if ((n == 2 || n == 3) && argv[0].equals("-tail")) {
        BinaryLogRetriever.tail(argv[1], logFileName);
        return;
    }

    // The undocumented -nosplash option is used for the native bundles.
    if (!(n == 1 && argv[0].equals("-nosplash"))) {
        splashScreen = new SplashScreen();
        splashScreen.splash();
    }
    extractNativeLibrary();

    // Start up the GUI.
    Application application = initApplication();

    // Default invocation: Load the default experiment.
    if (n == 0) {
        loadDefaultExperiment(application);
        application.show();
    }
    else if (argv[0].equals("-load")) {
        loadExperiment(application, argv[1]);
        application.show();
    }
    else if (argv[0].equals("-view")) {
        loadExperiment(application, argv[1]);
        application.showViewData();
    }
    else {
        assert false; // Should not happen.
    }

    // Dispose the splash screen later.
    EventQueue.invokeLater(new Runnable() {
        public void run() { if (splashScreen != null) splashScreen.dispose(); }
    });
}

/** Initialize the application and load the empty experiment. */
private static Application initApplication() {
    String systemLAFClass = UIManager.getSystemLookAndFeelClassName();
    loadLAF(systemLAFClass);

    Application application = new Application(new Mediator());
    application.setExperiment(Experiment.NULL);

    return application;
}

/** Load a resource as a text stream. */
private static Reader getResource(String resource) {
    Class mainClass = saces.app.Main.class;
    try {
        return new InputStreamReader(
            mainClass.getResource(resource).openStream(),
            Charset.forName("ISO-8859-1")
        );
    }
    catch (IOException ioe) {}
    catch (NullPointerException npe) {}

    return null;
}

/** Load an experiment. */
private static void loadExperiment(
    Application application, String experimentFileName
) {
    try {
        Experiment experiment = new Experiment(
            new FileReader(experimentFileName));
        application.setExperiment(experiment);
        if (experiment.getValidationProblem() != null)
            Util.showErrorMessage("Problem validating experiment",
                experiment.getValidationProblem());

        return;
    }
    catch (FileNotFoundException fnfe) {
        Util.showFileNotFoundMessage(experimentFileName);
    }
    catch (ExperimentInvalidException eie) {
        eie.printStackTrace();
        Util.showErrorMessage("Internal Problem loading file",
            "Internal problem loading file\n'" + experimentFileName
                + "',\n detail: " + eie.getCause());
    }
    catch (BadExperimentFileException befe) {
        Util.showErrorMessage("Bad Experiment file",
            "The file\n'" + experimentFileName
                + "'\ncouldn't be validated as\n"
                + "a sAcEs experiment XML file:\n\n"
                + befe.getMessage());
    }
    catch (SimulationPNPException spnpe) {
        // the user error message has been shown already.
        spnpe.printStackTrace();
    }

    // if there were problems (exceptions), start the NULL experiment.
    application.setExperiment(Experiment.NULL);
}

private static void loadDefaultExperiment(Application application) {
    try {
        application.loadExperiment(getResource("DefaultExperiment.xml"));
    }
    catch (BadExperimentFileException befe) {
        assert false;
        application.setExperiment(Experiment.NULL);
    }
}

// This is an ugly kludge neccessary for bundling in Windows:
// Java cannot use native libraries inside jar files. For the native
// library to work, it must be extracted from the jar file first, then
// written to the same directory the program is in.
private static void extractNativeLibrary() {
    URL resource = Main.class.getResource("/jogl.dll");
    if (resource != null && !new File("jogl.dll").exists()) {
        System.out.println("Extracting jogl.dll...");
        try {
            InputStream is = new BufferedInputStream(resource.openStream());
            OutputStream os = new BufferedOutputStream(
                new FileOutputStream("jogl.dll"));
            int b;
            while (-1 != (b = is.read())) os.write(b);
            is.close();
            os.close();
            System.out.println("jogl.dll extraction successful");
        }
        catch (IOException ioe) {
            System.out.println("jogl.dll extraction failed: " + ioe);
        }
    }
}

}