/*
 *                        s a c e s
 *      Simple Artificial Chemistry Experiment System
 *
 * A semester project Summer 2005 and a diploma project 2006
 *           Berne University of Applied Sciences
 *   HTI School of Engineering and Information Technology
 *             Anthony Aguillon, Daniel Noelpp
 *
 */
package saces.app.gui;

import saces.Util;
import saces.exp.Experiment;
import saces.exp.ParticleClass;
import saces.exp.Reaction;

import javax.swing.table.AbstractTableModel;
import java.awt.Color;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/** A table model for the particle classes tab in the Settings dialog. */
public class ParticleClassesTableModel extends AbstractTableModel {

private static final long serialVersionUID = 8766873425949682430L;
public static final int NAME = 0;
public static final int INITIAL_COUNT = 1;
public static final int ENERGY = 2;
public static final int RADIUS = 3;
public static final int MASS = 4;
public static final int COLOR = 5;
public static final int DELETE = 6;

private static final String[] columnNames = {
    "Name", "Initial Count", "Energy", "Radius",
    "Mass", "Color"
};
private static final Class[] columnClasses = {
    String.class, Integer.class, Float.class, Float.class,
    Float.class, Color.class
};

private final List<ParticleClass> particleClasses
    = new ArrayList<ParticleClass>();
private final Set<String> particleNames = new HashSet<String>();
private List<Reaction> reactions;

/**
 * Create a table model of the particle classes.
 *
 * @param experiment Experiment (must not be null)
 */
public ParticleClassesTableModel(Experiment experiment) {
    // modifiable unsealed copies of the particle classes
    int index = 0;
    for (ParticleClass pc : experiment.getParticleClasses().values()) {
        particleClasses.add(new ParticleClass(
            index, pc.getName(), pc.getRadius(), pc.getMass(),
            pc.getEnergy(), pc.getInitialCount(), pc.getColor()));
        index++;
    }

    particleNames.addAll(experiment.getParticleClasses().keySet());
}

/** Attach the reactions to the particle classes. Neccessary because
 * if removing a particle, the reactions must be verified not to contain
 * that particle class.
 */
public void setReactionList(List<Reaction> reactions) {
    this.reactions = reactions;
}

/** Get the updated particle classes. */
public List<ParticleClass> getParticleClasses() {
    return particleClasses;
}

public int getColumnCount() { return columnNames.length; }

public int getRowCount() { return particleClasses.size(); }

public Object getValueAt(int rowIndex, int columnIndex) {
    ParticleClass pc = particleClasses.get(rowIndex);
    switch (columnIndex) {
    case NAME:
        return pc.getName();
    case INITIAL_COUNT:
        return pc.getInitialCount();
    case ENERGY:
        return pc.getEnergy();
    case RADIUS:
        return pc.getRadius();
    case MASS:
        return pc.getMass();
    case COLOR:
        return pc.getColor();
    default:
        assert false : "columnIndex=" + columnIndex;
    }
    return null;
}

public Class<?> getColumnClass(int columnIndex) {
    return columnClasses[columnIndex];
}

public String getColumnName(int columnIndex) {
    return columnNames[columnIndex];
}

public boolean isCellEditable(int rowIndex, int columnIndex) {
    return true;
}

public void setValueAt(Object value, int rowIndex, int columnIndex) {
    ParticleClass pc = particleClasses.get(rowIndex);
    String name = pc.getName();
    float radius = pc.getRadius();
    float mass = pc.getMass();
    float energy = pc.getEnergy();
    int initialCount = pc.getInitialCount();
    Color color = pc.getColor();

    switch (columnIndex) {
    case NAME:
        if (name.equals(value)) return;

        String newName = (String)value;
        if (particleNames.contains(newName)) {
            Util.showErrorMessage("Name already taken", ""
                + "You already have a particle class with same name");
            return;
        }

        particleNames.remove(name);
        particleNames.add(newName);
        name = newName;
        break;
    case INITIAL_COUNT:
        initialCount = ((Integer)value).intValue();
        break;
    case ENERGY:
        energy = ((Float)value).floatValue();
        break;
    case RADIUS:
        radius = ((Float)value).floatValue();
        break;
    case MASS:
        mass = ((Float)value).floatValue();
        break;
    case COLOR:
        color = (Color)value;
        break;
    default:
        assert false : "columnIndex=" + columnIndex;
    }

    particleClasses.set(rowIndex, new ParticleClass(
        rowIndex, name, radius, mass, energy, initialCount, color));
}

/** Add a new row/particle class to the table. */
public void addNewParticleClass() {
    String name = "NewParticle";

    int discriminatingIndex = -1;
    if (particleNames.contains(name)) {
        discriminatingIndex = 0;
        while (particleNames.contains(name + discriminatingIndex))
            discriminatingIndex++;
    }

    if (discriminatingIndex >= 0) name += discriminatingIndex;

    ParticleClass pc = new ParticleClass(
        particleClasses.size(), name, 1, 1, 1, 0, Color.black);
    particleClasses.add(pc);
    particleNames.add(name);

    fireTableStructureChanged();
}

/** Remove the current row/particle class from the table. */
public void deleteParticleClass(int index) {
    if (index < 0) {
        Util.beep();
        return;
    }

    ParticleClass particleClass = particleClasses.get(index);
    for (Reaction reaction : reactions) {
        if (reaction.getReagents().contains(particleClass)) {
            Util.showErrorMessage("Particle class used in a reaction",
                "The particle class " + particleClass.getName()
                    + " is used in the reaction " + reaction.getDescription()
                    + ".\nPlease delete or modify the reaction first.");

            return;
        }
    }

    // Reindex the particle classes (somewhat inefficient)
    index = 0;
    for (ParticleClass pc : particleClasses) {
        particleClasses.set(index, new ParticleClass(
            index, pc.getName(), pc.getRadius(), pc.getMass(), pc.getEnergy(),
            pc.getInitialCount(), pc.getColor()));
        index++;
    }

    particleNames.remove(particleClass.getName());
    particleClasses.remove(index);
    fireTableStructureChanged();
}

public String toString() {
    return particleClasses.toString();
}
}
