/*
 *                        s a c e s
 *      Simple Artificial Chemistry Experiment System
 *
 * A semester project Summer 2005 and a diploma project 2006
 *           Berne University of Applied Sciences
 *   HTI School of Engineering and Information Technology
 *             Anthony Aguillon, Daniel Noelpp
 *
 */
package saces.app.gui;

import saces.Util;
import saces.exp.ExperimentInvalidException;
import saces.exp.ParticleClass;
import saces.exp.Reaction;

import javax.swing.table.AbstractTableModel;
import java.util.ArrayList;
import java.util.List;


/** A table model for reactions in the Settings dialog window. */
class ReactionTableModel extends AbstractTableModel {

private static final long serialVersionUID = 3425942685438876082L;
public static final int DESCRIPTION = 0;
public static final int EQUATION = 1;
public static final int PROBABILITY = 2;
public static final int ACTIVATION_ENERGY = 3;

private static final String[] columnNames = {
    "Description", "Equation", "Probability", "Activation Energy",
};
private static final Class[] columnClasses = {
    String.class, String.class, Float.class, Float.class,
};

private final ParticleClassesTableModel pctm;
private final List<Reaction> reactions = new ArrayList<Reaction>();

/**
 * Create a ReactionTableModel. The ParticleClassesTableModel is needed to
 * verify reaction equations entered in the table.
 */
public ReactionTableModel(
    List<Reaction> reactions, ParticleClassesTableModel pctm
) {
    this.pctm = pctm;

    // modifiable unsealed copies of the reactions
    int index = 0;
    for (Reaction reaction : reactions) {
        this.reactions.add(new Reaction(
            index, reaction.getDescription(), reaction.getProbability(),
            reaction.getActivationEnergy(), reaction.getEquation(),
            pctm.getParticleClasses()));
        index++;
    }
}

/** Get the reactions from the table. */
public List<Reaction> getReactions() { return reactions; }

public int getColumnCount() { return columnNames.length; }

public int getRowCount() { return reactions.size(); }

public Object getValueAt(int rowIndex, int columnIndex) {
    Reaction reaction = reactions.get(rowIndex);

    switch (columnIndex) {
    case DESCRIPTION:
        return reaction.getDescription();
    case EQUATION:
        return reaction.getEquation();
    case PROBABILITY:
        return reaction.getProbability();
    case ACTIVATION_ENERGY:
        return reaction.getActivationEnergy();
    default:
        assert false : "columnIndex=" + columnIndex;
    }
    return null;
}

public Class<?> getColumnClass(int columnIndex) {
    return columnClasses[columnIndex];
}

public String getColumnName(int columnIndex) {
    return columnNames[columnIndex];
}

public boolean isCellEditable(int rowIndex, int columnIndex) {
    return true;
}

public void setValueAt(Object value, int rowIndex, int columnIndex) {
    Reaction reaction = reactions.get(rowIndex);
    String description = reaction.getDescription();
    String equation = reaction.getEquation();
    float probability = reaction.getProbability();
    float activationEnergy = reaction.getActivationEnergy();

    switch (columnIndex) {
    case DESCRIPTION:
        description = (String)value;
        break;
    case EQUATION:
        equation = (String)value;
        break;
    case ACTIVATION_ENERGY:
        activationEnergy = (Float)value;
        break;
    case PROBABILITY:
        probability = (Float)value;
        break;
    default:
        assert false;
    }

    // A hack for catching probabilities > 1
    try {
        reactions.set(rowIndex, new Reaction(
            rowIndex, description, probability, activationEnergy,
            equation, pctm.getParticleClasses()));
    }
    catch (ExperimentInvalidException eie) {
        if (eie.getMessage().indexOf("probability") < 0) throw eie;
        setValueAt(0f, rowIndex, columnIndex);
        Util.beep();
    }

}

/** Add a new row/reaction to the table. */
public void addNewReaction() {
    if (pctm.getRowCount() < 1) {
        Util.showErrorMessage("Create particles first",
            "Please create particles first");
        return;
    }

    int index = reactions.size();
    String name = pctm.getParticleClasses().get(0).getName();
    reactions.add(new Reaction(index, "New Reaction", 1, 0,
        name + "+" + name + "=" + name + "+" + name, pctm.getParticleClasses()
    ));
    fireTableStructureChanged();
}

/** Delete the selected row/reaction from the table. */
public void deleteReaction(int index) {
    if (index < 0) {
        Util.beep();
        return;
    }

    Reaction oldReaction = reactions.remove(index);
    fireTableStructureChanged();

    // Reindex the reactions
    index = 0;
    for (Reaction reaction : reactions) {
        reactions.set(index, new Reaction(
            index, reaction.getDescription(), reaction.getProbability(),
            reaction.getActivationEnergy(), reaction.getEquation(),
            pctm.getParticleClasses()
        ));
        index++;
    }

    // Clear the reaction from all particle classes containing it.
    for (ParticleClass pc : pctm.getParticleClasses()) {
        pc.clearReaction(oldReaction);
    }
}

}
