/*
 *                        s a c e s
 *      Simple Artificial Chemistry Experiment System
 *
 * A semester project Summer 2005 and a diploma project 2006
 *           Berne University of Applied Sciences
 *   HTI School of Engineering and Information Technology
 *             Anthony Aguillon, Daniel Noelpp
 *
 */
package saces.app.gui;

import saces.Util;
import saces.app.Application;
import static saces.app.gui.SettingsButtonPanel.*;
import saces.app.gui.table.CardinalNumberCellEditor;
import saces.app.gui.table.ColorCellEditor;
import saces.app.gui.table.ColorCellRenderer;
import saces.app.gui.table.NonnegativeNumberCellEditor;
import saces.app.gui.table.PositiveNumberCellEditor;
import saces.exp.Experiment;
import saces.exp.ExperimentInvalidException;
import saces.file.BadExperimentFileException;
import saces.file.ExperimentSaver;
import saces.sim.SimulationPNPException;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusListener;
import java.awt.event.FocusEvent;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;


/** The settings dialog window. */
public class SettingsFrame extends JFrame
    implements ActionListener, FocusListener {

private static final long serialVersionUID = 6989592873118972460L;

private static final String NEWPARTICLE = "NewParticle";
private static final String DELETEPARTICLE = "DeleteParticle";
private static final String NEWREACTION = "NewReaction";
private static final String DELETEREACTION = "DeleteReaction";
private static final String NEWPROPERTY = "NewProperty";
private static final String DELETEPROPERTY = "DeleteProperty";

private JComponent root;
private SettingsButtonPanel settingsButtonPanel;

// Experiment tab
private JTextField descriptionField;
private JTextField initialTemperatureField;
private ColorField colorField;
private JTextField timeStepField;
private JTextField randomSeedField;
private JTextField heightField;
private JTextField widthField;
private JTextField depthField;

// Particles tab
private JTable particlesTable;
private ParticleClassesTableModel pctm;
private JButton newParticleButton;
private JButton deleteParticleButton;

// Reactions tab
private JTable reactionTable;
private ReactionTableModel rtm;
private JButton newReactionButton;
private JButton deleteReactionButton;

// Properties tab
private JTable propertiesTable;
private PropertiesTableModel ptm;
private JButton newPropertyButton;
private JButton deletePropertyButton;

private Application application;
private Experiment experiment;
private File experimentFile = new File(".");

private boolean inputFieldBlocked = false;

/** Create a settings window dialog attached to application. */
public SettingsFrame(Application application) {
    super("s A c E s :: Settings");
    this.application = application;

    getContentPane().add(root);
    setBounds(100, 20, 400, 250);
    setDefaultCloseOperation(HIDE_ON_CLOSE);

    colorField.setToolTipText("Color of the experiment box");
    settingsButtonPanel.addActionListener(this);
    settingsButtonPanel.setEnabled(SAVE, false);

    // --- to set inputFieldBlocked when focused
    initialTemperatureField.addFocusListener(this);
    timeStepField.addFocusListener(this);
    widthField.addFocusListener(this);
    heightField.addFocusListener(this);
    depthField.addFocusListener(this);
    randomSeedField.addFocusListener(this);

    // --- String up the new and delete buttons
    newParticleButton.addActionListener(this);
    newParticleButton.setActionCommand(NEWPARTICLE);
    newReactionButton.addActionListener(this);
    newReactionButton.setActionCommand(NEWREACTION);
    newPropertyButton.addActionListener(this);
    newPropertyButton.setActionCommand(NEWPROPERTY);
    deleteParticleButton.addActionListener(this);
    deleteParticleButton.setActionCommand(DELETEPARTICLE);
    deleteReactionButton.addActionListener(this);
    deleteReactionButton.setActionCommand(DELETEREACTION);
    deletePropertyButton.addActionListener(this);
    deletePropertyButton.setActionCommand(DELETEPROPERTY);

    // --- Preinitialize the particles table
    particlesTable.setDefaultEditor(Color.class, new ColorCellEditor());
    particlesTable.setDefaultEditor(Integer.class,
        new CardinalNumberCellEditor());
    particlesTable.setDefaultEditor(Float.class,
        new PositiveNumberCellEditor());
    particlesTable.setDragEnabled(false);
    particlesTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

    // --- Preinitialize the reactions table
    reactionTable.setDragEnabled(false);
    reactionTable.setDefaultEditor(Float.class,
        new NonnegativeNumberCellEditor());
    reactionTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

    // --- Preinitialize the properties table
    propertiesTable.setDragEnabled(false);
    propertiesTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
}

/** Get the current experiment from the dialog. */
public Experiment getExperiment() {
    return experiment;
}

/** Set the experiment to be edited in the dialog. */
public void setExperiment(Experiment experiment) {
    this.experiment = experiment;

    // --- Initialize the Experiment tab
    descriptionField.setText(experiment.getDescription());
    initialTemperatureField.setText("" + experiment.getInitialTemperature());
    initialTemperatureField.setInputVerifier(new NumberVerifier(0f));
    colorField.setColor(experiment.getColor());
    timeStepField.setText("" + experiment.getTimestep());
    timeStepField.setInputVerifier(new NumberVerifier(.0001f));
    randomSeedField.setText(experiment.hasRandomSeed()
        ? ("" + experiment.getRandomSeed()) : "");
    randomSeedField.setInputVerifier(NumberVerifier.emptyAllowed(true));
    heightField.setText("" + experiment.getHeight());
    heightField.setInputVerifier(new NumberVerifier(0f));
    widthField.setText("" + experiment.getWidth());
    widthField.setInputVerifier(new NumberVerifier(0f));
    depthField.setText("" + experiment.getDepth());
    depthField.setInputVerifier(new NumberVerifier(0f));

    // --- Initialize the Particles tab
    pctm = new ParticleClassesTableModel(experiment);
    particlesTable.setModel(pctm);
    particlesTable.setDefaultRenderer(Color.class, new ColorCellRenderer(pctm));

    // --- Initialize the Reactions tab
    rtm = new ReactionTableModel(experiment.getReactions(), pctm);
    reactionTable.setModel(rtm);
    pctm.setReactionList(rtm.getReactions());

    // --- Initialize the Properties tab
    ptm = new PropertiesTableModel(
        new HashMap<String, String>(experiment.getProperties()));
    propertiesTable.setModel(ptm);

}

public void focusGained(FocusEvent e) {
    inputFieldBlocked = true;
}

public void focusLost(FocusEvent e) {
    inputFieldBlocked = false;
}

/** Handle button clicks of the tool bar, like loading, saving etc. */
public void actionPerformed(ActionEvent e) {
    String cmd = e.getActionCommand();

    if (cmd.equals(LOAD)) {
        JFileChooser fileChooser = new JFileChooser(experimentFile);

        int result = fileChooser.showOpenDialog(this);
        if (result == JFileChooser.APPROVE_OPTION) {
            experimentFile = fileChooser.getSelectedFile();
            try {
                experiment = new Experiment(new FileReader(experimentFile));
                setExperiment(experiment);
                settingsButtonPanel.setEnabled(SAVE, true);
                if (experiment.getValidationProblem() == null) return;

                Util.showErrorMessage("Problem validating experiment",
                    experiment.getValidationProblem());
            }
            catch (FileNotFoundException fnfe) {
                Util.showFileNotFoundMessage(experimentFile.getPath());
            }
            catch (ExperimentInvalidException eie) {
                eie.printStackTrace();
                Util.showErrorMessage("Internal Problem loading file",
                    "Internal problem loading file\n'"
                        + experimentFile.getPath() + "',\n detail: "
                        + eie.getMessage());
            }
            catch (BadExperimentFileException befe) {
                Util.showErrorMessage("Bad Experiment file",
                    "The file\n'" + experimentFile.getPath()
                        + "'\ncouldn't be validated as\n"
                        + "a sAcEs experiment XML file:\n\n"
                        + befe.getMessage());
            }
            catch (SimulationPNPException spnpe) {
                // the user error message has been shown already
                spnpe.printStackTrace();
            }
            // NOTE: the try-catch block has some code duplication in
            // class Main.

            // if there were problems, set the experiment to NULL.
            application.setExperiment(Experiment.NULL);
        }
    }

    else if (cmd.equals(START)) {
        if (inputFieldBlocked) {
            Util.beep();
            return;
        }

        if (!Util.showYesNo("Reset simulation?",
            "This will reset the current simulation\n"
                + "AND the binary log.\n\n"
                + "Do you really want to start a simulation\n"
                + "with the new settings?")) return;

        readFromGUI();
        try {
            application.setExperiment(experiment);
            application.setRunning(true);
            setVisible(false);
        }
        catch (SimulationPNPException spnpe) {
            // the user error message has been shown already
            spnpe.printStackTrace();
            application.setExperiment(Experiment.NULL);
        }
        catch (NumberFormatException nfe) {
            Util.showErrorMessage("Bad number format",
                "Bad number format " + nfe.getMessage());
        }
    }

    else if (cmd.equals(SAVE)) {
        if (inputFieldBlocked) {
            Util.beep();
            return;
        }

        readFromGUI();
        saveExperiment();
    }

    else if (cmd.equals(SAVEAS)) {
        readFromGUI();
        JFileChooser fileChooser = new JFileChooser(experimentFile);

        int result = fileChooser.showSaveDialog(this);
        if (result == JFileChooser.APPROVE_OPTION) {
            experimentFile = fileChooser.getSelectedFile();
            saveExperiment();
        }
    }

    else if (cmd.equals(NEWPARTICLE)) {
        pctm.addNewParticleClass();
    }

    else if (cmd.equals(DELETEPARTICLE)) {
        pctm.deleteParticleClass(particlesTable.getSelectedRow());
    }

    else if (cmd.equals(NEWREACTION)) {
        rtm.addNewReaction();
    }

    else if (cmd.equals(DELETEREACTION)) {
        rtm.deleteReaction(reactionTable.getSelectedRow());
    }

    else if (cmd.equals(NEWPROPERTY)) {
        ptm.addNewProperty();
    }

    else if (cmd.equals(DELETEPROPERTY)) {
        ptm.deleteProperty(propertiesTable.getSelectedRow());
    }

    else {
        assert false; // Should not happen
    }
}

private void saveExperiment() {
    try {
        ExperimentSaver.saveAsXML(experiment, experimentFile);
    }
    catch (IOException ioe) {
        Util.showErrorMessage("Experiment not saved",
            "Experiment could not be saved to file\n"
                + experimentFile.getPath() + ":\n\n"
                + ioe.getMessage());
    }
}

private void readFromGUI() {
    try {
        boolean hasSeed = randomSeedField.getText().length() > 0;
        experiment = new Experiment(
            descriptionField.getText(),
            Float.parseFloat(initialTemperatureField.getText()),
            Float.parseFloat(timeStepField.getText()),
            Float.parseFloat(widthField.getText()),
            Float.parseFloat(heightField.getText()),
            Float.parseFloat(depthField.getText()),
            colorField.getColor(),
            hasSeed,
            hasSeed ? Integer.parseInt(randomSeedField.getText()) : 0,
            pctm.getParticleClasses(),
            rtm.getReactions(),
            ptm.getProperties()
        );
        setExperiment(experiment);
    }
    catch (UnsupportedOperationException uoe) {
        Util.showErrorMessage("Bad experiment", uoe.getMessage());
        uoe.printStackTrace();
        experiment = Experiment.NULL;
    }
}

public void setInputFieldBlocked(boolean inputFieldBlocked) {
    this.inputFieldBlocked = inputFieldBlocked;
}

}
