/*
 *                        s a c e s
 *      Simple Artificial Chemistry Experiment System
 *
 * A semester project Summer 2005 and a diploma project 2006
 *           Berne University of Applied Sciences
 *   HTI School of Engineering and Information Technology
 *             Anthony Aguillon, Daniel Noelpp
 *
 */
package saces.app.gui;

import static saces.app.gui.ViewDataModel.Source.*;
import saces.exp.Experiment;
import static saces.file.BinaryLog.*;
import saces.file.TailFileInputStream;

import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.LinkedList;

//Note: Contains some duplicate code as BinaryLogRetriever!

/** Data read from the binary log file is saved in this class. */
public class ViewDataModel extends ChangeSupport implements Runnable {

/** What is the source of a view data model change? */
public enum Source {

    ParticleClasses, Reactions, Measurements, Histogram, OtherData,
    NewExperiment;
}

private static abstract class Data {

    public final long timeStamp;
    public final String dateTime;

    public Data(long timeStamp, String dateTime) {
        this.timeStamp = timeStamp;
        this.dateTime = dateTime;
    }
}

/** A Data instance about a reaction. */
public static class ReactionData extends Data {

    public final int reactionIndex;
    public ReactionData(
        long timeStamp, String dateTime, int reactionIndex
    ) {
        super(timeStamp, dateTime);
        this.reactionIndex = reactionIndex;
    }
}

/** A Data instance about particle's particle classes. */
public static class ParticleData extends Data {

    public final int totalCount;
    public final int[] counts; // using particle class indices

    // the parameter counts must not be modified after constructor call
    public ParticleData(
        long timeStamp, String dateTime, int totalCount, int[] counts
    ) {
        super(timeStamp, dateTime);
        this.totalCount = totalCount;
        this.counts = counts;
    }
}

/** A Data instance about a speed distribution of the particles. */
public static class Histogram extends Data {

    public final float min, avg, max, top;
    public final int[] histogram;

    // the parameter histogram must not be modified after constructor call
    public Histogram(
        long timeStamp, String dateTime,
        float min, float avg, float max, float top,
        int[] histogram
    ) {
        super(timeStamp, dateTime);
        this.min = min;
        this.avg = avg;
        this.max = max;
        this.top = top;
        this.histogram = histogram;
    }
}

/** A Data instance about temperature. */
public static class Temperature extends Data {

    public final float temperature;

    public Temperature(
        long timeStamp, String dateTime, float temperature
    ) {
        super(timeStamp, dateTime);
        this.temperature = temperature;
    }
}

/** A Data instance about pressure. */
public static class Pressure extends Data {

    public final float pressure;

    public Pressure(
        long timeStamp, String dateTime, float pressure
    ) {
        super(timeStamp, dateTime);
        this.pressure = pressure;
    }
}

private static final long MASK = 0xfL;

private Experiment experiment = Experiment.NULL;
private DataInputStream dis;
private Thread readerThread;

private String dateTime = "";
private StringBuilder otherDataText;
private LinkedList<ReactionData> reactionData;
private LinkedList<ParticleData> particleData;
private LinkedList<Histogram> histograms;
private LinkedList<Temperature> temperatures;
private LinkedList<Pressure> pressures;

private int maxTotalParticleCount;
private float maxTemperature;
private float maxPressure;

/** Get the experiment the view data model is about. */
public Experiment getExperiment() { return experiment; }

/** Create a null, empty data view model. */
public ViewDataModel() {}

/** Create a view data model and start reading the binary log asynchronously. */
public ViewDataModel(Experiment experiment) {
    setExperiment(experiment);
}

/** Set the experiment. Reset the view data model and reread. */
public void setExperiment(Experiment experiment) {
    this.experiment = experiment;

    otherDataText = new StringBuilder();
    reactionData = new LinkedList<ReactionData>();
    particleData = new LinkedList<ParticleData>();
    histograms = new LinkedList<Histogram>();
    temperatures = new LinkedList<Temperature>();
    pressures = new LinkedList<Pressure>();

    maxTotalParticleCount = -1; // almost zeor (avoid division by zero)
    maxTemperature = .01f;
    maxPressure = .01f;

    try {
        String file = experiment.getString("BinaryLog", "saces.binlog");
        dis = new DataInputStream(
            new TailFileInputStream(new FileInputStream(file)));
    }
    catch (IOException ioe) { ioException(ioe); }

    if (readerThread != null) readerThread.interrupt();
    readerThread = new Thread(this);
    readerThread.start();

    fireStateChanged(NewExperiment);
}

/** Get the datetime string of the most recent data instance. */
public String getDateTime() { return dateTime; }

/**
 * Get the binary log as a string (except histograms and particle class
 * distributions)
 */
public String getOtherDataText() { return otherDataText.toString(); }

public LinkedList<ReactionData> getReactionData() { return reactionData; }

public LinkedList<ParticleData> getParticleData() { return particleData; }

public LinkedList<Histogram> getHistograms() { return histograms; }

public LinkedList<Temperature> getTemperatures() { return temperatures; }

public LinkedList<Pressure> getPressures() { return pressures; }

/** The maxmimum number of particles encountered so far. */
public float getMaxTotalParticleCount() { return maxTotalParticleCount; }

/** The maximum pressure encountered so far. */
public float getMaxPressure() { return maxPressure; }

/** The maximum temperature encountered so far. */
public float getMaxTemperature() { return maxTemperature; }

/** Asynchronously read from the binary log. Don't call. */
public void run() {
    try { Thread.sleep(250); } catch (InterruptedException ie) {}
    while (!Thread.interrupted()) readRecord();
    readerThread = null;
}

// NOTE: some code duplication with BinaryLogRetriever
private void ioException(IOException ioe) {
    otherDataText.append("\tIOEXCEPTION\t").append(ioe).append('\n');
    close();
}

private void error(String msg) {
    otherDataText.append("\tERROR\t").append(msg).append('\n');
    close();
}

private void close() {
    otherDataText.append("\tCLOSE\tBinary log file closed\n");
    if (dis != null)
        try { dis.close(); } catch (IOException ioe) {}
    fireStateChanged(OtherData);
    if (readerThread != null) readerThread.interrupt();
}

private int getRecordType() throws IOException {
    if (dis == null) return -1;
    long recordType = dis.readLong();
    if ((recordType & ~MASK ^ MAGIC) != 0L) {
        error(String.format("bad record type code 0x%016xL", recordType));
        return -1;
    }
    return (int)(recordType & MASK);
}

// NOTE: some duplicate code with BinaryLogRetriever.readRecord()
private void readRecord() {
    int pcCount = experiment.getParticleClassesCount();

    try {
        int recordType = getRecordType();
        if (recordType == -1) {
            try { Thread.sleep(100); } catch (InterruptedException ie) {}
            return;
        }

        long timeStamp = dis.readLong();
        dateTime = String.format("%1$TFT%1$TT.%1$TLUTC", timeStamp);

        switch (recordType) {
        case RECORD_MAGIC:
            otherDataText.append(dateTime).append("\tLOG_FILE_START\n");
            fireStateChanged(OtherData);
            return;

//        case RECORD_REACTION:
//            int reactionIndex = dis.readInt();
//            reactionData.add(
//                new ReactionData(timeStamp, dateTime, reactionIndex));
//            fireStateChanged(Reactions);
//            return;

        case RECORD_PARTICLECLASSES:
            int[] counts = new int[pcCount];
            int totalCount = 0;
            for (int i = 0; i < pcCount; i++) {
                counts[i] = dis.readInt();
                totalCount += counts[i];
            }
            if (maxTotalParticleCount < totalCount)
                maxTotalParticleCount = totalCount;

            particleData.add(new ParticleData(
                timeStamp, dateTime, totalCount, counts));
            fireStateChanged(ParticleClasses);

            // ignore padding of 4 bytes if number of particle classes is odd
            if (1 == pcCount % 2) dis.readInt();
            return;

        case RECORD_SNAPSHOT:
            int particleCount = dis.readInt();
            otherDataText.append(dateTime)
                .append('\t').append("Snapshot ParticleCount")
                .append('\t').append(particleCount).append('\n');
            for (int i = 0; i < particleCount; i++) {
                int pcIndex = dis.readInt();
                String pcName = experiment.getParticleClass(pcIndex).getName();
                otherDataText
                    .append('\t').append(pcName)
                    .append('\t').append(dis.readFloat())
                    .append('\t').append(dis.readFloat())
                    .append('\t').append(dis.readFloat())
                    .append('\t').append(dis.readFloat())
                    .append('\t').append(dis.readFloat())
                    .append('\t').append(dis.readFloat())
                    .append('\n');
            }

            // ignore padding of 4 bytes if number of particles is even
            if (0 == particleCount % 2) dis.readInt();
            return;

        case RECORD_FLOATPROPERTY:
        case RECORD_INTPROPERTY:
        case RECORD_STRINGPROPERTY:
            String key = dis.readUTF();
            Object value
                = recordType == RECORD_FLOATPROPERTY
                ? dis.readFloat()
                : recordType == RECORD_INTPROPERTY
                    ? dis.readInt()
                    : dis.readUTF();

            if (key.equals("Temperature")
                && recordType == RECORD_FLOATPROPERTY) {
                float temper = (Float)value;
                temperatures.add(new Temperature(timeStamp, dateTime, temper));
                if (temper > maxTemperature) maxTemperature = temper;
                fireStateChanged(Measurements);
            }
            if (key.equals("Pressure")
                && recordType == RECORD_FLOATPROPERTY) {
                float pressure = (Float)value;
                pressures.add(new Pressure(timeStamp, dateTime, pressure));
                if (pressure > maxPressure) maxPressure = pressure;
                fireStateChanged(Measurements);
            }

            otherDataText.append(dateTime).append('\t')
                .append(key).append('\t').append(value).append('\n');
            fireStateChanged(OtherData);

            // ignore padding to a length of a multiple of 8 bytes
            int length = getUTFLength(key);
            length += recordType == RECORD_STRINGPROPERTY
                ? getUTFLength(value.toString()) : 4;
            int padLength = getPadLength(length);
            for (int i = 0; i < padLength; i++) dis.readByte();
            return;

        case RECORD_HISTOGRAM:
            float min = dis.readFloat();
            float avg = dis.readFloat();
            float max = dis.readFloat();
            float top = dis.readFloat();
            int size = dis.readInt();

            int[] histogram = new int[size];
            for (int i = 0; i < size; i++) histogram[i] = dis.readInt();
            histograms.add(new Histogram(
                timeStamp, dateTime, min, avg, max, top, histogram));
            fireStateChanged(Histogram);

            // ignore padding of 4 bytes if number of intervals is even
            if (0 == size % 2) dis.readInt();
            return;

        default:
            error("bad record type " + recordType);
            return;

        }
    }
    catch (IOException ioe) { ioException(ioe); }
}

private int getUTFLength(String s) throws IOException {
    return s.getBytes("UTF-8").length + 2;
}

private int getPadLength(int length) {
    final int[] padTable = {0, 7, 6, 5, 4, 3, 2, 1};
    return padTable[length % 8];
}
}