/*
 *                        s a c e s
 *      Simple Artificial Chemistry Experiment System
 *
 * A semester project Summer 2005 and a diploma project 2006
 *           Berne University of Applied Sciences
 *   HTI School of Engineering and Information Technology
 *             Anthony Aguillon, Daniel Noelpp
 *
 */
package saces.exp;

import saces.Util;
import saces.file.BadExperimentFileException;
import saces.file.EasyXML;

import java.awt.Color;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;


/**
 * An immutable Experiment class with constant initial parameters like initial
 * pressure, text box dimensions, the particle classes and the number of
 * particles and finally the reaction rules. Dynamic changing parameters are
 * found in the class Simulation.
 */
public final class Experiment {

/** A null experiment (see Null Design Pattern) */
public final static Experiment NULL = new Experiment(
    "<none>", 0f, 0f, 0f, 0f, 0f, Color.black, false, 0,
    Collections.<ParticleClass>emptyList(),
    Collections.<Reaction>emptyList(),
    Collections.<String, String>emptyMap());

private final String description;
private final float initialTemperature;
private final float timestep;
private final Random random;
private final int randomSeed;
private final boolean hasSeed;
private final float width;
private final float height;
private final float depth;
private final Color color;
private final Map<String, ParticleClass> particleClasses;
private final ParticleClass[] particleClassArray;
private final List<Reaction> reactions;
private final Map<String, String> properties;
private final String validationProblem;

/**
 * Create an Experiment instance from an XML file given in the reader. There are
 * two ways in which the creation can fail: with an ExperimentInvalidException
 * for internal problems and with a validation problem (if mass preservation is
 * enabled, but some reactions violate mass preservation).
 *
 * @param reader The reader with the XML data for the Experiment instance
 * @throws BadExperimentFileException if the XML file has a bad format (it is
 * validated by XML Schema).
 * @throws ExperimentInvalidException if there is an internal problem (like bad
 * XML file [should not happen because it has been verified])
 */
public Experiment(Reader reader) throws BadExperimentFileException {
    this(new EasyXML(reader));
}

/**
 * Create an Experiment instance from an EasyXML instance. There are two ways in
 * which the creation can fail: with an ExperimentInvalidException for internal
 * problems and with a validation problem (if mass preservation is enabled, but
 * some reactions violate mass preservation).
 *
 * @param xml The EasyXML XML wrapper instance used to get the experiment
 * parameters from an XML file.
 * @throws ExperimentInvalidException if there is an internal problem (like bad
 * XML file [should not happen because it has been verified])
 */
public Experiment(EasyXML xml) {
    // properties first because the properties might influence the
    // interpretation of other data (for example whether invariants
    // are checked or not for reactions).
    properties = Collections.unmodifiableMap(readProperties(xml));

    description = xml.getString("Description");
    initialTemperature = xml.getFloat("InitialTemperature", 0f);
    timestep = xml.getFloat("TimeStep");
    width = xml.getFloat("Width");
    height = xml.getFloat("Height");
    depth = xml.getFloat("Depth");
    color = xml.getColor("Color");

    String seed = xml.getString("RandomSeed");
    if (seed.equals("")) {
        hasSeed = false;
        randomSeed = 0;
        random = new Random();
    }
    else {
        hasSeed = true;
        randomSeed = xml.getInt("RandomSeed");
        random = new Random(randomSeed);
    }

    particleClasses = Collections.unmodifiableMap(readParticleClasses(xml));
    particleClassArray = particleClasses.values().toArray(new ParticleClass[0]);
    reactions = Collections.unmodifiableList(readReactions(xml));

    // Search for a validation problem in reactions
    validationProblem = findValidationProblem();

    // Take care that the collection data structures aren't modifiable
    // even indirectly through their backing stores.
    seal();
}

/**
 * Create an Experiment instance specifying all experiment parameters (at once
 * because Experiment instances are immutable and it is not possible to change
 * the parameters after the construction of an instance).  There are two ways in
 * which the creation can fail: with an ExperimentInvalidException for internal
 * problems and with a validation problem (if mass preservation is enabled, but
 * some reactions violate mass preservation).
 *
 * @param description Experiment description
 * @param initialTemperature Initial pressure in the experiment box
 * @param timestep The "speed" at which the experiment will be running
 * @param width The width of the experiment box
 * @param height The height of  the experiment box
 * @param depth The depth of the experiment box
 * @param color The color of the experiment box walls
 * @param hasSeed Whether a random seed is to be used or not
 * @param randomSeed The random seed for repeatable experiments
 * @param particleClasses The particle classes
 * @param reactions The reaction rules
 * @param properties Additional parameters in key-value pairs
 * @throws ExperimentInvalidException if there is an internal problem (like bad
 * XML file [should not happen because it has been verified])
 */
public Experiment(
    String description, float initialTemperature,
    float timestep, float width, float height, float depth,
    Color color, boolean hasSeed, int randomSeed,
    List<ParticleClass> particleClasses,
    List<Reaction> reactions,
    Map<String, String> properties
) {
    this.description = description;
    this.initialTemperature = initialTemperature;
    this.timestep = timestep;
    this.width = width;
    this.height = height;
    this.depth = depth;
    this.color = color;

    this.hasSeed = hasSeed;
    this.randomSeed = randomSeed;
    if (hasSeed) random = new Random(randomSeed);
    else
        random = new Random();

    Map<String, ParticleClass>pcs = new LinkedHashMap<String, ParticleClass>();
    int index;

    // Verify the indices of the particle classes.
    index = 0;
    for (ParticleClass pc : particleClasses) {
        if (pc.getIndex() != index) throw new ExperimentInvalidException(""
            + "ParticleClass " + pc.getName() + " must have index "
            + index + " and not " + pc.getIndex());
        index++;
        pcs.put(pc.getName(), pc);
    }

    // Verify the indices and the reagents of the reactions (existence).
    index = 0;
    for (Reaction reaction : reactions) {
        if (reaction.getIndex() != index) throw new ExperimentInvalidException(
            "Reaction " + reaction.getDescription() + " must have index "
                + index + " and not " + reaction.getIndex());

        for (ParticleClass pc : reaction.getReagents())
            if (!pcs.containsValue(pc)) throw new ExperimentInvalidException(
                "Reaction " + reaction.getDescription()
                    + " uses nonexistent particle class " + pc);

        index++;
    }

    this.particleClasses = Collections.unmodifiableMap(pcs);
    this.particleClassArray = pcs.values().toArray(new ParticleClass[0]);
    this.reactions = Collections.unmodifiableList(
        new ArrayList<Reaction>(reactions));
    this.properties = Collections.unmodifiableMap(
        new HashMap<String, String>(properties));

    // Search for a validation problem in reactions
    validationProblem = findValidationProblem();

    // register the reactions with the particle classes for fast lookup
    for (ParticleClass pc : particleClasses)
        pc.clearReactions();
    for (Reaction reaction : reactions)
        reaction.registerReaction();

    // Take care that the collection data structures aren't modifiable
    // even indirectly through their backing stores.
    seal();
}

private String findValidationProblem() {
    String validationProblem = null;
    for (Reaction reaction : reactions) {
        validationProblem = reaction.getValidationProblem();
        if (validationProblem != null) break;
    }
    return validationProblem;
}

/** To make everything in the experiment unmodifiable */
private void seal() {
    for (ParticleClass pc : particleClasses.values()) pc.seal();

    for (int index = 0; index < particleClasses.size(); index++)
        assert index == particleClassArray[index].getIndex() : index;
}

/** Get the experiment description */
public String getDescription() { return description; }

/** Get the initial pressure in the experiment box */
public float getInitialTemperature() { return initialTemperature; }

/** Get the timestep, the "speed" at which the experiment will be running */
public float getTimestep() { return timestep; }

/** Get the color of the experiment box walls */
public Color getColor() { return color; }

/** Get the width of the experiment box */
public float getWidth() { return width; }

/** Get the height of  the experiment box */
public float getHeight() { return height; }

/** The depth of the experiment box */
public float getDepth() { return depth; }

/**
 * Get the random number generator for the experiment. Important for repeatable
 * experiments.
 */
public Random getRandom() { return random; }

/**
 * Is a random seed is to be used or not? For repeatable experiments this is
 * true.
 */
public boolean hasRandomSeed() { return hasSeed; }

/**
 * Get the random seed for repeatable experiments. Note, that the value is int,
 * not long.
 */
public int getRandomSeed() { return randomSeed; }

/** An unmodifiable map of particle classes of the experiment. */
public Map<String, ParticleClass> getParticleClasses() {
    return particleClasses;
}
/** Get a particle class with its name */
public ParticleClass getParticleClass(String name) {
    return particleClasses.get(name);
}

/** Get a particle class with its index. */
public ParticleClass getParticleClass(int index) {
    return particleClassArray[index];
}

/** Get the number of particle classes in this experiment. Convenience. */
public int getParticleClassesCount() {
    return particleClasses.size();
}

/** An unmodifiable list of reactions of the experiment. */
public List<Reaction> getReactions() {
    return reactions;
}

/** Get a list of properties keys. */
public List<String> getKeys() {
    return new ArrayList<String>(properties.keySet());
}

/** Get a string property with the key or null if no such property is defined. */
public String getString(String key) {
    return properties.get(key);
}

/** Get the string property or a default value. */
public String getString(String key, String defaultValue) {
    String value = properties.get(key);
    return value == null ? defaultValue : value;
}

/**
 * Return a boolean property. The value is true only if the strings match "1" or
 * "true" (case insensitive).
 */
public boolean getBoolean(String key, boolean defaultValue) {
    String value = properties.get(key);
    if (value == null) return defaultValue;
    if (value.equals("1")) return true;
    return Boolean.parseBoolean(value);
}

/** An unmodifiable map of properties of the experiment */
public Map<String, String> getProperties() {
    return properties;
}

/**
 * Get a float property with the key.
 *
 * @throws NumberFormatException if the value is not in the right format
 * @throws NullPointerException if no such property key is defined
 */
public float getFloat(String key) {
    return Float.parseFloat(getString(key));
}

/**
 * Get a float property with the key.
 *
 * @throws NumberFormatException if the value is not in the right format
 */
public float getFloat(String key, float defaultValue) {
    String value = getString(key);
    return value == null ? defaultValue : Float.parseFloat(value);
}

/**
 * Get a int property with the key.
 *
 * @throws NumberFormatException if the value is not in the right format
 */
public int getInt(String key, int defaultValue) {
    String value = getString(key);
    return value == null ? defaultValue : Integer.parseInt(value);
}

/**
 * If there was a validation problem (mass conservation), return a string
 * describing the validation problem, else return null.
 *
 * @return String describing validation problem or null
 */
public String getValidationProblem() {
    return validationProblem;
}

/** Write the experiment compact and completely to a string. */
public String toString() {
    StringBuilder buffer = new StringBuilder();

    buffer
        .append("Experiment{").append(getDescription())
        .append(": initialTemperature=").append(getInitialTemperature())
        .append("\nParticleClasses={");
    ;

    for (ParticleClass pc : getParticleClasses().values()) buffer.append(pc);
    buffer.append("\n}\nReactions={");
    for (Reaction r : getReactions()) buffer.append(r);
    buffer.append("\n}\nProperties={\n");
    for (Object key : properties.keySet())
        buffer.append(key).append('=')
            .append(properties.get(key)).append('\n');
    buffer.append("}}\n");

    return buffer.toString();
}

/** Read the experiment's particle classes. Internal for the constructor. */
private Map<String, ParticleClass> readParticleClasses(EasyXML xml) {
    Map<String, ParticleClass> pcs = new LinkedHashMap<String, ParticleClass>();
    int classCount = xml.getCount("ParticleClass");
    if (classCount == 0) return pcs;

    xml.enter("ParticleClass");
    for (int i = 0; i < classCount; i++) {
        ParticleClass pc = new ParticleClass(i, xml);
        pcs.put(pc.getName(), pc);
        xml.next();
    }
    xml.exit();
    return pcs;
}

/** Read the experiment's reaction. Internal for the constructor. */
private List<Reaction> readReactions(EasyXML xml) {
    List<Reaction> reactions = new ArrayList<Reaction>();
    int reactionCount = xml.getCount("Reaction");
    if (reactionCount == 0) return reactions;

    xml.enter("Reaction");
    for (int i = 0; i < reactionCount; i++) {
        reactions.add(new Reaction(xml, i, this));
        xml.next();
    }
    xml.exit();
    return reactions;
}

/** Read the experiment's properties. Internal for the constructor. */
private Map<String, String> readProperties(EasyXML xml) {
    Map<String, String> properties = new HashMap<String, String>();
    int propertyCount = xml.getCount("Property");
    if (propertyCount == 0) return properties;

    xml.enter("Property");
    for (int i = 0; i < propertyCount; i++) {
        properties.put(
            xml.getString("Key"), xml.getString("Value"));
        xml.next();
    }
    xml.exit();
    return properties;
}

/**
 * Find a list of possible reactions of type for two particles. The list is
 * cached for efficiency (because findReactions() is quite a hot method, called
 * a few times for every collision). Let us say, we have a collision of
 * particles A and B and want to get a list of merge reactions, call this method
 * with the reaction type Merge and the classes of particle A and B.
 *
 * @param type Reaction type of the reactions to be found.
 * @param pc1 Particle class of one of the particles.
 * @param pc2 Particle class of one of the particles. Null for Decay reactions.
 * @return The list of possible reactions.
 */
public Set<Reaction> findReactions(
    Reaction.ReactionType type, ParticleClass pc1, ParticleClass pc2
) {
    ReactionsCacheKey key = new ReactionsCacheKey(type, pc1, pc2);
    Set<Reaction> result = reactionsCache.get(key);
    if (result != null) return result;

    Set<Reaction> reactions1, reactions2;
    if (type == Reaction.ReactionType.Transform) {
        reactions1 = pc1.getTransformReactions();
        reactions2 = pc2.getTransformReactions();
    }
    else if (type == Reaction.ReactionType.Merge) {
        reactions1 = pc1.getMergeReactions();
        reactions2 = pc2.getMergeReactions();
    }
    else if (type == Reaction.ReactionType.Decay) {
        reactions1 = pc1.getDecayReactions();
        reactions2 = reactions1;
    }
    else throw new ExperimentInvalidException("unknown ReactionType " + type);

    if (pc1 == pc2) {
        result = new HashSet<Reaction>();
        for (Reaction reaction : reactions1)
            if (reaction.getEduct1() == pc1 && reaction.getEduct2() == pc2)
                result.add(reaction);
    }
    else {
        result = new HashSet<Reaction>(reactions1);
        result.retainAll(reactions2);
    }
    result = Collections.unmodifiableSet(result);

    if (Util.ASSERTS_ENABLED) {
        for (Reaction reaction : result) {
            assert reaction.getReactionType() == type;

            ParticleClass eductClass1 = reaction.getEduct1();
            ParticleClass eductClass2 = reaction.getEduct2();

            assert (pc1 == eductClass1 && pc2 == eductClass2)
                || (pc2 == eductClass1 && pc1 == eductClass2)
                : "reaction " + reaction + " for " + pc1 + " and " + pc2;
        }
    }

    Set<Reaction> old = reactionsCache.put(key, result);
    assert old == null;

    return result;
}

/**
 * A cache to find reactions efficient with lazy lookup. Needed by {@linkplain
 * Experiment#findReactions(Reaction.ReactionType, ParticleClass,
 * ParticleClass)}.
 */
private final Map<ReactionsCacheKey, Set<Reaction>>
    reactionsCache = new HashMap<ReactionsCacheKey, Set<Reaction>>();

/**
 * Keys for the reactions cache. The reactions cache caches all possible
 * reactions for a given triple of reaction type and two particle classes. For
 * example, a collision A + B has the following reactions A + B -> A + A, A + B
 * -> C + C, etc. They are found by the method {@linkplain
 * Experiment#findReactions(Reaction.ReactionType, ParticleClass,
 * ParticleClass)}.
 */
private static final class ReactionsCacheKey {

    final Reaction.ReactionType type;
    final ParticleClass pc1;
    final ParticleClass pc2;
    ReactionsCacheKey(
        Reaction.ReactionType type, ParticleClass pc1, ParticleClass pc2
    ) {
        this.type = type;
        this.pc1 = pc1;
        this.pc2 = pc2;
    }
    public int hashCode() {
        return (type.ordinal() & 0xf)
            | (pc1.getIndex() << 8) & 0xfff00
            | (pc2.getIndex() << 20) & 0xfff00000;
    }
    public boolean equals(Object obj) {
        if (!(obj instanceof ReactionsCacheKey)) return false;
        ReactionsCacheKey key = (ReactionsCacheKey)obj;
        return key.type == type && key.pc1 == pc1 && key.pc2 == pc2;
    }
    public String toString() {
        return type + "(" + pc1.getName() + "+" + pc2.getName() + ")@"
            + Integer.toHexString(hashCode());
    }
}
}