/*
 *                        s a c e s
 *      Simple Artificial Chemistry Experiment System
 *
 * A semester project Summer 2005 and a diploma project 2006
 *           Berne University of Applied Sciences
 *   HTI School of Engineering and Information Technology
 *             Anthony Aguillon, Daniel Noelpp
 *
 */
package saces.exp;


import saces.file.EasyXML;

import java.awt.Color;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;


/**
 * A particle class defines a particle "species". Particles of the same class
 * all have identical mass, for example. Mass and other parameters of particles
 * are stored here.
 * <p/>
 * The term "class" can be misunderstood by Java developers. A class is
 * something like a blueprint that defines common characteristics of objects.
 * This is also true for particles. A particle class is not a Java class,
 * however. It is a definition of common parameters for particles.
 */
public final class ParticleClass {

public static final String NAME_REGEX = "[\\p{L}\\p{N}_ ]+";

private final int index;
private final String name;
private final float radius;
private final float mass;
private final float energy;
private final int initialCount;
private final Color color;

private Set<Reaction> transformReactions = new HashSet<Reaction>();
private Set<Reaction> mergeReactions = new HashSet<Reaction>();
private Set<Reaction> decayReactions = new HashSet<Reaction>();

private static void eie(String msg) {
    throw new ExperimentInvalidException(msg);
}

/**
 * Create a particle class with parameters.
 *
 * @param index The particle class index in the experiment. Must be unique.
 * @param name The particle class name. Must be unique.
 * @param radius The radius of the particle sphere. Must be positive.
 * @param mass The mass of the particle. Must be positive.
 * @param energy The bound energy of the particle. Must be positive.
 * @param initialCount The initial count to start a simulation with.
 * @param color The sphere color. Note that lighting might change the shade a
 * bit.
 */
public ParticleClass(
    int index, String name, float radius, float mass, float energy,
    int initialCount, Color color
) {
    if (index < 0) eie("index=" + index);
    if (radius <= 0f) eie("radius=" + radius);
    if (mass <= 0f) eie("mass=" + mass);
    if (energy <= 0f) eie("energy=" + energy);
    if (initialCount < 0) eie("initialCount=" + initialCount);

    this.index = index;
    this.name = name;
    this.radius = radius;
    this.mass = mass;
    this.energy = energy;
    this.initialCount = initialCount;
    this.color = color;

    if (!name.matches('^' + NAME_REGEX + '$')) eie(
        "Name must only contain letters, digits, space and underscore");

    // don't seal yet because the reactions must be loaded first.
}

/**
 * Create a particle class from an XML file. Works only if the EasyXML instance
 * is currently inside an ParticleClass element (or there will be an
 * ExperimentInvalidException).
 *
 * @param xml An EasyXML instance currently inside an ParticleClass element.
 */
public ParticleClass(int number, EasyXML xml) {
    this(number, xml.getString("Name"), xml.getFloat("Radius"),
        xml.getFloat("Mass"), xml.getFloat("Energy"), xml.getInt(
        "InitialCount"), xml.getColor("Color"));
}

/**
 * Get the initial number of particles of this class a simulation is to be
 * started with.
 */
public int getInitialCount() { return initialCount; }

/** Get the particle class' index. */
public int getIndex() { return index; }

/** Get the particle class' name. */
public String getName() { return name; }

/** Get the particle color of this particle class. */
public Color getColor() { return color; }

/** Get the sphere radius of particles of this particle class. */
public float getRadius() { return radius; }

/** Get the particle mass of particles of this particle class. */
public float getMass() { return mass; }

/** Get the bound energy of particles of this particle class. */
public float getEnergy() { return energy; }


/** Transform reactions of this particle class as educt. */
public Set<Reaction> getTransformReactions() {
    // assert that the reactions are sealed.
    assert transformReactions.getClass() ==
        Collections.unmodifiableSet(new HashSet<Object>()).getClass()
        : transformReactions.getClass();

    return transformReactions;
}

/** Merge reactions of this particle class as educt. */
public Set<Reaction> getMergeReactions() { return mergeReactions; }

/** Decay reactions of this particle class as educt. */
public Set<Reaction> getDecayReactions() { return decayReactions; }

/**
 * During construction of an experiment this is needed to add a reaction to the
 * reactions list of a particle class as an educt. Don't call this method.
 */
void add(Reaction reaction) {
    // call only allowed by Reaction instances
    assert new Throwable().getStackTrace()[1].getClassName().equals(
        Reaction.class.getName());

    switch (reaction.getReactionType()) {
    case Transform:
        transformReactions.add(reaction);
        break;
    case Merge:
        mergeReactions.add(reaction);
        break;
    case Decay:
        decayReactions.add(reaction);
        break;
    default:
        assert false : reaction.getReactionType() + " of reaction " + reaction;
    }
}

void clearReactions() {
    transformReactions = new HashSet<Reaction>();
    mergeReactions = new HashSet<Reaction>();
    decayReactions = new HashSet<Reaction>();
}

/**
 * Only possible if not yet sealed. Needed by ReactionTableModel if a reaction
 * has been deleted from the Reactions tab.
 */
public void clearReaction(Reaction reaction) {
    transformReactions.remove(reaction);
    mergeReactions.remove(reaction);
    decayReactions.remove(reaction);
}

/**
 * To make the reactions list unmodifiable. Can be called repeatedly, only the
 * first call has an effect.
 */
void seal() {
    if (transformReactions instanceof HashSet)
        transformReactions = Collections.unmodifiableSet(transformReactions);
    if (mergeReactions instanceof HashSet)
        mergeReactions = Collections.unmodifiableSet(mergeReactions);
    if (decayReactions instanceof HashSet)
        decayReactions = Collections.unmodifiableSet(decayReactions);
}

/** Print the particle class in compact but readable form to a string. */
public String toString() {
    return new StringBuilder(getName()).append('{')
        .append(" initialCount=").append(getInitialCount())
        .append(" energy=").append(getEnergy())
        .append(" radius=").append(getRadius())
        .append(" mass=").append(getMass())
        .append(" color=").append(
        String.format("#%06x", getColor().getRGB() & 0xffffff))
        .append("}").toString();
}

/** Equality is based on index. */
public boolean equals(Object obj) {
    return ((ParticleClass)obj).index == index;
}

/** Use the index as hash code. */
public int hashCode() {
    return index;
}
}

