/*
 *                        s a c e s
 *      Simple Artificial Chemistry Experiment System
 *
 * A semester project Summer 2005 and a diploma project 2006
 *           Berne University of Applied Sciences
 *   HTI School of Engineering and Information Technology
 *             Anthony Aguillon, Daniel Noelpp
 *
 */
package saces.exp;

import saces.Util;
import static saces.exp.Reaction.ReactionType.*;
import saces.file.EasyXML;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


/**
 * A reaction instance defines a reaction in the simulation. It contains
 * reaction parameters: the reagents, activation energy, reaction probability
 * etc.
 */
public final class Reaction {


/** Experiment property key whether mass preservation is validated. */
public static final String
    VALIDATE_MASS_PRESERVATION = "ValidateMassConservation";

/** Experiment property key whether activation energy is validated. */
public static final String
    VALIDATE_ACTIVATION_ENERGY = "ValidateActivationEnergy";

/** What types of reaction are available in 'saces'? */
public enum ReactionType {

    /** Decay Reaction type with Educt -> Product1 + Product2 */
    Decay(1, 2),

    /** "Normal" Reaction type with Educt1 + Educt2 -> Product1 + Product2 */
    Transform(2, 2),

    /** Merge Reaction type with Educt1 + Educt2 -> Product */
    Merge(2, 1);

    private final int eductCount;
    private final int productCount;

    private ReactionType(int eductCount, int productCount) {
        this.eductCount = eductCount;
        this.productCount = productCount;
    }

    /** Number of educts of this reaction type. Either 1 or 2. */
    public int getEductCount() { return eductCount; }

    /** Number of products of this reaction type. Either 1 or 2. */
    public int getProductCount() { return productCount; }

    /** Number of reagents of this reaction type. Either 3 or 4. */
    public int getReagentCount() { return eductCount + productCount; }
}

private final int index;
private final float activationEnergy;
private final float probability;
private final String description;
private final ReactionType reactionType;
private final ParticleClass[] reagents;
private final String validationProblem;

/**
 * Create a Reaction instance from parameters.
 *
 * @param index Reaction index, must be unique.
 * @param description A short description of the reaction.
 * @param probability Reaction probability. Must be between 0 and 1 inclusive.
 * @param activationEnergy 0 or a positive value (0 means: not defined).
 * @param reactionType The type of the reaction.
 * @param reagents An array of reactions (with the right count of them).
 * @param experiment The experiment the reaction is part of.
 * @throws ExperimentInvalidException if some combination of values is not
 * allowed or applicable. This exception must be reported to the end user (only
 * its message text).
 */
public Reaction(
    int index, String description, float probability, float activationEnergy,
    ReactionType reactionType, ParticleClass[] reagents, Experiment experiment
) {
    this.index = index;
    this.description = description;
    this.probability = probability;
    this.activationEnergy = activationEnergy;
    this.reactionType = reactionType;
    this.reagents = reagents;

    validationProblem = checkClassInvariants(experiment);
    registerReaction();
}

/**
 * Read a Reaction instance from XML.
 *
 * @param xml XML wrapper to read from
 * @param experiment Needed for experiment properties and the particle classes
 * @throws ExperimentInvalidException if some combination of values is not
 * allowed or applicable. This exception must be reported to the end user (only
 * its message text).
 */
public Reaction(EasyXML xml, int index, Experiment experiment) {
    this.index = index;
    this.description = xml.getString("Description");
    this.probability = xml.getFloat("Probability", 1f);
    this.activationEnergy = xml.getFloat("ActivationEnergy", 0);
    this.reactionType = ReactionType.valueOf(xml.getString("ReactionType"));

    List<ParticleClass> reagents = new ArrayList<ParticleClass>();
    int reagentCount = xml.getCount("Reagent");
    if (reagentCount > 0) {
        xml.enter("Reagent");
        for (int i = 0; i < reagentCount; i++) {
            String name = xml.getString("Name");
            ParticleClass reagent = experiment.getParticleClass(name);
            if (reagent == null) throw new ExperimentInvalidException(
                "no such reagent: " + name);
            reagents.add(reagent);
            xml.next();
        }
        xml.exit();
    }
    this.reagents = reagents.toArray(new ParticleClass[reagentCount]);

    validationProblem = checkClassInvariants(experiment);
    registerReaction();
}

// We really need a modifiable version of the experiment. The problem
// is the verification whether the particle classes are really defined.
// If the user changed the particle classes then edited the reaction,
// one cannot use the particle classes from the experiment instance
// but needs the particle class list from the gui. This leads so some
// kludges. It would better if we had the time to define a modifiable
// version.
/**
 * Create a reaction from an equation string. Used by the Settings frame.
 *
 * @param equation in the form A + B -> C + D (= or == is also allowed).
 */
public Reaction(
    int index, String description, float probability, float activationEnergy,
    String equation, List<ParticleClass> availableReagents
) {
    this.index = index;
    this.description = description;
    this.probability = probability;
    this.activationEnergy = activationEnergy;

    reactionType = findReactionType(equation);
    reagents = findReagents(equation, reactionType, availableReagents);

    // Kludgey
    checkClassInvariants1();
    validationProblem = null;
}

/** The reaction's index in the experiment's reaction list. */
public int getIndex() { return index; }

/** The reaction description (usually a sentence or so). */
public String getDescription() { return description; }

/**
 * The energy required for the reaction to take effect. If zero, then activation
 * energy is not defined for this reaction.
 */
public float getActivationEnergy() { return activationEnergy; }

/** The probability for the reaction to take effect. */
public float getProbability() { return probability; }

/** The reaction type. */
public ReactionType getReactionType() { return reactionType; }

/** One of the reaction educts. */
public ParticleClass getEduct1() { return reagents[0]; }

/** The other educt if there is one (for Decay reactions it's null). */
public ParticleClass getEduct2() {
    return reactionType == Decay ? null : reagents[1];
}

/** One of the reaction products. */
public ParticleClass getProduct1() {
    return reagents[reactionType == Decay ? 1 : 2];
}

/** The other product if there is one (for Merge reactions it's null). */
public ParticleClass getProduct2() {
    return reactionType == Merge ? null
        : reagents[reactionType == Decay ? 2 : 3];
}

/** One of the reagents (educt or product) of the reaction. */
public ParticleClass getReagent(int index) { return reagents[index]; }

/** Unmodifiable list of reagents (educts and products) of the reaction. */
public List<ParticleClass> getReagents() {
    return Collections.unmodifiableList(Arrays.asList(reagents));
}

/** Get a validation problem message (null if there is no problem) */
public String getValidationProblem() { return validationProblem; }

/**
 * The mass change in a reaction. This should be zero because of mass
 * conservation. An increase in mass would be positive.
 */
public float getMassChange() {
    ParticleClass reag0 = getReagent(0);
    ParticleClass reag1 = getReagent(1);
    float deltaMass = -reag0.getMass() + getReagent(2).getMass();

    switch (reactionType) {
    case Decay:
        return deltaMass + reag1.getMass();
    case Merge:
        return deltaMass - reag1.getMass();
    case Transform:
        return deltaMass - reag1.getMass() + getReagent(3).getMass();
    default:
        assert false : reactionType;
    }
    return Float.NaN;
}

/**
 * The reaction bound energy difference. Reactions which free energy have
 * positive energy difference (energy comes from the reaction).
 */
public float getEnergyDifference() {
    ParticleClass reag0 = getReagent(0);
    ParticleClass reag1 = getReagent(1);
    float delta = reag0.getEnergy() - getReagent(2).getEnergy();

    switch (reactionType) {
    case Decay:
        return delta - reag1.getEnergy();
    case Merge:
        return delta + reag1.getEnergy();
    case Transform:
        return delta + reag1.getEnergy() - getReagent(3).getEnergy();
    default:
        assert false : reactionType;
    }
    return Float.NaN;
}

/** Get equation as string */
public String getEquation() {
    StringBuilder builder = new StringBuilder();
    builder.append(getEduct1().getName());
    if (getEduct2() != null)
        builder.append(" + ").append(getEduct2().getName());
    builder.append(" --> ").append(getProduct1().getName());
    if (getProduct2() != null)
        builder.append(" + ").append(getProduct2().getName());
    return builder.toString();
}

static private final String nameRegex = "(" + ParticleClass.NAME_REGEX + ")";
static private final String equalRegex = "(-+>|=+)";
// <Name> + <Name> -> <Name> + <Name>
static private final String transformRegex = String.format(
    "^%1$s\\+%1$s%2$s%1$s\\+%1$s$", nameRegex, equalRegex);
// <Name> + <Name> -> <Name>
static private final String mergeRegex = String.format(
    "^%1$s\\+%1$s%2$s%1$s$", nameRegex, equalRegex);
// <Name> -> <Name> + <Name>
static private final String decayRegex = String.format(
    "^%1$s%2$s%1$s\\+%1$s$", nameRegex, equalRegex);
static private final Pattern transformPattern = Pattern.compile(transformRegex);
static private final Pattern mergePattern = Pattern.compile(mergeRegex);
static private final Pattern decayPattern = Pattern.compile(decayRegex);


/** Parse the equation formula */
public ParticleClass[] findReagents(
    String equation, ReactionType type, List<ParticleClass> particleClasses
) {
    Matcher matcher;
    String[] reagentNames = null;

    switch (type) {
    case Transform:
        matcher = transformPattern.matcher(equation);
        if (matcher.matches())
            reagentNames = new String[] {
                matcher.group(1), matcher.group(2),
                matcher.group(4), matcher.group(5),
            };
        break;

    case Merge:
        matcher = mergePattern.matcher(equation);
        if (matcher.matches())
            reagentNames = new String[] {
                matcher.group(1), matcher.group(2), matcher.group(4),
            };
        break;

    case Decay:
        matcher = decayPattern.matcher(equation);
        if (matcher.matches())
            reagentNames = new String[] {
                matcher.group(1), matcher.group(3), matcher.group(4),
            };
        break;

    default:
        assert false;
        return null;
    }

    assert reagentNames != null : "didn't match but should";

    ParticleClass[] reagents = new ParticleClass[reagentNames.length];
    int index = 0;
    for (String name : reagentNames) {
        // Kludgey here (but in the last weeks there was not enough time to
        // create a modifiable model of the experiment just for editing)
        // Better would be to use a map and not a list.
        //reagents[index] = experiment.getParticleClass(name);
        name = name.trim();
        for (ParticleClass pc : particleClasses)
            if (pc.getName().equals(name)) {
                reagents[index] = pc;
                break;
            }

        if (reagents[index] == null) throw new ExperimentInvalidException(
            "no such reagent: " + name);
        index++;
    }
    return reagents;
}

/**
 * Parse the equation string and find its reaction type or throw an
 * IllegalArgumentException.
 */
public ReactionType findReactionType(String equation) {
    if (transformPattern.matcher(equation).matches()) return Transform;
    if (mergePattern.matcher(equation).matches()) return Merge;
    if (decayPattern.matcher(equation).matches()) return Decay;

    throw new IllegalArgumentException(equation + " not a reaction equation");
}

/** Print a compact, but complete reaction description to a string */
public String toString() {
    StringBuilder buffer = new StringBuilder();

    buffer.append(getReactionType())
        .append('{').append(getDescription())
        .append(' ').append(getEquation())
        .append(" probability=").append(getProbability())
        .append("; activationEnergy=").append(getActivationEnergy())
        ;

    return buffer.append('}').toString();
}

/** Compare based on the reaction index. */
public boolean equals(Object obj) {
    return ((Reaction)obj).index == index;
}

/** Use the reaction index as hash code. */
public int hashCode() {
    return index;
}

private static void iae(String msg) {
    throw new ExperimentInvalidException(msg);
}

// A simpler version of class invariants.
private void checkClassInvariants1() {
    if (probability < 0 || probability > 1) iae("probability=" + probability);
    if (activationEnergy < 0) iae("activationEnergy=" + activationEnergy);
    int count = reactionType.getReagentCount();
    if (reagents.length != count) iae("reagents.length=" + reagents.length);

    for (int i = 0; i < reagents.length; i++)
        if (reagents[i] == null) throw new NullPointerException(
            "reagents[" + i + "]");
}

/**
 * Check Reaction class invariants. These are: <ul> <li>probability between 0
 * and 1 inclusive <li>non-negative activation energy <li>reagent count as
 * expected (4 for Transform reaction, for example) <li>mass preservation in
 * reaction (optional) <li>activation energy not smaller than negative energyDiff
 * (optional)</ul>
 *
 * @param experiment
 * @throws ExperimentInvalidException invariant violations. They never should
 * happen. No need to catch them, because they indicate internal problems.
 */
private String checkClassInvariants(Experiment experiment) {
    checkClassInvariants1();

    String validationProblem = null;

    // Mass preservation is an important property of physics. Reactions
    // should not violate mass preservation. But because entering complex
    // reaction systems with mass preservation is cumbersome, validation
    // of mass preservation is an optional step.
    if (experiment.getBoolean(VALIDATE_MASS_PRESERVATION, false))
        if (Math.abs(getMassChange()) > Util.EPSILON)
            validationProblem = "Reaction " + getDescription()
                + " violates mass preservation. Mass change is "
                + getMassChange()
                + " (positive means increasing mass)";

    // Reactions with negative enthalpies are reaction which "convert"
    // kinetic energy to bound energy. Suppose two fast particle collide
    // they can merge into a particle with high internal energy.
    // Such reactions aren't possible if there is not enough kinetic
    // energy. Activation energy is the barrier for such reaction. It
    // shouldn't be too low, especially not lower than the negative
    // energyDiff.
    // But because some artificial chemistries treat the activation
    // energy differently (especially Lotka-Volterra-systems) validation
    // of activation energy is an optional step.
    if (experiment.getBoolean(VALIDATE_ACTIVATION_ENERGY, false))
        if (-getEnergyDifference() > getActivationEnergy())
            validationProblem = "Reaction " + getDescription()
                + " is a reaction with negative energy difference, but has"
                + "insufficient activation energy (must be "
                + (-getEnergyDifference()) + " or higher but is "
                + getActivationEnergy() + ").";

    return validationProblem;
}

// register reaction with educts only (for a fast lookup of reactions)
void registerReaction() {
    try {
        getEduct1().add(this);
        if (getEduct2() != null) getEduct2().add(this);
    }
    catch (UnsupportedOperationException uoe) {
        throw new ExperimentInvalidException("Experiment already sealed: "
            + getEduct1().getName()
            + (getEduct2() == null ? "" : ", " + getEduct2().getName()));
    }
}
}