package saces.file;

import saces.Util;
import static saces.Util.*;
import saces.sim.Particle;

import java.io.BufferedOutputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;


/**
 * The binary log is a file to which a running simulation writes its data,
 * measurements etc. in a binary format. Java has a well-defined internal
 * representation of binary data. In dropping the conversion of data to text,
 * performace is gained for the simulation animation.
 * <p/>
 * The binary log uses a record-oriented format with a timestamp for each
 * record. A stream of different records, whose ordering is free, is produced.
 * The first record, however, should be a magic record. All record start with an
 * eight-byte magic number, whose last nybble is the record identifier. The
 * timestamp is an eight-byte number of milliseconds since midnight, January 1st
 * 1970 as produced by the Java method {@linkplain System#currentTimeMillis()}.
 * <p/>
 * It is assumed that the data viewer or tail dumper has access to the
 * experiment definition. Without the experiment definition the binary data is
 * uninterpretable, because only indices of particle classes or of reactions are
 * written.
 */
public class BinaryLog {

private static class NullOutputStream extends OutputStream {

    public void write(int b) throws IOException {}
}

/** The magic number of the 'saces' binary log */
public static final long MAGIC = 0x73416345732100f0L;
/** The record identifier nybble for the magic record. */
public static final int RECORD_MAGIC = 0;
/** The record identifier nybble for the reaction record (UNSUPPORTED). */
public static final int RECORD_REACTION = 1;
/** The record identifier nybble for the particle classes record. */
public static final int RECORD_PARTICLECLASSES = 2;
/** The record identifier nybble for the snapshot record. */
public static final int RECORD_SNAPSHOT = 3;
/** The record identifier nybble for the histogram record. */
public static final int RECORD_HISTOGRAM = 4;
/** The record identifier nybble for the float property record. */
public static final int RECORD_FLOATPROPERTY = 13;
/** The record identifier nybble for the int property record. */
public static final int RECORD_INTPROPERTY = 14;
/** The record identifier nybble for the string property record. */
public static final int RECORD_STRINGPROPERTY = 15;

private final DataOutputStream dos;

/** A dummy binary log which discards any data written to it. */
public static final BinaryLog NULL = new BinaryLog();

private BinaryLog() {
    dos = new DataOutputStream(new NullOutputStream());
}

/** Create a binary log (truncating an existing file). */
private BinaryLog(String fileName) {
    try {
        dos = new DataOutputStream(new BufferedOutputStream(
            new FileOutputStream(fileName)));

        // Important if writing to a file and a BinaryLogRetriever is reading
        // from that file: Ensure that BinaryLogRetriever doesn't miss the file
        // truncate. In most rare cases, if the thread is interrupted, the
        // sleep could be too short and the BinaryLogRetriever misses the
        // truncate. Bad luck, so what.
        try { Thread.sleep(TailFileInputStream.POLL_SLEEP_TIME + 10); }
        catch (InterruptedException ie) { }

        dos.writeLong(MAGIC);
        dos.writeLong(System.currentTimeMillis());
    }
    catch (IOException ioe) {
        fileName = new File(fileName).getAbsolutePath();
        Util.showErrorMessage("Binary log file not created",
            "Binary log file " + fileName
                + "\ncouldn't be created:\n\n" + ioe
                + "\n\nPerhaps running from CD?"
                + "\nData viewing will not be possible.");
        throw new IllegalArgumentException(ioe);
    }
}

public static BinaryLog create(String fileName) {
    try {
        return new BinaryLog(fileName);
    }
    catch (IllegalArgumentException iae) {
        return BinaryLog.NULL;
    }
}

/** Close the binary log. Do not forget to close or some data might be dropped. */
public void close() {
    try { dos.close(); }
    catch (IOException ioe) {
        System.err.println("ObjectOutputStream.close(): " + ioe);
    }
}

/** Immediately flush buffered data. */
public void flush() {
    try { dos.flush(); }
    catch (IOException ioe) {
        System.err.println("ObjectOutputStream.flush(): " + ioe);
    }
}

// Pad length bytes.
private void pad(int length) throws IOException {
    for (int i = 0; i < length; i++) dos.writeByte(0);
}

// We have a length. Pad so that the length gets a multiple of 8 bytes.
private void padWith(int length) throws IOException {
    final int[] padTable = {0, 7, 6, 5, 4, 3, 2, 1};
    pad(padTable[length % 8]);
}

// Write in UTF format, but get the number of bytes written.
private int writeUTF(String s) throws IOException {
    byte[] bs = s.getBytes("UTF-8");
    int length = bs.length;
    if (length > Short.MAX_VALUE) throw new IllegalArgumentException(
        "too long a string, length: " + length);

    dos.writeShort(length);
    dos.write(bs);

    return length + 2;
}

//// Format:
//// long recordType, long timeStamp, int reactionNumber,
//public synchronized void writeReactionRecord(int reactionNumber) {
//    try {
//        dos.writeLong(RECORD_REACTION | MAGIC);
//        dos.writeLong(System.currentTimeMillis());
//        dos.writeInt(reactionNumber);
//    }
//    catch (IOException ioe) { handleIOException(ioe); }
//}

// Format:
// long recordType, long timeStamp, n * int particleCount
// n is the number of particle classes as specified in the experiment
// padded up, so that the record length is a multiple of 8.
/** Write a record to contain the number of particles per particle class. */
public synchronized void writeParticleClassesRecord(int[] particleCounts) {
    try {
        dos.writeLong(RECORD_PARTICLECLASSES | MAGIC);
        dos.writeLong(System.currentTimeMillis());
        for (int count : particleCounts) dos.writeInt(count);

        // pad up with 4 bytes if the number of particle classes is odd
        if (1 == particleCounts.length % 2) pad(4);
    }
    catch (IOException ioe) { handleIOException(ioe); }
}

// Format:
// long recordType, long timeStamp
// int m = particleCount, m * {
//     int particleClassNumber, 3 * float position, 3 * float velocity
// }
// padded up, so that the record length is a multiple of 8.
/** Write a snapshot containing position, velocity and class of all particles. */
public synchronized void writeSnapshot(Particle[] particles) {
    if (particles.length == 0) return;

    try {
        dos.writeLong(RECORD_SNAPSHOT | MAGIC);
        dos.writeLong(System.currentTimeMillis());
        dos.writeInt(particles.length);
        for (Particle p : particles) {
            dos.writeInt(p.particleClass.getIndex());
            dos.writeFloat(p.position[X]);
            dos.writeFloat(p.position[Y]);
            dos.writeFloat(p.position[Z]);
            dos.writeFloat(p.velocity[X]);
            dos.writeFloat(p.velocity[Y]);
            dos.writeFloat(p.velocity[Z]);
        }

        // pad up with 4 bytes if the numer of particles is even
        if (0 == particles.length % 2) pad(4);
    }
    catch (IOException ioe) { handleIOException(ioe); }
}

// Format:
// long recordType, long timeStamp
// javaUTF key, javaUTF value
// padded up, so that the record length is a multiple of 8.
/** Write a property: a name-value pair. */
public synchronized void writeProperty(String key, Object value) {
    try {
        int length;
        if (value instanceof Float) {
            dos.writeLong(RECORD_FLOATPROPERTY | MAGIC);
            dos.writeLong(System.currentTimeMillis());
            length = writeUTF(key) + 4;
            dos.writeFloat((Float)value);
        }
        else if (value instanceof Integer) {
            dos.writeLong(RECORD_INTPROPERTY | MAGIC);
            dos.writeLong(System.currentTimeMillis());
            length = writeUTF(key) + 4;
            dos.writeInt((Integer)value);
        }
        else {
            dos.writeLong(RECORD_STRINGPROPERTY | MAGIC);
            dos.writeLong(System.currentTimeMillis());
            length = writeUTF(key) + writeUTF(value.toString());
        }
        padWith(length);
    }
    catch (IOException ioe) { handleIOException(ioe); }
}

// Format:
// long recordType, long timeStamp
// float minSpeed, float avgSpeed, float maxSpeed
// float histogramTop, int k = count, k * int histogram
/** Write a speed distribution (histogram). */
public synchronized void writeHistogram(
    float min, float avg, float max, float top, int[] histogram
) {
    if (histogram.length == 0) return;

    try {
        dos.writeLong(RECORD_HISTOGRAM | MAGIC);
        dos.writeLong(System.currentTimeMillis());
        dos.writeFloat(min);
        dos.writeFloat(avg);
        dos.writeFloat(max);
        dos.writeFloat(top);
        dos.writeInt(histogram.length);
        for (int i : histogram) dos.writeInt(i);

        // pad up with 4 bytes if the number of histogram values is even
        if (0 == histogram.length % 2) pad(4);
    }
    catch (IOException ioe) { handleIOException(ioe); }
}

// An IO exception occurred. Close binary log and show error message.
private void handleIOException(IOException ioe) {
    close();
    Util.showErrorMessage("Error writing to binary log file",
        "Error writing to binary log file:\n" + ioe);
}

}
