package saces.file;

import static saces.file.BinaryLog.*;
import saces.exp.ParticleClass;
import saces.exp.Experiment;

import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;


// Note: some code duplication with ViewDataModel
/** A helper class to convert binary log in a text form better suitable
 * for processing by third-party software.
 */
public class BinaryLogRetriever {

public static final long MASK = 0xfL;

private final DataInputStream dis;
private final PrintStream out;
private final Experiment experiment;
private boolean closed = true;

/** Create a binary log retriever instance from file to write to the
 * print stream.
 * @param filename The binary log file name.
 * @param out The print stream to write to.
 * @param experiment The experiment about which the binary log is.
 */
public BinaryLogRetriever(
    String filename, PrintStream out, Experiment experiment
) {
    this.out = out == null ? System.out : out;
    DataInputStream dis = null;
    try {
        new File(filename).createNewFile();
        dis = new DataInputStream(
            new TailFileInputStream(new FileInputStream(filename)));
        closed = false;
    }
    catch (IOException ioe) {
        ioException(ioe);
    }

    this.dis = dis;
    this.experiment = experiment;
}

/** An additional entry point to retrieve the binary log. */
public static void main(String[] args) {
    String logFileName = "saces.binlog";
    String experimentFileName = "SampleExperiments/DefaultExperiment.xml";

    if (args.length > 0) experimentFileName = args[0];
    if (args.length > 1) logFileName = args[1];

    tail(logFileName, experimentFileName);
}

/** Tail dump the binary log of the experiment.
 *
 * @param experimentFileName The file name of the experiment XML file.
 * @param logFileName The file name of the binary log.
 */
public static void tail(String experimentFileName, String logFileName) {
    Experiment experiment;
    try {
        experiment = new Experiment(new FileReader(experimentFileName));
    }
    catch (FileNotFoundException fnfe) {
        System.err.println(experimentFileName + " not found");
        return;
    }
    catch (BadExperimentFileException befe) {
        System.err.println(experimentFileName + " couldn't be validated: "
            + befe.getMessage());
        return;
    }

    BinaryLogRetriever binaryLog
        = new BinaryLogRetriever(logFileName, System.out, experiment);
    binaryLog.startDecode();
}

// Asynchronously start decoding the binary log
private void startDecode() {
    new Thread(new Runnable() {
        public void run() {
            while (!isClosed()) { readRecord(); }
        }
    }).start();
}

// Has the binary log been closed?
private boolean isClosed() { return closed; }

// An error occurred. Print it and close binary log.
private void error(String msg) {
    out.println("ERROR\t" + msg);
    close();
}

// Close the binary log.
private void close() {
    try { dis.close(); } catch (IOException ioe) { }
    out.println("CLOSE\t");
    closed = true;
}

// An io exception ocurred. Print it and close binary log.
private void ioException(IOException ioe) {
    out.println("EXCEPTION\t" + ioe);
    StackTraceElement[] st = ioe.getStackTrace();
    for (int i = 0; i < st.length; i++) out.println("STACKTRACE\t" + st[i]);
    close();
}

// Read a binary log record.
private void readRecord() {
    if (closed) return;

//    List<Reaction> reactions = experiment.getReactions();
    Map<String, ParticleClass> pcs = experiment.getParticleClasses();
    List<ParticleClass> pcList = new ArrayList<ParticleClass>(pcs.values());

    StringBuilder sb = new StringBuilder();
    try {
        int recordType = getRecordType();
        long timeStamp = dis.readLong();
        String dateTime = String.format("%1$TFT%1$TT.%1$TLUTC", timeStamp);

        switch (recordType) {
        case -1:
            return;

        case RECORD_MAGIC:
            out.println(sb.append("LOG_FILE_START\t").append(dateTime));
            break;

//        case RECORD_REACTION:
//            Reaction reaction = reactions.get(dis.readInt());
//            String description = reaction.getDescription();
//
//            sb.append("REACTION")
//                .append('\t').append(dateTime)
//                .append('\t').append(description);
//
//            out.println(sb);
//            break;

        case RECORD_PARTICLECLASSES:
            sb.append("PARTICLECLASSES").append('\t').append(dateTime);
            out.println(sb);
            sb.delete(0, sb.length());

            for (int i = 0; i < pcs.size(); i++) {
                String name = pcList.get(i).getName();
                int countInClass = dis.readInt();
                sb.append("PARTICLECLASS")
                    .append('\t').append(name)
                    .append('\t').append(countInClass);
                out.println(sb);
                sb.delete(0, sb.length());
            }

            // ignore padding of 4 bytes if number of particle classes is odd
            if (1 == pcs.size() % 2) dis.readInt();

            break;

        case RECORD_SNAPSHOT:
            int particleCount = dis.readInt();
            sb.append("SNAPSHOT")
                .append('\t').append(dateTime)
                .append('\t').append(particleCount);
            out.println(sb);
            sb.delete(0, sb.length());

            for (int i = 0; i < particleCount; i++) {
                String name = pcList.get(dis.readInt()).getName();
                sb.append("PARTICLE")
                    .append("\t").append(name)
                    .append("\t").append(dis.readFloat())
                    .append("\t").append(dis.readFloat())
                    .append("\t").append(dis.readFloat())
                    .append("\t").append(dis.readFloat())
                    .append("\t").append(dis.readFloat())
                    .append("\t").append(dis.readFloat());
                out.println(sb);
                sb.delete(0, sb.length());
            }

            // ignore padding of 4 bytes if number of particles is even
            if (0 == particleCount % 2) dis.readInt();
            break;

        case RECORD_FLOATPROPERTY:
        case RECORD_INTPROPERTY:
        case RECORD_STRINGPROPERTY:
            String key = dis.readUTF();
            Object value
                = recordType == RECORD_FLOATPROPERTY
                ? dis.readFloat()
                : recordType == RECORD_INTPROPERTY
                    ? dis.readInt()
                    : dis.readUTF();

            sb.append("PROPERTY")
                .append('\t').append(dateTime)
                .append('\t').append(key)
                .append('\t').append(value);

            out.println(sb);

            // ignore padding to a length of a multiple of 8 bytes
            int length = getUTFLength(key);
            length += recordType == RECORD_STRINGPROPERTY
                ? getUTFLength(value.toString()) : 4;
            int padLength = getPadLength(length);
            for (int i = 0; i < padLength; i++) dis.readByte();
            break;

        case RECORD_HISTOGRAM:
            float min = dis.readFloat();
            float avg = dis.readFloat();
            float max = dis.readFloat();
            float top = dis.readFloat();
            int size = dis.readInt();

            sb.append("HISTOGRAM")
                .append('\t').append(dateTime)
                .append('\t').append(min)
                .append('\t').append(avg)
                .append('\t').append(max)
                .append('\t').append(top)
                .append('\t').append(size);
            out.println(sb);
            sb.delete(0, sb.length());

            for (int i = 0; i < size; i++) {
                sb.append("INTERVAL")
                    .append('\t').append(dis.readInt());
                out.println(sb);
                sb.delete(0, sb.length());
            }

            // ignore padding of 4 bytes if number of intervals is even
            if (0 == size % 2) dis.readInt();
            break;

        default:
            error("bad record type " + recordType);
            break;

        }
    }
    catch (IOException ioe) { ioException(ioe); }
}

// Decide which record type the current record is.
private int getRecordType() throws IOException {
    long recordType = dis.readLong();
    if ((recordType & ~MASK ^ MAGIC) != 0L) {
        error(String.format("bad record type code 0x%x", recordType));
        return -1;
    }
    return (int)(recordType & MASK);
}

// Get the length of a string as UTF-8 in bytes, including the length tag.
private int getUTFLength(String s) throws IOException {
    return s.getBytes("UTF-8").length + 2;
}

// If we are at length, tell how much bytes to pad.
private int getPadLength(int length) {
    final int[] padTable = {0, 7, 6, 5, 4, 3, 2, 1};
    return padTable[length % 8];
}

}

