/*
 *                        s a c e s
 *      Simple Artificial Chemistry Experiment System
 *
 * A semester project Summer 2005 and a diploma project 2006
 *           Berne University of Applied Sciences
 *   HTI School of Engineering and Information Technology
 *             Anthony Aguillon, Daniel Noelpp
 *
 */
package saces.file;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import javax.xml.XMLConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;
import java.awt.Color;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;


/**
 * EasyXML provides easy and compact reading access to XML files for
 * persistence. The {@code enter()}, {@code next()} and {@code exit()} methods
 * navigate around in the document tree and the {@code get}... methods retrieve
 * data from the attributes of the current element.
 * <p/>
 * XML Schema Validation is done in EasyXML producing {@linkplain
 * BadExperimentFileException} on failure.
 */
public class EasyXML {

/**
 * A special buffered reader as a workaround to a Schema Validating bug. DOM
 * parsing and schema validation must be done as separate steps because of this
 * problem. This means that the reader must be read twice. Experiment XML files
 * shouldn't exceed around 100 kB.
 * <p/>
 * Because Schema Validating closes the source, closing is prevented and made
 * possible by another method.
 */
private static class SpecialReader extends BufferedReader {

    static int BUFFER_SIZE = 100 * 1024 + 1;
    SpecialReader(Reader r) throws IOException {
        super(r, BUFFER_SIZE);
        mark(BUFFER_SIZE);
    }
    public void close() throws IOException {
        // don't close
    }
    void reallyClose() throws IOException {
        super.close();
    }
}

private Element currentElement;
private Document document;

/**
 * Open an XML file using a file name.
 *
 * @throws BadExperimentFileException if there was a problem like validation, IO
 * etc.
 */
public EasyXML(String fileName) throws BadExperimentFileException {
    try {
        init(new FileReader(fileName));
    }
    catch (FileNotFoundException e) {
        throw new IllegalArgumentException(e);
    }
}

/**
 * Open an XML file using a Reader (character input stream).
 *
 * @throws BadExperimentFileException if there was a problem like validation, IO
 * etc.
 */
public EasyXML(Reader reader) throws BadExperimentFileException {
    init(reader);
}

// Factored out as a private method because it is not possible
// to use try-catch with a call to this().
private void init(Reader reader) throws BadExperimentFileException {
    Throwable t = null;
    try {
        // prepare XML Schema validation
        Source schemaSource = new StreamSource(
            getClass().getResource("experiment.xsd").openStream());
        SchemaFactory schemaFactory = SchemaFactory.newInstance(
            XMLConstants.W3C_XML_SCHEMA_NS_URI);
        Schema schema = schemaFactory.newSchema(schemaSource);
        Validator validator = schema.newValidator();

        SpecialReader sr = new SpecialReader(reader);
        validator.validate(new StreamSource(sr));
        sr.reset();

        // Then build DOM
        DocumentBuilderFactory
            factory = DocumentBuilderFactory.newInstance();
        DocumentBuilder builder = factory.newDocumentBuilder();
        document = builder.parse(new InputSource(sr));
        sr.reallyClose();
    }
    catch (SAXException saxe) { t = saxe; }
    catch (ParserConfigurationException pce) { t = pce; }
    catch (IOException ioe) { t = ioe; }

    if (t != null) {
        throw new BadExperimentFileException(t);
    }

    currentElement = document.getDocumentElement();
}

/** Get the count of child elements of that tag name */
public int getCount(String tagName) {
    return currentElement.getElementsByTagName(tagName).getLength();
}

/** Enter the first child element with this tag tagName */
public void enter(String tagName) {
    Node node = currentElement.getElementsByTagName(tagName).item(0);
    if (node instanceof Element) currentElement = (Element)node;
    else throw new IllegalArgumentException("no such element " + tagName);
}

/**
 * Go to the next sibling. If there's none, just return false.
 *
 * @return true if successful.
 */
public boolean next() {
    Node node = currentElement.getNextSibling();
    while (node != null) {
        if (node instanceof Element) {
            currentElement = (Element)node;
            return true;
        }
        node = node.getNextSibling();
    }
    return false;
}

/** Exit the element and continue with the parent */
public void exit() {
    Node node = currentElement.getParentNode();
    if (node instanceof Element) currentElement = (Element)node;
    else throw new IllegalArgumentException("no parent element");
}

/** Get a string as an attribute value */
public String getString(String attribute) {
    return currentElement.getAttribute(attribute);
}

/** Get a floating point number as an attribute value */
public float getFloat(String attribute) {
    String s = getString(attribute);
    try {
        float f = Float.parseFloat(s);
        if (!Float.isNaN(f) && !Float.isInfinite(f)) return f;
    }
    catch (NumberFormatException nfe) {}

    // rethrow
    throw new NumberFormatException("'" + s + "' for attribute " + attribute);
}

/**
 * Get a floating point number as an optional attribute value. If there is no
 * such attribute, return the default value.
 */
public float getFloat(String attribute, float defaultValue) {
    return getString(attribute).equals("")
        ? defaultValue : getFloat(attribute);
}

/** Get an integer number as an attribute value */
public int getInt(String attribute) {
    try {
        return Integer.parseInt(getString(attribute));
    }
    catch (NumberFormatException nfe) {
        throw new NumberFormatException(
            "'" + nfe.getMessage() + "' for attribute " + attribute);
    }
}

/**
 * Get color as an attribute value. Color is in one of three formats #RGB,
 * #RRGGBB (where R,G,B are a hex digit) or (r g b), where the three floating
 * point numbers r, g, b in range [0..1] are separated by spaces and the tripel
 * enclosed in parentheses. Examples: {@code #ade}, {@code #18A070} or {@code (1
 * .5 .1)}.
 *
 * @return a three component float array with red, green and blue components in
 *         range [0..1].
 *
 * @throws NumberFormatException if the attribute is not in such a format
 */
public Color getColor(String attribute) {
    try {
        return getColorRaw(attribute);
    }
    catch (NumberFormatException nfe) {
        throw new NumberFormatException("bad color specification '"
            + getString(attribute) + "' for attribute " + attribute);
    }
}

// Get the color as an attribute value. NumberFormatException is about
// a substring of the color value, not the whole color value string.
// Like this: (0.5 0.5 .) ==> NumberFormatException(".") instead of
// NumberFromatException("(0.5 0.5 .)") confusing the user. That's why
// there is a wrapper around this method.
private Color getColorRaw(String attribute) {
    String value = getString(attribute).trim();
    int len = value.length();
    final int[] oneDigitHex = {
        0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
        0x88, 0x99, 0xaa, 0xbb, 0xcc, 0xdd, 0xee, 0xff,
    };

    if (len > 0 && value.charAt(0) == '#') {
        // parse #RGB
        if (len == 4) return new Color(
            oneDigitHex[Integer.parseInt(value.substring(1, 2), 16)],
            oneDigitHex[Integer.parseInt(value.substring(2, 3), 16)],
            oneDigitHex[Integer.parseInt(value.substring(3, 4), 16)]);

            // parse #RRGGBB
        else if (len == 7) return new Color(
            Integer.parseInt(value.substring(1, 3), 16),
            Integer.parseInt(value.substring(3, 5), 16),
            Integer.parseInt(value.substring(5, 7), 16));
    }

    // parse (r g b)
    else if (len > 1
        && value.charAt(0) == '(' && value.charAt(len - 1) == ')'
        ) {
        String[] rgb = value.substring(1, len - 1).split("[ ]+", 3);
        return new Color(
            Float.parseFloat(rgb[0]),
            Float.parseFloat(rgb[1]),
            Float.parseFloat(rgb[2]));
    }

    throw new NumberFormatException("'" + value + "' not a color format");
}

/** Get a text node */
public String getText() {
    return currentElement.getTextContent();
}

}