package saces.file;

import saces.exp.Experiment;
import saces.exp.ParticleClass;
import saces.exp.Reaction;

import java.awt.Color;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;


/** Write experiments to experiment XML files validatable by XML Schema. */
public class ExperimentSaver {


/**
 * Write experiments to experiment XML files validatable by XML Schema.
 *
 * @param experiment The experiment instance to be written.
 * @param file The destination file of the experiment XML file.
 */
public static void saveAsXML(
    Experiment experiment, File file
) throws IOException {
    BufferedWriter bw = new BufferedWriter(new FileWriter(file));

    bw.write("<?xml version='1.0' encoding='ISO-8859-1'?>\n"
        + "<!--\n"
        + "                          s a c e s\n"
        + "        Simple Artificial Chemistry Experiment System\n"
        + "\n"
        + "   A semester project Summer 2005 and a diploma project 2006\n"
        + "             Berne University of Applied Sciences\n"
        + "     HTI School of Engineering and Information Technology\n"
        + "               Anthony Aguillon, Daniel Noelpp\n"
        + "\n"
        + "-->\n"
        + "<Experiment\n"
        + "  xmlns='http://saces.yce.ch/experiment.xsd'\n");

    bw.write("  Description='");
    bw.write(experiment.getDescription());
    bw.write("'\n");

    bw.write("  InitialTemperature='");
    bw.write("" + experiment.getInitialTemperature());
    bw.write("'\n");

    bw.write("  TimeStep='");
    bw.write("" + experiment.getTimestep());
    bw.write("'\n");

    if (experiment.hasRandomSeed()) {
        bw.write("  RandomSeed='");
        bw.write("" + experiment.getRandomSeed());
        bw.write("'\n");
    }

    bw.write("  Width='");
    bw.write("" + experiment.getWidth());
    bw.write("'\n");

    bw.write("  Height='");
    bw.write("" + experiment.getHeight());
    bw.write("'\n");

    bw.write("  Depth='");
    bw.write("" + experiment.getDepth());
    bw.write("'\n");

    bw.write("  Color='");
    bw.write(toString(experiment.getColor()));
    bw.write("'>\n");

    // don't use the map because it is not sorted (maybe?)
    for (int i = 0; i < experiment.getParticleClassesCount(); i++)
        save(experiment.getParticleClass(i), bw);

    for (Reaction reaction : experiment.getReactions())
        save(reaction, bw);

    for (String key : experiment.getProperties().keySet()) {
        bw.write("  <Property\n");

        bw.write("    Key='");
        bw.write(key);
        bw.write("'\n");

        bw.write("    Value='");
        bw.write(experiment.getString(key));
        bw.write("'/>\n");
    }

    bw.write("</Experiment>\n");
    bw.flush();
    bw.close();
}

// Save a particle class.
private static void save(
    ParticleClass pc, BufferedWriter bw
) throws IOException {
    bw.write("  <ParticleClass\n");

    bw.write("    Name='");
    bw.write(pc.getName());
    bw.write("'\n");

    bw.write("    InitialCount='");
    bw.write("" + pc.getInitialCount());
    bw.write("'\n");

    bw.write("    Radius='");
    bw.write("" + pc.getRadius());
    bw.write("'\n");

    bw.write("    Mass='");
    bw.write("" + pc.getMass());
    bw.write("'\n");

    bw.write("    Energy='");
    bw.write("" + pc.getEnergy());
    bw.write("'\n");

    bw.write("    Color='");
    bw.write(toString(pc.getColor()));
    bw.write("'/>\n");
}

// Save a reaction.
private static void save(
    Reaction reaction, BufferedWriter bw
) throws IOException {
    bw.write("  <Reaction\n");

    bw.write("    Description='");
    bw.write(reaction.getDescription());
    bw.write("'\n");

    bw.write("    Probability='");
    bw.write("" + reaction.getProbability());
    bw.write("'\n");

    bw.write("    ReactionType='");
    bw.write("" + reaction.getReactionType());
    bw.write("'\n");

    if (reaction.getActivationEnergy() > 0) {
        bw.write("    ActivationEnergy='");
        bw.write("" + reaction.getActivationEnergy());
        bw.write("'\n");
    }

    bw.write(">\n");

    for (ParticleClass reagent : reaction.getReagents()) {
        bw.write("    <Reagent Name='");
        bw.write(reagent.getName());
        bw.write("'/>\n");
    }

    bw.write("  </Reaction>\n");
}

// Convert a color to #RRGGBB format.
private static String toString(Color color) {
    return String.format("#%02x%02x%02x",
        color.getRed(), color.getGreen(), color.getBlue());
}

}
