package saces.gl;
/*
 *                       s a c e s
 *      Simple Artificial Chemistry Experiment System
 *
 * A semester project Summer 2005 and a diploma project 2006
 *           Berne University of Applied Sciences
 *   HTI School of Engineering and Information Technology
 *             Anthony Aguillon, Daniel Noelpp
 *
 */

import java.awt.*;


/**
 *
 *  Implements the algorithm described in:
 *  See paper "A Study in Interactive 3-D Rotation Using 2-D Control Devices"
 *  (published in the ACM Siggraph '88 . Membership and purchase required) or
 *  see [http://www.mactech.com/articles/develop/issue_14/Chen_final.html] by
 *  Michael Chen, S. Joy Mountford, and Abigail Sellen
 *  This is also known by as the [cue ball] or [virtual track ball] interface.
 *
 **/
public class BoundingSphere {

private static final int X = 0;
private static final int Y = 1;
private static final int Z = 2;

// Storage for 3D point information.
private final float[] op = new float[3];
private final float[] oq = new float[3];
private final float[] a = new float[3];


/**
*  Calculate a rotation matrix based on the axis and angle of rotation
*  from the last 2 locations of the mouse relative to the Virtual
*  Sphere cue circle.
*
*  @param  pnt1       : The 1st mouse location in the window.
*  @param  pnt2       : The 2nd mouse location in the window.
*  @param  cueCenter  : The center of the virtual sphere in the window.
*  @param  cueRadius  : The radius of the virtual sphere.
*  @param  rotMatrix  : Preallocated rotation matrix to be filled in
*                       by this method.  Must have 16 floating point
*                       elements.
*  @return rotation Matrix
*
**/
public float[] makeRotationMatrix(Point pnt1, Point pnt2, Point cueCenter,
  int cueRadius, float[] rotMatrix) {

  // Project mouse points to 3-D points on the +z hemisphere of a unit sphere.
  pointOnUnitSphere (pnt1, cueCenter, cueRadius, op);
  pointOnUnitSphere (pnt2, cueCenter, cueRadius, oq);

  setRotationMatrix(rotMatrix, op, oq);

  return rotMatrix;
}

/**
*  Project a 2D point on a circle to a 3D point on the +z hemisphere of a
*  unit sphere. If the 2D point is outside the circle, it is first mapped
*  to the nearest point on the circle before projection.
*
*  @param  p         : Window point to be projected onto the sphere.
*  @param  cueCenter : Location of center of virtual sphere in window.
*  @param  cueRadius : The radius of the virtual sphere.
*  @param  v         : Storage for the 3D projected point created by this method.
**/
private static void pointOnUnitSphere(Point p, Point cueCenter, int cueRadius, float[] v) {

  /** Turn the mouse points into vectors relative to the center of the circle
    *  and normalize them.  Flip the y value (Model has +y upward)
    */
  float vx = (p.x - cueCenter.x)/(float)cueRadius;
  float vy = (cueCenter.y - p.y)/(float)cueRadius;
  float lengthSqared = vx*vx + vy*vy;

  float vz = 0;

  if (lengthSqared < 1)
    vz = (float)Math.sqrt(1.0 - lengthSqared);

  else {
    float length = (float)Math.sqrt(lengthSqared);
    vx /= length;
    vy /= length;
  }

  v[X] = vx;
  v[Y] = vy;
  v[Z] = vz;
}

/**
*  Computes a rotation matrix that map (rotate) vectors op onto oq.
*  The rotation is about an axis perpendicular to op and oq.
*
*
*  @param  rotationMatrix  : The 16 element rotation matrix to be filled in.
*  @param  op              : The 1st 3D vector.
*  @param  oq              : The 2nd 3D vector.
*
*
* */
private void setRotationMatrix(float[] rotationMatrix, float[] op, float[] oq) {

  Util.crossProduct3D(op, oq, a);
  float s = Util.length3D(a);
  float c = Util.dotProduct3D(op, oq);
  float t = 1 - c;

  float ax = a[X];
  float ay = a[Y];
  float az = a[Z];

  if (s > 0) {
    ax /= s;
    ay /= s;
    az /= s;
  }

  float tax = t*ax;
  float taxay = tax*ay, taxaz = tax*az;
  float saz = s*az, say = s*ay;
  rotationMatrix[0] = tax*ax + c;
  rotationMatrix[1] = taxay + saz;
  rotationMatrix[2] = taxaz - say;

  float tay = t*ay;
  float tayaz = tay*az;
  float sax = s*ax;
  rotationMatrix[4] = taxay - saz;
  rotationMatrix[5] = tay*ay + c;
  rotationMatrix[6] = tayaz + sax;

  rotationMatrix[8] = taxaz + say;
  rotationMatrix[9] = tayaz - sax;
  rotationMatrix[10] = t*az*az + c;

  rotationMatrix[3] = rotationMatrix[7] = rotationMatrix[11] =
  rotationMatrix[12] = rotationMatrix[13] = rotationMatrix[14] = 0;
  rotationMatrix[15] = 1;
}
}

