package saces.gl;
/*
 *                       s a c e s
 *      Simple Artificial Chemistry Experiment System
 *
 * A semester project Summer 2005 and a diploma project 2006
 *           Berne University of Applied Sciences
 *   HTI School of Engineering and Information Technology
 *             Anthony Aguillon, Daniel Noelpp
 *
 */

/**
 * Matrix manipulation functions
 *
 */
public class Matrix {

public static final float PI = (float)Math.PI;

/**
 *  Converts an angle in radians to degrees.
 *
 *  @param  rad  : The angle in radians to be converted.
 *  @return The angle in degrees.
**/
public static float rad2Deg(float rad) {

  return rad * 180/PI;

}

/**
*  Convert an angle in degrees to radians.
*
*  @param  deg  : The angle in degrees to be converted.
*  @return The angle in radians.
**/
public static float deg2Rad(float deg) {

  return deg*PI/180;

}


/**
*  Create a new 4x4 transformation matrix set to the identity matrix.
*      Equiv. to: T = I
*
*      I =  1   0   0   0
*           0   1   0   0
*           0   0   1   0
*           0   0   0   1
*
*  @return A reference to a 4x4 identity matrix is returned.
**/
public static float[] identity() {

  float[] mtx = new float[16];
  mtx[0] = mtx[5] = mtx[10] = mtx[15] = 1;

  return mtx;

}



/**
 *  Multiplies two 4x4 matricies together.
 *
 *  @param  mtx1  : The 1st matrix being multiplied.
 *  @param  mtx2  : The 2nd matrix being multiplied.
 *  @return The output matrix containing the input matrices multiplied.
 **/
public static float[] multiply(float mtx1[], float mtx2[]) {

  float nmtx[] = new float[16];

  float m0 = mtx1[0], m1 = mtx1[1], m2 = mtx1[2], m3 = mtx1[3];
  nmtx[0]  = m0*mtx2[0]  + m1*mtx2[4]  + m2*mtx2[8]   + m3*mtx2[12];
  nmtx[1]  = m0*mtx2[1]  + m1*mtx2[5]  + m2*mtx2[9]   + m3*mtx2[13];
  nmtx[2]  = m0*mtx2[2]  + m1*mtx2[6]  + m2*mtx2[10]  + m3*mtx2[14];
  nmtx[3]  = m0*mtx2[3]  + m1*mtx2[7]  + m2*mtx2[11]  + m3*mtx2[15];

  m0 = mtx1[4];	m1 = mtx1[5];	m2 = mtx1[6];	m3 = mtx1[7];
  nmtx[4]  = m0*mtx2[0]  + m1*mtx2[4]  + m2*mtx2[8]   + m3*mtx2[12];
  nmtx[5]  = m0*mtx2[1]  + m1*mtx2[5]  + m2*mtx2[9]   + m3*mtx2[13];
  nmtx[6]  = m0*mtx2[2]  + m1*mtx2[6]  + m2*mtx2[10]  + m3*mtx2[14];
  nmtx[7]  = m0*mtx2[3]  + m1*mtx2[7]  + m2*mtx2[11]  + m3*mtx2[15];

  m0 = mtx1[8];	m1 = mtx1[9];	m2 = mtx1[10];	m3 = mtx1[11];
  nmtx[8]  = m0*mtx2[0]  + m1*mtx2[4]  + m2*mtx2[8]  + m3*mtx2[12];
  nmtx[9]  = m0*mtx2[1]  + m1*mtx2[5]  + m2*mtx2[9]  + m3*mtx2[13];
  nmtx[10] = m0*mtx2[2]  + m1*mtx2[6]  + m2*mtx2[10] + m3*mtx2[14];
  nmtx[11] = m0*mtx2[3]  + m1*mtx2[7]  + m2*mtx2[11] + m3*mtx2[15];

  m0 = mtx1[12];	m1 = mtx1[13];	m2 = mtx1[14];	m3 = mtx1[15];
  nmtx[12] = m0*mtx2[0] + m1*mtx2[4] + m2*mtx2[8]  + m3*mtx2[12];
  nmtx[13] = m0*mtx2[1] + m1*mtx2[5] + m2*mtx2[9]  + m3*mtx2[13];
  nmtx[14] = m0*mtx2[2] + m1*mtx2[6] + m2*mtx2[10] + m3*mtx2[14];
  nmtx[15] = m0*mtx2[3] + m1*mtx2[7] + m2*mtx2[11] + m3*mtx2[15];

  return nmtx;
}

/**
 *      T = T*Rx
 *
 *      Rx =  1   0           0         0
 *            0   cos(a)      sin(a)    0
 *            0   -sin(a)     cos(a)    0
 *            0   0           0         1
 *
 *  @param  angle  : The rotation angle about the X-axis in radians.
 *  @param  mtx    : The 4x4 matrix to be multiplied by the rotation matrix.
 *  @return A reference to the output matrix "mtx" is returned.
 **/
public static float[] rotateX(float angle, float mtx[]) {

  float rsin = (float)Math.sin(angle);
  float rcos = (float)Math.cos(angle);

  for (int i = 0 ; i < 4 ; i++) {
    int i1 = i*4 + 1;
    int i2 = i1 + 1;
    float t = mtx[i1];
    mtx[i1] = t*rcos - mtx[i2]*rsin;
    mtx[i2] = t*rsin + mtx[i2]*rcos;
  }

  return mtx;
}

/**
 *      T = T*Ry
 *
 *      Ry =  1   0           0         0
 *            0   cos(a)      sin(a)    0
 *            0   -sin(a)     cos(a)    0
 *            0   0           0         1
 *
 *  @param  angle : The rotation angle about the Y-axis in radians.
 *  @param  mtx   : The 4x4 matrix to be multiplied by the rotation matrix.
 *  @return A reference to the output matrix "mtx" is returned.
 **/
public static float[] rotateY(float angle, float mtx[]) {

  float rsin = (float)Math.sin(angle);
  float rcos = (float)Math.cos(angle);

  for (int i = 0 ; i < 4 ; i++) {
    int i0 = i*4;
    int i2 = i0 + 2;
      float t = mtx[i0];
      mtx[i0] = t*rcos + mtx[i2]*rsin;
      mtx[i2] = mtx[i2]*rcos - t*rsin;
  }

  return mtx;
}

/**
 *      T = T*Rz
 *
 *      Rz =  1   0           0         0
 *            0   cos(a)      sin(a)    0
 *            0   -sin(a)     cos(a)    0
 *            0   0           0         1
 *
 *  @param  angle : The rotation angle about the Y-axis in radians.
 *  @param  mtx   : The 4x4 matrix to be multiplied by the rotation matrix.
 *  @return A reference to the output matrix "mtx" is returned.
 **/
public static float[] rotateZ(float angle, float mtx[]) {

  float rsin = (float)Math.sin(angle);
  float rcos = (float)Math.cos(angle);

  for (int i = 0 ; i < 4 ; i++) {
    int i0 = i*4;
    int i1 = i0 + 1;
      float t = mtx[i0];
      mtx[i0] = t*rcos - mtx[i1]*rsin;
      mtx[i1] = t*rsin + mtx[i1]*rcos;
  }

  return mtx;
}

}
