package saces.gl;
/*
 *                        s a c e s
 *      Simple Artificial Chemistry Experiment System
 *
 * A semester project Summer 2005 and a diploma project 2006
 *           Berne University of Applied Sciences
 *   HTI School of Engineering and Information Technology
 *             Anthony Aguillon, Daniel Noelpp
 *
 */

import net.java.games.jogl.*;
import net.java.games.jogl.impl.windows.PIXELFORMATDESCRIPTOR;

import java.awt.event.*;
import java.awt.*;

/**
 * GL routines which handle gl initialization,
 * reflection rendering, scene rendering, polygon rendering
 * viewing mode changes and gl state variables. It implements
 * the GLEventListener interface but is aggregated by the mediator
 * the GLEventListener interface but is aggregated by the mediator
 * which delegates GLEventListener method calls instead.
 *
 */
public class Routines implements GLEventListener, MouseListener,
  MouseWheelListener, MouseMotionListener, KeyListener {

// no lighting, flat shading
public static final int DETAIL_LEVEL_LOW    = 0;
// lighting, smooth shading
public static final int DETAIL_LEVEL_MEDIUM = 1;
// reflections, lighting, smooth shading
public static final int DETAIL_LEVEL_HIGH   = 2;

// current rendering detail level
public int detailLevel;
// gl lighting object
public Light light;
// reference to the gl bounding box
private Box box;
// gl viewing object (camera pos, frustum, etc.)
public View view;
// Display panel for fps, particle count, etc.
public ProcessDataPanel processDataPanel;
// Display panel for axis labels in orthographic viewing mode.
public AxisLabelPanel axisLabelPanel;
// virtual sphere
public BoundingSphere boundingSphere;
// GL drawing surface component
private GLCanvas glCanvas;
// GL animator thread
private Animator animator;
// virtual sphere helpers
private Point prevMouse;
private float[] mouseMatrix;
public float[] boundingSphereRotationMatrix;
private Point cueCenter = new Point();
private int cueRadius;

// animation control flags
public boolean isAnimatorRunning;
public boolean hasBeenRunning;

// cached screen width, set by reshape method
private int screenWidth;
// cached screen height, set by reshape method
private int screenHeight;

/**
 *  Constructor
 *
 *  @param detailLevel : see DETAIL_LEVEL enum
 *
 */
public Routines(int detailLevel) {

  GLCapabilities caps = new GLCapabilities();
  caps.setDoubleBuffered(true);
  caps.setHardwareAccelerated(true);

  glCanvas = GLDrawableFactory.getFactory().createGLCanvas(caps);

  animator                     = new Animator(glCanvas);
  boundingSphere               = new BoundingSphere();
  light                        = new Light();
  processDataPanel             = new ProcessDataPanel();
  axisLabelPanel               = new AxisLabelPanel();
  mouseMatrix                  = new float[16];
  boundingSphereRotationMatrix = Matrix.identity();

  this.detailLevel             = detailLevel;
  isAnimatorRunning            = false;
  hasBeenRunning               = false;

  glCanvas.addMouseListener(this);
  glCanvas.addMouseWheelListener(this);
  glCanvas.addMouseMotionListener(this);
  glCanvas.addKeyListener(this);

}

/**
 *   Set the simulation bounding box
 *
 *   @param box : Box
 *
 */
public void setBox(Box box) {
  this.box = box;
  view = new View(box.getDimensions());
}

/**
 *   Set the rendering detail level
 *
 *   @param detailLevel : see DETAIL_LEVEL enum
 *
 */
public void setDetailLevel(int detailLevel) {
  this.detailLevel = detailLevel;
}

public GLCanvas getCanvas() {
  return glCanvas;
}

/**
 * Starts the animation thread.
 *
 **/
public void startAnimationThread(boolean start) {
  if(start && !isAnimatorRunning) {
    if (!hasBeenRunning) {
      animator.start();
      hasBeenRunning = true;
    }
  }
}

/**
 *   GLEventListener method
 *
 *   @param glDrawable : GLDrawable
 *
 */
public void init(GLDrawable glDrawable) {

  GL gl = glDrawable.getGL();

  // set background (clear) color
  gl.glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

  // enable depth test
  gl.glEnable(GL.GL_DEPTH_TEST);
  // enable culling
  gl.glEnable(GL.GL_CULL_FACE);
  // enable smooth polygons
  gl.glEnable(GL.GL_POLYGON_SMOOTH);

  // clear stencil to contain 0s
  gl.glClearStencil(0);
  // set polygon fill mode to front, back will be culled
  gl.glPolygonMode(GL.GL_FRONT, GL.GL_FILL);
  // hint for nicest polygon quality.
  gl.glHint(GL.GL_POLYGON_SMOOTH_HINT, GL.GL_NICEST);

  light.initialize(gl, box.getDimensions());

 }

public void display(GLDrawable glDrawable) {
  // implemented by the mediator
}

/**
 * Canvas reshape.
 *
 **/
public void reshape(GLDrawable glDrawable, int i, int i1, int i2, int i3) {

  processDataPanel.reshape(i2, i3);
  axisLabelPanel.reshape(i2, i3);
  view.setViewport(glDrawable, i2, i3);

  screenWidth = i2;
  screenHeight = i3;

  applyDetailLevelChanges(glDrawable.getGL());
}

public void displayChanged(GLDrawable glDrawable, boolean b, boolean b1) {
  // implemented in the mediator
}

public void mouseDragged(MouseEvent e) {

  Point newMouse = e.getPoint();

  if (newMouse.x != prevMouse.x || newMouse.y != prevMouse.y) {

    // Use bounding sphere to determine an incremental rotation matrix.
    boundingSphere.makeRotationMatrix(prevMouse, newMouse, cueCenter,
      cueRadius, mouseMatrix);

    // Multiply overall rotation matrix by incremental rotation-matrix
    // for new rotation.
    boundingSphereRotationMatrix = Matrix.multiply(
      boundingSphereRotationMatrix, mouseMatrix);

    fixRotationMatrix();

    prevMouse = newMouse;
  }
}

public void mousePressed(MouseEvent e) {
  prevMouse = e.getPoint();
}

/**
 *  Mouse wheel has moved. The mouse wheel rotation
 *  currently moves the gl camera to and from the
 *  viewing center
 *
 *  @param e : see MouseMotionListener interface
 *
 **/
public void mouseWheelMoved(MouseWheelEvent e) {

  int steps = e.getWheelRotation();

  if(steps < 0){
    view.stepCloserToCenter(steps);
  }
  else{
    view.stepAwayFromCenter(steps);
  }
}

/**
 * Applies detail level modifications in the OpenGL context.
 *
 **/
public void applyDetailLevelChanges(GL gl) {

  switch (detailLevel)
  {
    case Routines.DETAIL_LEVEL_LOW :
      gl.glDisable(GL.GL_DEPTH_TEST);
      gl.glDisable(GL.GL_LIGHTING);
      gl.glDisable(GL.GL_CULL_FACE);
      gl.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_LINE);
    break;

    case Routines.DETAIL_LEVEL_MEDIUM :
      gl.glEnable(GL.GL_DEPTH_TEST);
      gl.glShadeModel(GL.GL_SMOOTH);
      gl.glEnable(GL.GL_LIGHTING);
      gl.glEnable(GL.GL_CULL_FACE);
      gl.glCullFace(GL.GL_BACK);
      gl.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_FILL);
    break;

    case Routines.DETAIL_LEVEL_HIGH :
      gl.glEnable(GL.GL_DEPTH_TEST);
      gl.glShadeModel(GL.GL_SMOOTH);
      gl.glEnable(GL.GL_LIGHTING);
      gl.glEnable(GL.GL_CULL_FACE);
      gl.glCullFace(GL.GL_BACK);
      gl.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_FILL);
    break;

    default:
    break;
  }
}

/**
 * Applies a view change in the OpenGL context.
 *
 **/
public void applyViewChange(GLDrawable glDrawable, int screenWidth,
                            int screenHeight) {

  GL gl = glDrawable.getGL();

  if(view.getViewMode() == View.ORTHOGRAPHIC) {
      gl.glEnable(GL.GL_SCISSOR_TEST);
      gl.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_LINE);
  }
  else {
      gl.glDisable(GL.GL_SCISSOR_TEST);
      gl.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_FILL);
      view.resetPerspective();
  }

  reshape(glDrawable, 0, 0, screenWidth, screenHeight);

}

/**
 * Draws data panels containing data such as frame rate, particle count,
 * Axis captions.
 *
 **/
public void drawPanels(GLDrawable glDrawable, int viewportId) {

  switch(viewportId) {

    case View.VIEWPORT_ID_TOP_LEFT :
      axisLabelPanel.setHorizontalAxisName("x");
      axisLabelPanel.setVerticalAxisName("y");
      axisLabelPanel.draw(glDrawable);
    break;

    case View.VIEWPORT_ID_TOP_RIGHT :
      axisLabelPanel.setHorizontalAxisName("z");
      axisLabelPanel.setVerticalAxisName("y");
      axisLabelPanel.draw(glDrawable);
    break;

    case View.VIEWPORT_ID_BOTTOM_LEFT :
      axisLabelPanel.setHorizontalAxisName("x");
      axisLabelPanel.setVerticalAxisName("z");
      axisLabelPanel.draw(glDrawable);
    break;

    case View.VIEWPORT_ID_BOTTOM_RIGHT :


    break;
  }

}

/**
 * Applies the state changes in the OpenGL context, required to
 * set up the orthogonal view.
 *
 **/
public void applyOrthoViewportSetup(GLDrawable glDrawable, int viewportId) {

  int d;
  GL gl = glDrawable.getGL();

  if (screenWidth > screenHeight) {
     d = screenHeight;
  }
  else {
     d = screenWidth;
  }

  view.setupOrthoViewport(gl, d, d, viewportId);

}

/**
 * Applies the state changes in the OpenGL context, required to
 * set up the default perspective view.
 *
 **/
public void applyPerspectiveViewportSetup(GLDrawable glDrawable) {

  GL gl = glDrawable.getGL();

  view.updateViewpoint(glDrawable);
  gl.glMultMatrixf(boundingSphereRotationMatrix);

  // make center of box { 0,0,0 } to coincide with bounding sphere
  float[] dims = box.getDimensions();
  gl.glTranslatef(-dims[0] / 2f, -dims[1] / 2f, -dims[2] / 2f);
}

/**
 * Recalculates the rotation matrix (for the virtual sphere) and
 * corrects any rounding errors which might have occured.
 *
 **/
private void fixRotationMatrix() {

  boundingSphereRotationMatrix[3]  = 0.0f;
  boundingSphereRotationMatrix[7]  = 0.0f;
  boundingSphereRotationMatrix[11] = 0.0f;
  boundingSphereRotationMatrix[12] = 0.0f;
  boundingSphereRotationMatrix[13] = 0.0f;
  boundingSphereRotationMatrix[14] = 0.0f;
  boundingSphereRotationMatrix[15] = 1.0f;

  float fac;

  if((fac = (float)Math.sqrt
      ((boundingSphereRotationMatrix[0] * boundingSphereRotationMatrix[0]) +
       (boundingSphereRotationMatrix[4] * boundingSphereRotationMatrix[4]) +
       (boundingSphereRotationMatrix[8] * boundingSphereRotationMatrix[8])))
         != 1.0f) {

           if (fac != 0.0f) {
             fac = 1.0f / fac;
             boundingSphereRotationMatrix[0] *= fac;
             boundingSphereRotationMatrix[4] *= fac;
             boundingSphereRotationMatrix[8] *= fac;
           }
  }

  if ((fac = (float)Math.sqrt
       ((boundingSphereRotationMatrix[1] * boundingSphereRotationMatrix[1]) +
        (boundingSphereRotationMatrix[5] * boundingSphereRotationMatrix[5]) +
        (boundingSphereRotationMatrix[9] * boundingSphereRotationMatrix[9])))
          != 1.0f) {

            if (fac != 0.0f) {
              fac = 1.0f / fac;
              boundingSphereRotationMatrix[1] *= fac;
              boundingSphereRotationMatrix[5] *= fac;
              boundingSphereRotationMatrix[9] *= fac;
            }
  }

  if ((fac = (float)Math.sqrt
      ((boundingSphereRotationMatrix[2] * boundingSphereRotationMatrix[2]) +
       (boundingSphereRotationMatrix[6] * boundingSphereRotationMatrix[6]) +
       (boundingSphereRotationMatrix[10] * boundingSphereRotationMatrix[10])))
          != 1.0f) {

          if (fac != 0.0f) {
            fac = 1.0f / fac;
            boundingSphereRotationMatrix[2] *= fac;
            boundingSphereRotationMatrix[6] *= fac;
            boundingSphereRotationMatrix[10] *= fac;
          }
  }
}

/**
 * Sets up the  sphere cue center and radius.
 *
 *  @param w bounding width of experiment box
 *  @param h bounding height of experiment box
 *
 **/
public void setupBoundingSphere(int w, int h) {
  cueCenter.x = w / 2;
  cueCenter.y = h / 2;
  cueRadius = Math.min(w - 20, h - 20) / 2;
}

// see MouseListener
public void mouseEntered(MouseEvent e) {}
public void mouseExited(MouseEvent e) {}
public void mouseClicked(MouseEvent e) {}
public void mouseReleased(MouseEvent e) {}

// see MouseMotionListener
public void mouseMoved(MouseEvent e) {}

public void keyPressed(KeyEvent e) {

  switch(e.getKeyCode())
  {
    case KeyEvent.VK_UP :
      view.stepCloserToCenter(-1);
    break;

    case KeyEvent.VK_DOWN :
     view.stepAwayFromCenter(1);
    break;

    default:
  }
}

public void keyTyped(KeyEvent e) {}
public void keyReleased(KeyEvent e) {}
}

