package saces.gl;
/*
*                       s a c e s
*      Simple Artificial Chemistry Experiment System
*
* A semester project Summer 2005 and a diploma project 2006
*           Berne University of Applied Sciences
*   HTI School of Engineering and Information Technology
*             Anthony Aguillon, Daniel Noelpp
*
*/

/**
 * A class of utility functions used in 3D modeling.
 *
 */
public class Util {

private static final int X = 0;
private static final int Y = 1;
private static final int Z = 2;

/**
 *
 * Normalizes the vector in vector and returns it in output
 *
 * @param vector : float[3] non-normalized vector
 * @param output : float[3] normalized vector
 * @return output
 *
 */
public static float[] normalize(float[] vector, float[] output) {

  float v0 = vector[X], v1 = vector[Y], v2 = vector[Z];
  float length = (float)Math.sqrt( v0*v0 + v1*v1 + v2*v2 );

  if (length == 0)
    length = 1;

  output[X] = v0/length;
  output[Y] = v1/length;
  output[Z] = v2/length;

  return output;
}

/**
 *
 * Calculates the normal vector of the surface described by the
 * three points pnt1, pnt2 and point3
 *
 * @param pnt1 : A point on the plane
 * @param pnt2 : Another point on the plane
 * @param pnt3 : Another point on the plane
 * @param output : The normal vector of the plane defined by pnt1, pnt2 and pn3.
 *
 * @return output
 *
 */
public static float[] normal(float[] pnt1, float[] pnt2, float[] pnt3,
                             float[] output) {

  float v1x = pnt1[X] - pnt2[X];
  float v1y = pnt1[Y] - pnt2[Y];
  float v1z = pnt1[Z] - pnt2[Z];

  float v2x = pnt2[X] - pnt3[X];
  float v2y = pnt2[Y] - pnt3[Y];
  float v2z = pnt2[Z] - pnt3[Z];

  output[X] = v1y*v2z - v1z*v2y;
  output[Y] = v1z*v2x - v1x*v2z;
  output[Z] = v1x*v2y - v1y*v2x;

  return normalize(output, output);
}

/**
 *
 * Calculates the distance of a point to the center (0, 0, 0), or
 * the length of the vector
 *
 * @param vector : point
 *
 * @return length
 *
 */
public static float length3D(float[] vector) {

  double dx = vector[X];
  double dy = vector[Y];
  double dz = vector[Z];

  return (float)Math.sqrt(dx*dx + dy*dy + dz*dz);
}

/**
 *
 * Calculates the dot product of the vectors a and b.
 *
 * @param a : vector 1
 * @param b : vector 2
 *
 * @return dot product of vector 1 with vector 2
 *
 */
public static float dotProduct3D(float[] a, float[] b) {

  return (a[X]*b[X] + a[Y]*b[Y] + a[Z]*b[Z]);
}

/**
 *
 * Calculates the cross product of the vectors a and b.
 *
 * @param A : vector 1
 * @param B : vector 2
 *
 * @return dot product of vector 1 with vector 2
 *
 */
public static float[] crossProduct3D(float[] A, float[] B,
                                           float[] AcrossB) {

  AcrossB[X] = A[Y] * B[Z] - A[Z] * B[Y];
  AcrossB[Y] = A[Z] * B[X] - A[X] * B[Z];
  AcrossB[Z] = A[X] * B[Y] - A[Y] * B[X];

  return AcrossB;
}

/**
 *
 * Calculates the distance of a point to the center (0, 0, 0).
 *
 * @param dx : X coordinate of point
 * @param dy : Y coordinate of point
 * @param dz : Z coordinate of point
 *
 * @return distance
 *
 */
public static float distance(double dx, double dy, double dz) {

  return (float)Math.sqrt(dx*dx + dy*dy + dz*dz);
}

/**
 *
 * Calculates the distance between two points.
 *
 * @param a : x, y and z components of point 1
 * @param b : x, y, and z components of point 2
 *
 * @return distance
 *
 */
public static float distance(float[] a, float[] b) {

  float a1 = a[X] - b[X];
  float a2 = a[Y] - b[Y];
  float a3 = a[Z] - b[Z];

  a1 *= a1;
  a2 *= a2;
  a3 *= a3;

  return (float)Math.sqrt(a1 + a2 + a3);
}

}


