package saces.gl;
/*
 *                        s a c e s
 *      Simple Artificial Chemistry Experiment System
 *
 * A semester project Summer 2005 and a diploma project 2006
 *           Berne University of Applied Sciences
 *   HTI School of Engineering and Information Technology
 *             Anthony Aguillon, Daniel Noelpp
 *
 */

import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;
import net.java.games.jogl.GLDrawable;
import saces.sim.Particle;

/**
 *  The view class handles GL viewing aspects such as viewport,
 *  viewpoint, viewing type (orthographic, perspective, etc.),
 *  camera positioning (transformations).
 *
 *
 * */
public class View
{

// Constants for viewMode
public static final int DEFAULT_PERSPECTIVE = 1;
public static final int ORTHOGRAPHIC = 2;
public static final int PARTICLE_PERSPECTIVE = 3;

// One of the static final definitions
// DEFAULT_PERSPECTIVE, ORTHOGRAPHIC, PARTICLE_PERSPECTIVE
private int viewMode;
// Near clipping plane for perspective view
private float nearClippingPlane;
// Far clipping plance for perspective view
private float farClippingPlane;
// Distance from the camera position to the viewing center
private float distanceToCenterFactor;
// Random particle to follow when in PARTICLE_PERSPECTIVE
private Particle particle;
// Viewport angle constant
private static final float viewportAngle = 45.0f;
// Maximum distance of camera from viewing center
private final float DISTANCE_TO_CENTER_FACTOR_MAX = 3.0f;
// Minimum distance of camera to viewing center
private final float DISTANCE_TO_CENTER_FACTOR_MIN = 1.0f;
// Steps when moving camera to or from the viewing center
private final float DISTANCE_STEP = 0.1f;
// The experiment bounding box
private float box[];
// Margin for orthographic view (view frame)
private final float orthoMargin = 1.5f;

// Viewport location identifiers for orthographic view mode
public static final int VIEWPORT_ID_TOP_LEFT     = 0;
public static final int VIEWPORT_ID_TOP_RIGHT    = 1;
public static final int VIEWPORT_ID_BOTTOM_LEFT  = 2;
public static final int VIEWPORT_ID_BOTTOM_RIGHT = 3;

// Camera position
public float[] position;
// Vanishing point position
public float[] vanishingPointCoordinates;
// Vector pointing upward
public float[] upVector;

/**
 * Constructor
 *
 * @param box : The experiment bounding box
 *
 */
public View(float[] box)
{

  this.box = new float[3];
  this.box[0] = box[0];
  this.box[1] = box[1];
  this.box[2] = box[2];

  viewMode = DEFAULT_PERSPECTIVE;

  position = new float[3];
  upVector = new float[3];
  vanishingPointCoordinates = new float[3];
  distanceToCenterFactor = DISTANCE_TO_CENTER_FACTOR_MAX;

  resetPerspective();

}

/**
 * Resets the perspective. The vanishing point reset to the
 * center of the experiment bounding box
 *
 */
public void resetPerspective() {

  nearClippingPlane = 1.0f;
  farClippingPlane = getMaxDimension(this.box) * 5.0f;

  this.vanishingPointCoordinates[0] = box[0] / 2.0f;
  this.vanishingPointCoordinates[1] = box[1] / 2.0f;
  this.vanishingPointCoordinates[2] = box[2] / 2.0f;

  this.position[0] = this.vanishingPointCoordinates[0]
          * distanceToCenterFactor;
  this.position[1] = this.vanishingPointCoordinates[1]
          * distanceToCenterFactor;
  this.position[2] = this.vanishingPointCoordinates[2]
          * distanceToCenterFactor;

  // up is the model's Y+ axis
  this.upVector[0] = 0.0f;
  this.upVector[1] = 1.0f;
  this.upVector[2] = 0.0f;

}

/**
 * Moves the camera closer to the center.
 *
 *  @param steps : The steps are multiplied by the <steps>
 *                 factor to get actual distance
 */
public void stepCloserToCenter(int steps) {

  distanceToCenterFactor += (DISTANCE_STEP * (float)steps);
  if(distanceToCenterFactor < DISTANCE_TO_CENTER_FACTOR_MIN){
    distanceToCenterFactor = DISTANCE_TO_CENTER_FACTOR_MIN;
  }
  resetPerspective();

}

/**
 * Moves the camera away from the center.
 *
 *  @param steps : The steps are multiplied by the <steps>
 *                 factor to get actual distance
 */
public void stepAwayFromCenter(int steps) {

  distanceToCenterFactor -= (DISTANCE_STEP * (float)steps * -1);
  if(distanceToCenterFactor > DISTANCE_TO_CENTER_FACTOR_MAX){
    distanceToCenterFactor = DISTANCE_TO_CENTER_FACTOR_MAX;
  }
  resetPerspective();
}

/**
 * Returns the largest value in f
 *
 *  @param f : Array of any length
 *
 */
public float getMaxDimension(float[] f) {

  float fMax = .0f;

  for (float aF : f) {
    if (aF > fMax) {
      fMax = aF;
    }
  }

  return fMax;
}

/**
 * Sets the viewport according to :
 *   height and width, and viewMode
 *
 *  @param gLDrawable : GLDrawable
 *  @param width : new viewport width, should equal to screen width
 *  @param height : new viewport height, should equal to screen height
 *
 */
public void setViewport(GLDrawable gLDrawable, int width, int height) {

  GL gl = gLDrawable.getGL();
  GLU glu = gLDrawable.getGLU();

  if(viewMode == View.ORTHOGRAPHIC) {

   gl.glMatrixMode(GL.GL_PROJECTION);
   gl.glLoadIdentity();

   float maxDim = getMaxDimension(box);

   gl.glOrtho(-orthoMargin, maxDim + orthoMargin, -orthoMargin,
           maxDim + orthoMargin, -(maxDim + orthoMargin),
           maxDim + orthoMargin);
 }
 else{

   float aspectRatio = (float) width / (float) height;

   gl.glViewport(0, 0, width, height);
   gl.glMatrixMode(GL.GL_PROJECTION);
   gl.glLoadIdentity();

   glu.gluPerspective(viewportAngle, aspectRatio,
           nearClippingPlane, farClippingPlane);

 }

 gl.glMatrixMode(GL.GL_MODELVIEW);
 gl.glLoadIdentity();

}

/**
 *  Sets the view mode
 *
 *  @param viewMode : DEFAULT_PERSPECTIVE or
 *                    ORTHOGRAPHIC or
 *                    PARTICLE_PERSPECTIVE
 *
 */
public void setViewMode(int viewMode) {
  this.viewMode = viewMode ;
}


  /**
   *  Sets up the one of the four viewports for orthographic projection
   *  The screen is divided into four equal rectangles.
   *
   *  @param gl : GL
   *  @param screenWidth : viewport width, equals to screen width
   *  @param screenHeight : viewport width, equals to screen height
   *
   */
  public void setupOrthoViewport(GL gl, int screenWidth,
                              int screenHeight, int viewportId) {

    switch(viewportId) {

      case VIEWPORT_ID_TOP_LEFT :

        gl.glViewport(0, (int)(screenHeight / 2f),
          (int)(screenWidth / 2f),
          (int)(screenHeight / 2f));

        gl.glScissor(0, (int)(screenHeight / 2f),
          (int)(screenWidth / 2f),
          (int)(screenHeight / 2f));

      break;

      case VIEWPORT_ID_TOP_RIGHT :

        gl.glViewport(screenWidth / 2,
          screenHeight / 2,
          screenWidth / 2, screenHeight / 2);
        gl.glScissor(screenWidth / 2,
          screenHeight / 2,
          screenWidth / 2, screenHeight / 2);

      break;

      case VIEWPORT_ID_BOTTOM_LEFT:

        gl.glViewport(0, 0,
          screenWidth / 2, screenHeight / 2);
        gl.glScissor(0, 0,
          screenWidth / 2, screenHeight / 2);

      break;

      case VIEWPORT_ID_BOTTOM_RIGHT:

        gl.glViewport(screenWidth / 2,
          0, screenWidth / 2, screenHeight / 2);
        gl.glScissor(screenWidth / 2,
          0, screenWidth / 2, screenHeight / 2);

      break;

      default:

    }
  }

/**
 *  Sets the particle to be followed in particle perspective mode.
 *
 *  @param particle : The particle to be followd
 *
 */
public void setParticle(Particle particle) {
  nearClippingPlane = particle.radius + 0.2f;
  this.particle = particle;
}

/**
 *  Sets the viewpoint or camera position according to the view mode
 *
 *  @param gLDrawable : GLDrawable
 *
 */
public void updateViewpoint(GLDrawable gLDrawable) {

  if(viewMode == PARTICLE_PERSPECTIVE)
    updateCameraPosition();

  GLU glu = gLDrawable.getGLU();
  GL gl = gLDrawable.getGL();

  gl.glLoadIdentity();

  glu.gluLookAt(position[0], position[1],
          position[2],
          0,
          0,
          0,
          upVector[0],
          upVector[1],
          upVector[2]);
}

public void updateCameraPosition() {

  if(particle != null) {
    position[0] = particle.position[0];
    position[1] = particle.position[1];
    position[2] = particle.position[2];
  }

}

/**
 *  Returns the view mode
 *
 *  @return  DEFAULT_PERSPECTIVE
 *  @return  ORTHOGRAPHIC
 *  @return  PARTICLE_PERSPECTIVE
 *
 */
public int getViewMode() {
  return viewMode;
}


}

