/*
 *                        s a c e s
 *      Simple Artificial Chemistry Experiment System
 *
 * A semester project Summer 2005 and a diploma project 2006
 *           Berne University of Applied Sciences
 *   HTI School of Engineering and Information Technology
 *             Anthony Aguillon, Daniel Noelpp
 *
 */
package saces.pnp;

import saces.Util;
import static saces.Util.*;
import saces.sim.Collision;
import saces.sim.Particle;
import saces.sim.Partition;
import saces.sim.Simulation;

import static java.lang.Math.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;


/**
 * Detect particle collisions but only with particles in the same partition.
 * Neglect particles overlapping from the neighbouring partitions.
 * <p/>
 * Partitionized Collision Detection works almost in linear time if the number
 * of partitions is made dependent of the number of particles. Let's say we have
 * 10 particles every partition on average. Then we loop for each partition and
 * do collision detection for these 10 partitions only.
 */
public class DetectorPartitionized implements Detector {

/**
 * The name of the count-per-partition property. What is this property?
 * <p/>
 * How many particles does a partition contain on average? Let's say for example
 * we have 2500 particles and want an average count of 15 particles per
 * partition. We need 167 partitions (rounding to whole numbers). We could split
 * in the x-axis 6 times, in the y-axis 5 times and in the z-axis again 6 times
 * and have thusly 180 partitions and have 13.9 particles per partition on
 * average. This example shows that this field is only a hint and not a strict
 * requirement (which would be impossible because of rounding and even
 * worse, because many integers aren't factors of three integers).
 */
public static final String COUNT_PER_PARTITION
    = "DetectorPartitionized.particleCountPerPartition";

//private static final String ALREADY_COLLIDED = "AlreadyCollided";

public Collision[] detect(Particle[] particles, Simulation simulation) {
    if (simulation.getPartitions() == null) initPartitions(simulation);
    else
        repartitionize(particles, simulation);

    // It is estimated that about a fifth of the particles collide.
    int initialCapacity = particles.length / 5;
    List<Collision> collisions = new ArrayList<Collision>(initialCapacity);

    // The MARK object marks particles already in collision. It is
    // recreated for each call of detect(), so no need to remove MARK
    // objects to clear the collision marks after detection.
    final Object MARK = new Object();
    for (Particle p1 : particles) {
        for (Particle p2 : p1.partition.particles) {
            if (p1 == p2) continue;
            if (p1.alreadyCollided == MARK) continue;
            if (p2.alreadyCollided == MARK) continue;
            //if (p1.get(ALREADY_COLLIDED) == MARK) continue;
            //if (p2.get(ALREADY_COLLIDED) == MARK) continue;

            // A Particle doesn't collide with itself
            assert p1 != p2;

            if (p1.overlaps(p2)) {
                // The collision detector should only report particles ahich
                // really are in the simulation. This is checked here.
                assert simulation.contains(p1) : simulation.missing(p1);
                assert simulation.contains(p2) : simulation.missing(p2);

                collisions.add(new Collision(p1, p2));
                p1.alreadyCollided = MARK;
                p2.alreadyCollided = MARK;
                //p1.put(ALREADY_COLLIDED, MARK);
                //p2.put(ALREADY_COLLIDED, MARK);
            }
        }
    }

    // Assert that no particle is in more than one collision pair.
    // If we don't do this, the merge reaction will make problems if a
    // particle is already merged and yet found in another collision pair.
    if (ASSERTS_ENABLED) {
        Set<Particle> collidedParticles = new HashSet<Particle>();
        for (Collision collision : collisions) {
            assert !collidedParticles.contains(collision.educt1);
            assert !collidedParticles.contains(collision.educt2);

            collidedParticles.add(collision.educt1);
            collidedParticles.add(collision.educt2);
        }
    }

    return collisions.toArray(new Collision[0]);
}

// ATTENTION: some duplicate code (because of efficiency) between
// initPartitions() and repartitionize()!
// BE CAREFUL TO DUPLICATE THE MODIFICATIONS AS WELL!!
private void initPartitions(Simulation simulation) {
    int particleCount = simulation.getTotalParticleCount();

    int countPerPartition = simulation.getExperiment()
        .getInt(COUNT_PER_PARTITION, 12);
    int partitionCount = particleCount / countPerPartition;
    if (partitionCount < 1) partitionCount = 1;

    simulation.getLogger().writeProperty("RequestedPartitionCount",
        partitionCount);

    // Now find three numbers nx, ny, nz which multiplied are roughly
    // the same as partitionCount. It is impossible to get the exact
    // number because many numbers aren't factors of three numbers.
    int nx = (int)round(cbrt(partitionCount));
    int ny = (int)round(sqrt(((double)partitionCount) / nx));
    int nz = (int)round(((double)partitionCount) / nx / ny);
    partitionCount = nx * ny * nz;

    simulation.getLogger().writeProperty("PartitionCount[X]", nx);
    simulation.getLogger().writeProperty("PartitionCount[Y]", ny);
    simulation.getLogger().writeProperty("PartitionCount[Z]", nz);
    simulation.getLogger().writeProperty("AverageParticleCountPerPartition",
        ((float)particleCount) / partitionCount);

    Partition[][][] partitions = new Partition[nx][][];
    for (int ix = 0; ix < nx; ix++) {

        partitions[ix] = new Partition[ny][];
        for (int iy = 0; iy < ny; iy++) {

            partitions[ix][iy] = new Partition[nz];
            for (int iz = 0; iz < nz; iz++) {

                partitions[ix][iy][iz] = new Partition(ix, iy, iz,
                    simulation);
            }
        }
    }
    simulation.setPartitions(partitions);

    // initial partitionizing of the particles
    float[] bounds = simulation.cache.bounds;

    for (Particle particle : simulation.getParticles()) {

        float[] p = particle.position;
        int ix = (int)floor(p[X] * nx / bounds[X]);
        int iy = (int)floor(p[Y] * ny / bounds[Y]);
        int iz = (int)floor(p[Z] * nz / bounds[Z]);

        try {
            particle.partition = partitions[ix][iy][iz];
        }
        catch (IndexOutOfBoundsException ioobe) {
            particle.partition = partitions[0][0][0];
            simulation.getLogger().writeProperty(
                "Partitions.IndexOutOfBoundsException",
                "ix=" + ix + " iy=" + iy + " iz=" + iz + " for " + particle);
        }
        particle.partition.particles.add(particle);
    }
}

private void repartitionize(Particle[] particles, Simulation simulation) {

    Partition[][][] partitions = simulation.getPartitions();

    int nx = partitions.length - 1;
    int ny = partitions[0].length - 1;
    int nz = partitions[0][0].length - 1;
    float w = simulation.cache.bounds[X];
    float h = simulation.cache.bounds[Y];
    float d = simulation.cache.bounds[Z];

    for (Particle particle : particles) {
        float[] p = particle.position;
        int ix = (int)floor(p[X] * nx / w);
        int iy = (int)floor(p[Y] * ny / h);
        int iz = (int)floor(p[Z] * nz / d);

        Partition newPartition;
        try {
            newPartition = partitions[ix][iy][iz];
        }
        catch (IndexOutOfBoundsException ioobe) {
            newPartition = partitions[0][0][0];
            simulation.getLogger().writeProperty(
                "Partitions.IndexOutOfBoundsException",
                "ix=" + ix + " iy=" + iy + " iz=" + iz + " for " + particle
                    + "at " + (new Throwable().getStackTrace()[0]));
        }
        Partition oldPartition = particle.partition;
        if (oldPartition == newPartition) continue;

        // Attention: particle count may decrease during the simulation
        // This has to work for small particle counts as well
        // Attention: initpartitionize() is not called again NOW
        if (oldPartition != null) {
            boolean ok = oldPartition.particles.remove(particle);
            assert ok;
        }
        if (newPartition != null) {
            newPartition.particles.add(particle);
            particle.partition = newPartition;
        }

    }

    if (Util.ASSERTS_ENABLED) {
        // Do particles' partitions contain the particle?
        for (Particle p : simulation.getParticles())
            assert p.partition.particles.contains(p) : p;

        // Do the partitions contain all particles?
        Set<Particle> partitionized = new HashSet<Particle>();
        partitions = simulation.getPartitions();
        for (Partition[][] partitions2 : partitions)
            for (Partition[] partitions1 : partitions2)
                for (Partition partition : partitions1)
                    partitionized.addAll(partition.particles);

        Set<Particle> lostParticles = new HashSet<Particle>(partitionized);
        lostParticles.removeAll(Arrays.asList(simulation.getParticles()));
        assert lostParticles.isEmpty() : lostParticles;

        // Are all particles in partitions?
        List<Particle> particleList = Arrays.asList(simulation.getParticles());
        lostParticles = new HashSet<Particle>(particleList);
        lostParticles.removeAll(partitionized);
        assert lostParticles.isEmpty() : lostParticles;
    }
}

}
