/*
 *                        s a c e s
 *      Simple Artificial Chemistry Experiment System
 *
 * A semester project Summer 2005 and a diploma project 2006
 *           Berne University of Applied Sciences
 *   HTI School of Engineering and Information Technology
 *             Anthony Aguillon, Daniel Noelpp
 *
 */
package saces.pnp;

import saces.Util;
import static saces.Util.*;
import saces.exp.Experiment;
import saces.sim.Collision;
import saces.sim.Particle;
import saces.sim.Simulation;

import java.util.Random;


/**
 * Randomly distribute particles: position has equal distribution and the
 * velocity components have gaussian distribution around 0 with standard
 * deviation equal to the experiment property {@code "DistributorRandom.stdDev"}.
 * This results in a Maxwell-Boltzmann distribution of the particle speeds. The
 * pressure in the experiment is being ignored by this distributor.
 */
public class DistributorRandom implements Distributor {

/**
 * A sanity check so that there aren't too many and too big spheres in the box.
 * The ratio of the volume of the box and of the spheres shouldn't exceed
 * VOLUME_RATIO : 1.
 */
public static final float VOLUME_RATIO = 15f;

public static final String
    KEY_DISTRIBUTOR_RANDOM_STDDEV = "DistributorRandom.stdDev";
public static final float
    DEFAULT_DISTRIBUTOR_RANDOM_STDEF = 1f;


/**
 * Randomly distribute particles. The standard deviation of the gaussian
 * distribution of the velocity components is the experiment property {@code
 * "DistributorRandom.stdDev"}.
 * <p/>
 * These are the invariants: <ul><li>No particles collides with another. <li>No
 * particle touches a box wall or even penetrates it.</ul>
 *
 * @throws IllegalArgumentException if there are too many or too big particles
 * (the volume ratio was exceeded or a particle couldn't be placed without
 * intersecting another particle).
 */
public void distribute(Particle[] particles, Simulation simulation) {
    distributePosition(particles, simulation);
   distributeVelocity(particles, simulation,
       simulation.getExperiment().getFloat(
            KEY_DISTRIBUTOR_RANDOM_STDDEV,
            DEFAULT_DISTRIBUTOR_RANDOM_STDEF));
}

public void distributePosition(Particle[] particles, Simulation simulation) {
    Experiment exp = simulation.getExperiment();
    Random random = exp.getRandom();

    // A sanity check that there aren't too many and too big spheres
    // Compare the total volume sum of the bounding boxes with the
    // volume of  the box. The ratio shouldn't exceed 1 : VOLUME_RATIO.
    // This is neccessary because in the distribution we make sure
    // that no two spheres overlap. If we discover that two spheres
    // overlap we redistribute the second sphere and this get increasingly
    // difficult when the box gets fuller and fuller.
    float boxVolume = exp.getWidth() * exp.getHeight() * exp.getDepth();
    float particlesVolume = 0f;
    for (Particle particle : particles) {
        float radius = particle.radius;
        particlesVolume += radius * radius * radius;
    }
    float ratio = VOLUME_RATIO * particlesVolume / boxVolume;
    if (ratio > 1)
        throw new IllegalArgumentException(""
            + "Too small a box (need to enlarge by ca. "
            + String.format("%1.1f", ratio) + " times)");

    // Now that we have enough space for the spheres we can try to
    // distribute randomly without they overlapping each other.
    for (Particle particle : particles) {
        float radius = particle.radius;
        float[] center = particle.position;
        float[] bounds = {
            exp.getWidth() - 2 * radius,
            exp.getHeight() - 2 * radius,
            exp.getDepth() - 2 * radius,
        };

        // random position and check overlap
        int count = 0;
        randomLoop:
        while (true) {
            if (count++ > 100) { // after a 100 tries, give up
                throw new IllegalArgumentException(""
                    + "Not enough place for this particle: " + particle);
            }

            center[X] = random.nextFloat() * bounds[X] + radius;
            center[Y] = random.nextFloat() * bounds[Y] + radius;
            center[Z] = random.nextFloat() * bounds[Z] + radius;

            for (Particle particle2 : particles) {
                if (particle2 == particle) break randomLoop;
                if (particle.overlaps(particle2)) continue randomLoop;
            }
        }
    }

    // don't do invariant checks if that compile-time constant is false
    if (!Util.ASSERTS_ENABLED) return;

    if (ASSERTS_ENABLED) {
        for (Particle p : particles) {
            float r = p.radius;
            float[] center = p.position;
            float[] bounds = {exp.getWidth(), exp.getHeight(), exp.getDepth()};

            // No particle is (partially) outside the bounds
            assert center[X] - r >= 0;
            assert center[X] + r <= bounds[X];
            assert center[Y] - r >= 0;
            assert center[Y] + r <= bounds[Y];
            assert center[Z] - r >= 0;
            assert center[Z] + r <= bounds[Z];
        }

        // No particle is collided with another
        DetectorSimple scd = new DetectorSimple();
        Collision[] collided = scd.detect(particles, simulation);
        assert collided.length == 0;
    }
}
/**
 * Randomly distribute particles. The standard deviation of the gaussian
 * distribution of the velocity components is given as parameter stdDev. This
 * method is different to {@linkplain DistributorMaxwellBoltzmann}, which
 * assigns a standard deviation dependent on mass of the particle.
 * <p/>
 * These are the invariants: <ul><li>No particles collides with another. <li>No
 * particle touches a box wall or even penetrates it.</ul>
 *
 * @throws IllegalArgumentException if there are too many or too big particles
 * (the volume ratio was exceeded or a particle couldn't be placed without
 * intersecting another particle).
 */
public void distributeVelocity(
    Particle[] particles, Simulation simulation, float stdDev
) {
    Random random = simulation.getExperiment().getRandom();

    for (Particle particle : particles) {
        float[] velocity = particle.velocity;

        // random velocity distributed gaussian around 0 with stdDev
        velocity[X] = (float)random.nextGaussian() * stdDev;
        velocity[Y] = (float)random.nextGaussian() * stdDev;
        velocity[Z] = (float)random.nextGaussian() * stdDev;
    }
}

}
