/*
 *                        s a c e s
 *      Simple Artificial Chemistry Experiment System
 *
 * A semester project Summer 2005 and a diploma project 2006
 *           Berne University of Applied Sciences
 *   HTI School of Engineering and Information Technology
 *             Anthony Aguillon, Daniel Noelpp
 *
 */

package saces.pnp;

import static saces.Util.*;
import saces.exp.ParticleClass;
import saces.exp.Reaction;
import static saces.exp.Reaction.ReactionType.Merge;
import saces.sim.Collision;
import saces.sim.Particle;
import saces.sim.Simulation;

import java.util.Random;
import java.util.Set;


/**
 * This merger ignores momentum preservation and accelerates or decelerates the
 * merged particle. It assumes that the momentum violations of the mergers
 * cancel out each other.
 */
public class MergerEnergyConservation implements Merger {

// Warning: some duplicate code with transform() and decay()
public void merge(Collision[] collisions, Simulation simulation) {
    Random random = simulation.getExperiment().getRandom();
    int count = 0;

    for (Collision collision : collisions) {
        Particle e1 = collision.educt1;
        Particle e2 = collision.educt2;
        assert e1 != e2; // particle doesn't collide with itself

        if (collision.alreadyMerged) continue;
        assert simulation.contains(e1);
        assert simulation.contains(e2);

        final float kineticEnergy
            = (e1.mass * e1.speedSq + e2.mass * e2.speedSq) / 2;
        final ParticleClass pc1 = e1.particleClass;
        final ParticleClass pc2 = e2.particleClass;

        // Intersection of all merge reactions of both particles
        Set<Reaction> reactions = simulation.getExperiment()
            .findReactions(Merge, pc1, pc2);

        // To avoid the use of the random number generator if there are
        // no reactions.
        if (reactions.isEmpty()) continue;
        float probability = random.nextFloat();

        for (Reaction reaction : reactions) {
            // Verify whether we really have the right reaction type
            assert reaction.getReactionType() == Merge;

            // Verify whether we really have the right educts (classes)
            if (ASSERTS_ENABLED) {
                ParticleClass eductClass1 = reaction.getEduct1();
                ParticleClass eductClass2 = reaction.getEduct2();
                assert (pc1 == eductClass1 && pc2 == eductClass2)
                    || (pc2 == eductClass1 && pc1 == eductClass2);
            }

            // Try one merge reaction after another till the probability
            // range is hit.
            if (probability > reaction.getProbability()) {
                probability -= reaction.getProbability();
                continue;
            }

            // Then give up if kinetic energy of collision is not enough
            if (kineticEnergy < reaction.getActivationEnergy()) break;

            // And give up if there is not enough energy for negative diff.
            final float newEnergy = kineticEnergy + reaction.getEnergyDifference();
            if (newEnergy <= 0) break;

            // All is well, now do a merge reaction
            Particle p = simulation.createParticle(reaction.getProduct1());
            simulation.removeParticles(e1, e2);
            simulation.addParticle(p);

            // Assign velocity and position to products
            p.position[X] = (e1.position[X] + e2.position[X]) / 2f;
            p.position[Y] = (e1.position[Y] + e2.position[Y]) / 2f;
            p.position[Z] = (e1.position[Z] + e2.position[Z]) / 2f;

            float m1 = e1.mass;
            float m2 = e2.mass;
            float m3 = p.mass;

            float[] v = p.velocity;
            v[X] = (m1 * e1.velocity[X] + m2 * e2.velocity[X]) / m3;
            v[Y] = (m1 * e1.velocity[Y] + m2 * e2.velocity[Y]) / m3;
            v[Z] = (m1 * e1.velocity[Z] + m2 * e2.velocity[Z]) / m3;

            float momentumSpeedSq = v[X] * v[X] + v[Y] * v[Y] + v[Z] * v[Z];
            float energySpeedSq = 2 * newEnergy / p.mass;
            float factor = (float)Math.sqrt(energySpeedSq / momentumSpeedSq);

            v[X] *= factor;
            v[Y] *= factor;
            v[Z] *= factor;

            if (ASSERTS_ENABLED) {
                float energy = p.mass
                    * (v[X] * v[X] + v[Y] * v[Y] + v[Z] * v[Z]) / 2;
                assert eq(newEnergy, energy);
            }

            // Assign properties to collision so that other steps
            // don't consider this collision object
            collision.product1 = null;
            collision.product2 = null;
            collision.alreadyMerged = true;

            count++;
            break;
            // Don't handle any more merge reactions for this collision
        }
    }
    int oldCount = (Integer)simulation.getProperty(MERGE_COUNT);
    simulation.putProperty(MERGE_COUNT, oldCount + count);
}

}

