package saces.pnp;

import static saces.Util.*;
import saces.sim.Particle;
import saces.sim.Collision;
import saces.sim.Simulation;

import java.util.Random;


/**
 * A collision response where the particles tend to stick to each other. This
 * class has a funny history: After a talk with Dr P. Schwab this class has been
 * implemented as ResponseSchwab but the first run showed the bug that particles
 * often stick to each other. This is nice to look at and sometimes hilarously
 * funny when these groups of particles sticking to each other collide with the
 * walls or with other particles or groups &mdash; the particles swing or are
 * something like digested into the biggest particle and disappear inside it. We
 * decided to keep this bug as a funny feature of sAcEs.
 * <p/>
 * There were two bugs: a minor glitch (forgot a minus sign, see comment in
 * source code with (*)) so that the particles weren't moving away from each
 * other in the center of mass referenc model and a secondly, more serious one:
 * At the time of collision detection the particles were already overlapping
 * (sometimes very much when they were moving fast and collision detection
 * kicked in late). In principle it is possible to ignore this effect and say
 * that particles are made of an elastic material. But what if one of the
 * particles collides with another particle even before it left the other
 * particle? So we must calculate the time back to the exact point in time when
 * the two particles touched each other and calculate forward to the original
 * time.
 * <p/>
 * The "bugs" are kept as "features". This class is kept simple, too, and
 * doesn't contain optimizations (like pre-created vector arrays).
 */
public class ResponseFunnySticky implements Response {

public void response(Collision[] collisions, Simulation simulation) {
    Random random = simulation.getExperiment().getRandom();
    for (Collision collision : collisions) {
        if (collision.alreadyMerged) continue;

        Particle p1 = collision.educt1;
        Particle p2 = collision.educt2;

        // don't do a second collision response if already done
        if (p1.get(RESPONSE_DONE_WITH) == p2) continue;
        p1.put(RESPONSE_DONE_WITH, p2);
        p2.put(RESPONSE_DONE_WITH, p1);

        // handle mass
        float m1 = p1.mass;
        float m2 = p2.mass;
        float m = m1 + m2;

        // handle velocity
        float[] v1 = p1.velocity;
        float[] v2 = p2.velocity;
        float v1X = v1[X];
        float v1Y = v1[Y];
        float v1Z = v1[Z];
        float v2X = v2[X];
        float v2Y = v2[Y];
        float v2Z = v2[Z];

        // velocity of mass center
        float vcmX = (m1 * v1X + m2 * v2X) / m;
        float vcmY = (m1 * v1Y + m2 * v2Y) / m;
        float vcmZ = (m1 * v1Z + m2 * v2Z) / m;

        // translate velocities to mass center reference frame
        float uX = v1X - vcmX;
        float uY = v1Y - vcmY;
        float uZ = v1Z - vcmZ;
        float wX = v2X - vcmX;
        float wY = v2Y - vcmY;
        float wZ = v2Z - vcmZ;
        float u = (float)Math.sqrt(uX * uX + uY * uY + uZ * uZ);
        float w = +(float)Math.sqrt(wX * wX + wY * wY + wZ * wZ);
        // here --^-- the plus sign is the only difference to
        // ResponseSchwab (instead of minus sign)

        // get random vector
        float nX = 2 * random.nextFloat() - 1;
        float nY = 2 * random.nextFloat() - 1;
        float nZ = 2 * random.nextFloat() - 1;
        float n1 = (float)Math.sqrt(nX * nX + nY * nY + nZ * nZ);
        nX /= n1;
        nY /= n1;
        nZ /= n1;

        // calculate response velocities and save them
        float[] pv1 = collision.product1.velocity;
        float[] pv2 = collision.product2.velocity;
        pv1[X] = vcmX + u * nX;
        pv1[Y] = vcmY + u * nY;
        pv1[Z] = vcmZ + u * nZ;
        pv2[X] = vcmX + w * nX;
        pv2[Y] = vcmY + w * nY;
        pv2[Z] = vcmZ + w * nZ;
    }
}

}
