/*
 *                        s a c e s
 *      Simple Artificial Chemistry Experiment System
 *
 * A semester project Summer 2005 and a diploma project 2006
 *           Berne University of Applied Sciences
 *   HTI School of Engineering and Information Technology
 *             Anthony Aguillon, Daniel Noelpp
 *
 */
package saces.pnp;

import static saces.Util.*;
import saces.sim.Collision;
import saces.sim.Particle;
import saces.sim.Simulation;

import java.util.Random;


/**
 * An efficient Monte-Carlo collision response given by Dr Peter Schwab, our
 * diploma adivsor. It uses as reference frame the mass center of two particles
 * and uses an equal distribution of the reflection angle.
 * <p>
 * It is assumed, however, that bound energy difference is zero and that all
 * educt and product masses are the same.
 */
public class ResponseSchwab implements Response {

public void response(Collision[] collisions, Simulation simulation) {
    float momentumX = 0, momentumY = 0, momentumZ = 0;
    Random random = simulation.getExperiment().getRandom();
    for (Collision collision : collisions) {
        if (collision.alreadyMerged) continue;

        Particle p1 = collision.educt1;
        Particle p2 = collision.educt2;

        // don't do a second collision response if already done
        if (p1.get(RESPONSE_DONE_WITH) == p2) continue;
        p1.put(RESPONSE_DONE_WITH, p2);
        p2.put(RESPONSE_DONE_WITH, p1);

        // handle mass
        float m1 = p1.mass;
        float m2 = p2.mass;
        float m = m1 + m2;

        // handle velocity
        float[] v1 = p1.velocity;
        float[] v2 = p2.velocity;
        float v1X = v1[X];
        float v1Y = v1[Y];
        float v1Z = v1[Z];
        float v2X = v2[X];
        float v2Y = v2[Y];
        float v2Z = v2[Z];

        // momentum BEFORE collision (to assert momentum preservation)
        if (ASSERTS_ENABLED && ASSERT_ENERGY) {
            momentumX = m1 * v1[X] + m2 * v2[X];
            momentumY = m1 * v1[Y] + m2 * v2[Y];
            momentumZ = m1 * v1[Z] + m2 * v2[Z];
        }

        // velocity of mass center
        float vcmX = (m1 * v1X + m2 * v2X) / m;
        float vcmY = (m1 * v1Y + m2 * v2Y) / m;
        float vcmZ = (m1 * v1Z + m2 * v2Z) / m;

        // translate velocities to mass center reference frame
        float uX = v1X - vcmX;
        float uY = v1Y - vcmY;
        float uZ = v1Z - vcmZ;
        float wX = v2X - vcmX;
        float wY = v2Y - vcmY;
        float wZ = v2Z - vcmZ;

        float u = (float)Math.sqrt(uX * uX + uY * uY + uZ * uZ);
        float w = -(float)Math.sqrt(wX * wX + wY * wY + wZ * wZ);

        // in mass center reference frame the momenta are equal
        assert eq(m1 * u, -m2 * w);

        // get random vector
        float nX = 2 * random.nextFloat() - 1;
        float nY = 2 * random.nextFloat() - 1;
        float nZ = 2 * random.nextFloat() - 1;
        float n1 = (float)Math.sqrt(nX * nX + nY * nY + nZ * nZ);
        nX /= n1;
        nY /= n1;
        nZ /= n1;

        // calculate response velocities and save them
        float[] pv1 = collision.product1.velocity;
        float[] pv2 = collision.product2.velocity;
        pv1[X] = vcmX + u * nX;
        pv1[Y] = vcmY + u * nY;
        pv1[Z] = vcmZ + u * nZ;
        pv2[X] = vcmX + w * nX;
        pv2[Y] = vcmY + w * nY;
        pv2[Z] = vcmZ + w * nZ;

        // assert momentum preservation (assuming no mass changes and energyDiff)
        if (ASSERTS_ENABLED && ASSERT_ENERGY) {
            assert eq(momentumX, m1 * pv1[X] + m2 * pv2[X]);
            assert eq(momentumY, m1 * pv1[Y] + m2 * pv2[Y]);
            assert eq(momentumZ, m1 * pv1[Z] + m2 * pv2[Z]);
        }
    }
}

}
