/*
 *                        s a c e s
 *      Simple Artificial Chemistry Experiment System
 *
 * A semester project Summer 2005 and a diploma project 2006
 *           Berne University of Applied Sciences
 *   HTI School of Engineering and Information Technology
 *             Anthony Aguillon, Daniel Noelpp
 *
 */
package saces.pnp;

import static saces.Util.*;
import saces.sim.Collision;
import saces.sim.Particle;
import saces.sim.Simulation;

import java.util.Random;
import java.util.Arrays;


/**
 * An efficient Monte-Carlo collision response given by Dr Peter Schwab, our
 * diploma adivsor. It uses as reference frame the mass center of two particles
 * and uses an equal distribution of the reflection angle.
 * <p/>
 */
public class ResponseSchwab2 implements Response {

public void response(Collision[] collisions, Simulation simulation) {
    double momentumX = 0, momentumY = 0, momentumZ = 0, kinEnergy = 0;
    double totalEnergyInCollisions = 0;

    if (ASSERTS_ENABLED) {
        for (Collision collision : collisions) {
            Particle e1 = collision.educt1;
            Particle e2 = collision.educt2;
            float[] v1 = e1.velocity;
            float[] v2 = e2.velocity;
            double v1Sq = v1[X] * v1[X] + v1[Y] * v1[Y] + v1[Z] * v1[Z];
            double v2Sq = v2[X] * v2[X] + v2[Y] * v2[Y] + v2[Z] * v2[Z];
            totalEnergyInCollisions = totalEnergyInCollisions
                + e1.mass * v1Sq / 2 + e1.particleClass.getEnergy()
                + e2.mass * v2Sq / 2 + e2.particleClass.getEnergy();
        }
    }

    Random random = simulation.getExperiment().getRandom();
    for (Collision collision : collisions) {
        if (collision.alreadyMerged) continue;

        final Particle p1 = collision.educt1;
        final Particle p2 = collision.educt2;
        final Particle p3 = collision.product1;
        final Particle p4 = collision.product2;
        final double dH = collision.energyDiff;

        // handle mass
        final double m1 = p1.mass;
        final double m2 = p2.mass;
        final double m3 = p3.mass;
        final double m4 = p4.mass;

        // handle velocity
        final float[] v1 = p1.velocity;
        final float[] v2 = p2.velocity;
        final float[] v3 = p3.velocity;
        final float[] v4 = p4.velocity;

        // open block because many local variables
        {
            final double v1X = v1[X];
            final double v1Y = v1[Y];
            final double v1Z = v1[Z];
            final double v2X = v2[X];
            final double v2Y = v2[Y];
            final double v2Z = v2[Z];

            // momentum BEFORE collision (to assert momentum preservation)
            // energy BEFORE collision (to assert energy conservation)
            if (ASSERTS_ENABLED) {
                momentumX = m1 * v1X + m2 * v2X;
                momentumY = m1 * v1Y + m2 * v2Y;
                momentumZ = m1 * v1Z + m2 * v2Z;
                kinEnergy = m1 * p1.speedSq / 2 + m2 * p2.speedSq / 2;

                // assert speedSq
                double speedSq1 = v1X * v1X + v1Y * v1Y + v1Z * v1Z;
                double speedSq2 = v2X * v2X + v2Y * v2Y + v2Z * v2Z;
                assert eq(p1.speedSq, speedSq1);
                assert eq(p2.speedSq, speedSq2);

                // assert bound energy difference
                float e1 = p1.storedEnergy;
                float e2 = p2.storedEnergy;
                float e3 = p3.storedEnergy;
                float e4 = p4.storedEnergy;
                assert eq(dH, e1 + e2 - e3 - e4);
            }

            // velocity of mass center
            final double m = m1 + m2;
            final double vcmX = (m1 * v1X + m2 * v2X) / m;
            final double vcmY = (m1 * v1Y + m2 * v2Y) / m;
            final double vcmZ = (m1 * v1Z + m2 * v2Z) / m;

            // translate velocities to mass center reference frame
            final double uX = v1X - vcmX;
            final double uY = v1Y - vcmY;
            final double uZ = v1Z - vcmZ;
            final double wX = v2X - vcmX;
            final double wY = v2Y - vcmY;
            final double wZ = v2Z - vcmZ;
            final double uSq = uX * uX + uY * uY + uZ * uZ;
            final double wSq = wX * wX + wY * wY + wZ * wZ;

            // consider bound energy difference and mass changes
            final double massRatio = m3 * m4 / (m3 + m4);
            final double f1 = uSq * m * m1 / m2;
            final double f2 = wSq * m * m2 / m1;

            // calculate the new speeds speed3 and speed4
            assert f1 >= -2 * dH;
            assert f2 >= -2 * dH;
            final double speed3 = Math.sqrt(massRatio * (f1 + 2 * dH)) / m3;
            final double speed4 = -Math.sqrt(massRatio * (f2 + 2 * dH)) / m4;

            // are the momenta after collision equal?
            assert eq(m3 * speed3, -m4 * speed4);

            // get random vector
            double nX = 2 * random.nextDouble() - 1;
            double nY = 2 * random.nextDouble() - 1;
            double nZ = 2 * random.nextDouble() - 1;
            final double n = Math.sqrt(nX * nX + nY * nY + nZ * nZ);
            nX /= n;
            nY /= n;
            nZ /= n;

            // calculate response velocities and save them
            v3[X] = (float)(vcmX + speed3 * nX);
            v3[Y] = (float)(vcmY + speed3 * nY);
            v3[Z] = (float)(vcmZ + speed3 * nZ);
            v4[X] = (float)(vcmX + speed4 * nX);
            v4[Y] = (float)(vcmY + speed4 * nY);
            v4[Z] = (float)(vcmZ + speed4 * nZ);
        }

        // calculate speed squares
        p3.speedSq = v3[X] * v3[X] + v3[Y] * v3[Y] + v3[Z] * v3[Z];
        p4.speedSq = v4[X] * v4[X] + v4[Y] * v4[Y] + v4[Z] * v4[Z];

        // assert momentum preservation (assuming no mass changes and en.diff.)
        if (ASSERTS_ENABLED) {
            assert eq(momentumX, m3 * v3[X] + m4 * v4[X]);
            assert eq(momentumY, m3 * v3[Y] + m4 * v4[Y]);
            assert eq(momentumZ, m3 * v3[Z] + m4 * v4[Z]);
            assert eq(kinEnergy + dH,
                m3 * p3.speedSq / 2 + m4 * p4.speedSq / 2);

        }
    }

    if (ASSERTS_ENABLED) {
        // Total energy of educts should be same as products
        double totalEnergyInCollisionsAfter = 0;
        for (Collision collision : collisions) {
            if (collision.alreadyMerged) continue;

            Particle p1 = collision.product1;
            Particle p2 = collision.product2;
            float[] v1 = p1.velocity;
            float[] v2 = p2.velocity;
            double v1Sq = (v1[X] * v1[X] + v1[Y] * v1[Y] + v1[Z] * v1[Z]) / 2;
            double v2Sq = (v2[X] * v2[X] + v2[Y] * v2[Y] + v2[Z] * v2[Z]) / 2;
            totalEnergyInCollisionsAfter = totalEnergyInCollisionsAfter
                + p1.mass * v1Sq + p1.particleClass.getEnergy()
                + p2.mass * v2Sq + p2.particleClass.getEnergy();
        }

        if (!eq(totalEnergyInCollisions, totalEnergyInCollisionsAfter)
            && count < 10) {
            double a = totalEnergyInCollisionsAfter - totalEnergyInCollisions;
            double b = totalEnergyInCollisionsAfter + totalEnergyInCollisions;
            double percent = 200f * a / b;
            System.out.print("at " + new Throwable().getStackTrace()[0]
                + ": totalEnergyInCollisions=");
            System.out.printf("% 8.2f (%+7.2f%%),%ncollisions=%s%n",
                totalEnergyInCollisions, percent,
                Arrays.asList(collisions).toString());
            count++;
        }
    }
}

private static int count = 0;

}