/*
 *                        s a c e s
 *      Simple Artificial Chemistry Experiment System
 *
 * A semester project Summer 2005 and a diploma project 2006
 *           Berne University of Applied Sciences
 *   HTI School of Engineering and Information Technology
 *             Anthony Aguillon, Daniel Noelpp
 *
 */
package saces.pnp;

import saces.Util;
import static saces.Util.*;
import saces.exp.ParticleClass;
import saces.exp.Reaction;
import static saces.exp.Reaction.ReactionType.Transform;
import saces.sim.Collision;
import saces.sim.Particle;
import saces.sim.Simulation;

import java.util.Random;
import java.util.Set;

/** Does a sensible minimum for handling transform reactions:
 * Iterates through all collision pairs and looks for each
 * particle pair whether they have common transform reactions
 * as an educt. Then probability, activation and collision
 * energy is verified.
 */
public class TransformerSimple implements Transformer {

public void transform(Collision[] collisions, Simulation simulation) {
    Random random = simulation.getExperiment().getRandom();
    int count = 0;

    int particleCount = 0;
    if (ASSERTS_ENABLED) particleCount = simulation.getTotalParticleCount();

    for (Collision collision : collisions) {
        Particle e1 = collision.educt1;
        Particle e2 = collision.educt2;
        assert e1 != e2; // particle doesn't collide with itself

        if (collision.alreadyMerged) continue;

        // Sadly these assertions are triggered (rarely)
        //assert simulation.contains(e1);
        //assert simulation.contains(e2);
        if (!simulation.contains(e1)) simulation.missing(e1);
        if (!simulation.contains(e2)) simulation.missing(e2);
        if (!simulation.contains(e1) || !simulation.contains(e2))
            break;

        final float kineticEnergy
            = (e1.mass * e1.speedSq + e2.mass * e2.speedSq) / 2;
        final ParticleClass pc1 = e1.particleClass;
        final ParticleClass pc2 = e2.particleClass;

        // Intersection of all transform reactions of both particles
        Set<Reaction> reactions = simulation.getExperiment()
            .findReactions(Transform, pc1, pc2);

        // To avoid the use of the random number generator if there are
        // no reactions.
        if (reactions.isEmpty()) continue;
        float probability = random.nextFloat();

        for (Reaction reaction : reactions) {
            // Verify whether we really have the right reaction type
            assert reaction.getReactionType() == Transform;

            // Verify whether we really have the right educts (classes)
            if (ASSERTS_ENABLED) {
                ParticleClass eductClass1 = reaction.getEduct1();
                ParticleClass eductClass2 = reaction.getEduct2();
                assert (pc1 == eductClass1 && pc2 == eductClass2)
                    || (pc2 == eductClass1 && pc1 == eductClass2);
            }

            // Try one transform reaction after another till the probability
            // range is hit.
            if (probability > reaction.getProbability()) {
                probability -= reaction.getProbability();
                continue;
            }

            // Then give up if kinetic energy of collision is not enough
            if (kineticEnergy < reaction.getActivationEnergy()) break;


            // Then give up if collision energy is not enough
            float dE = reaction.getEnergyDifference();
            if (collision.getCollisionEnergy() < -2 * dE) break;

            // All is well, now do a transform reaction
            Particle p1 = simulation.createParticle(reaction.getProduct1());
            Particle p2 = simulation.createParticle(reaction.getProduct2());
            simulation.removeParticles(e1, e2);
            simulation.addParticles(p1, p2);

            // Assign educt velocity and position to products
            p1.position[X] = e1.position[X];
            p1.position[Y] = e1.position[Y];
            p1.position[Z] = e1.position[Z];
            p1.velocity[X] = e1.velocity[X];
            p1.velocity[Y] = e1.velocity[Y];
            p1.velocity[Z] = e1.velocity[Z];
            p2.position[X] = e2.position[X];
            p2.position[Y] = e2.position[Y];
            p2.position[Z] = e2.position[Z];
            p2.velocity[X] = e2.velocity[X];
            p2.velocity[Y] = e2.velocity[Y];
            p2.velocity[Z] = e2.velocity[Z];

            // Assign properties to collision for response()
            collision.product1 = p1;
            collision.product2 = p2;

            collision.energyDiff = dE;

            count++;
            break;
            // Don't handle any more transform reactions for this collision
        }
    }
    int oldCount = (Integer)simulation.getProperty(TRANSFORM_COUNT);
    simulation.putProperty(TRANSFORM_COUNT, oldCount + count);

// Transform reactions don't change the particle count
    if (Util.ASSERTS_ENABLED)
        assert particleCount == simulation.getTotalParticleCount();
}

}
