/*
*                        s a c e s
*      Simple Artificial Chemistry Experiment System
*
* A semester project Summer 2005 and a diploma project 2006
*           Berne University of Applied Sciences
*   HTI School of Engineering and Information Technology
*             Anthony Aguillon, Daniel Noelpp
*
*/
package saces.sim;


import net.java.games.jogl.GL;
import net.java.games.jogl.GLDrawable;
import net.java.games.jogl.GLEventListener;
import saces.Util;
import saces.exp.Experiment;
import saces.exp.ParticleClass;
import saces.gl.Box;
import saces.gl.DisplayList;
import saces.gl.DisplayListElement;
import saces.gl.Drawable;
import saces.gl.Routines;
import saces.gl.Sphere;
import saces.gl.View;
import saces.pnp.Decayer;
import saces.pnp.Detector;
import saces.pnp.Distributor;
import saces.pnp.Merger;
import saces.pnp.Reflector;
import saces.pnp.Response;
import saces.pnp.Transformer;

import java.awt.Component;
import java.util.Random;
import java.util.Set;
import java.util.Locale;


/**
 * The mediator mediates between gl and the simulation logic. This class is the
 * only class in package simulation which uses saces.gl classes. All other
 * classes use OpenGL indirectly through an instance of this class.
 */
public class Mediator implements GLEventListener {

// simulation data
private Simulation simulation;

// pnp interfaces
private Reflector reflector;
private Detector detector;
private Merger merger;
private Transformer transformer;
private Response response;
private Decayer decayer;
private Positioner positioner;

private boolean initialRun;
// do not draw when an experiment is set
private boolean doNotUpdate;

private Box box;
private Routines routines;

private boolean viewChanged;
private boolean detailLevelChanged;
private boolean clearOnNextIteration;

// must be set to true if a new experiment has been set
private boolean initializeExperiment = true;

private int screenWidth;
private int screenHeight;

private Random random;
private DisplayList displayList;

public static final int DEFAULT_PERSPECTIVE = View.DEFAULT_PERSPECTIVE;
public static final int ORTHOGRAPHIC = View.ORTHOGRAPHIC;
public static final int PARTICLE_PERSPECTIVE = View.PARTICLE_PERSPECTIVE;

public static final int DETAIL_LEVEL_LOW = Routines.DETAIL_LEVEL_LOW;
public static final int DETAIL_LEVEL_MEDIUM = Routines.DETAIL_LEVEL_MEDIUM;
public static final int DETAIL_LEVEL_HIGH = Routines.DETAIL_LEVEL_HIGH;

public static final float SIMULATION_MAX_SLOWING_FACTOR =
    Simulation.SIMULATION_MAX_SLOWING_FACTOR;
public static final float SIMULATION_MIN_SLOWING_FACTOR =
    Simulation.SIMULATION_MIN_SLOWING_FACTOR;

public Mediator() {

    simulation = new Simulation();
    positioner = new Positioner();
    displayList = new DisplayList();
    routines = new Routines(Routines.DETAIL_LEVEL_MEDIUM);
    random = new Random();
    viewChanged = false;
    detailLevelChanged = true;
    clearOnNextIteration = false;
    initialRun = true;

    routines.getCanvas().addGLEventListener(this);
}

/** Simulation getter. */
public Simulation getSimulation() {
    return simulation;
}

/** Get the component where OpenGL drawing is happening. */
public Component getGLComponent() {
    return routines.getCanvas();
}

public void init(GLDrawable gLDrawable) {
    routines.init(gLDrawable);
}

/**
 * Set a detail level (of painting the simulation)
 *
 * @param detailLevel : int DETAIL_LEVEL_LOW | DETAIL_LEVEL_MEDIUM |
 * DETAIL_LEVEL_HIGH
 */
public void setDetailLevel(int detailLevel) {
    routines.setDetailLevel(detailLevel);
    detailLevelChanged = true;
}

/** Clear the experiment */
public synchronized void clearExperiment() {
    clearOnNextIteration = true;
    initialRun = true;
}

public void display(GLDrawable glDrawable) {

    if(doNotUpdate) return;

    GL gl = glDrawable.getGL();

    // Attention: Might be a bottleneck.
    if (initializeExperiment) {
        displayList.add(glDrawable, box);
        displayList.initializeParticleClassLists(glDrawable,
            getExperiment().getParticleClasses());
    }

    // pseudo queuing
    if (clearOnNextIteration) {
        simulation.clear();
        init(glDrawable);

        if (gl.glIsEnabled(GL.GL_LIGHTING)) {
            routines.light.draw(gl);
        }

        clearOnNextIteration = false;
        return;
    }

    // If simulation is halted, don't calculate the simulation loop.
    if (simulation.isRunning() && routines.isAnimatorRunning) {
        Particle[] particles;
        particles = simulation.getParticles();

        positioner.position(particles, simulation);
        simulation.updateTime();
        reflector.reflect(particles, simulation);
        simulation.measure();

        Collision[] collisions = detector.detect(particles, simulation);
        merger.merge(collisions, simulation);
        transformer.transform(collisions, simulation);
        response.response(collisions, simulation);

        // Some particles have been removed, other added
        particles = simulation.getParticles();
        decayer.decay(particles, simulation);
    }

    // Detail level has been changed by the user.
    if (detailLevelChanged) {
        routines.applyDetailLevelChanges(gl);
        detailLevelChanged = false;
    }

    // View mode has been changed by the user.
    if (viewChanged) {
        routines.applyViewChange(glDrawable, screenWidth, screenHeight);
        viewChanged = false;
    }

    // Paint display data: running time, number of particles, speed, etc.
    int fps = simulation.getFrameRate();
    if (fps > 0) {
        routines.processDataPanel.setFPS(String.valueOf(
            simulation.getFrameRate()));
    }
    // Attention: might be a bottleneck. String.format() is expensive.
    routines.processDataPanel.setRunningtime(String.format(
        "%1$tH:%1$tM:%1$tS:%1$tL", simulation.getRunningTime()));
    routines.processDataPanel.setParticleCount(
        String.valueOf(simulation.getTotalParticleCount()));
    routines.processDataPanel.setSpeedPercentage(
        String.valueOf((int)(simulation.getSlowingFactor() * 100.0f)) + "%");

    // Orthographic view mode with rendering three times.
    Set<Particle> particleSet = simulation.getParticleSet();
    if (routines.view.getViewMode() == View.ORTHOGRAPHIC) {

        gl.glViewport(0, 0, screenWidth, screenHeight);
        gl.glScissor(0, 0, screenWidth, screenHeight);

        gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
        routines.applyOrthoViewportSetup(glDrawable, View.VIEWPORT_ID_TOP_LEFT);

        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glPushMatrix();
        gl.glLoadIdentity();

        drawScene(glDrawable, particleSet);
        routines.drawPanels(glDrawable, View.VIEWPORT_ID_TOP_LEFT);

        gl.glPopMatrix();

        routines.applyOrthoViewportSetup(glDrawable,
            View.VIEWPORT_ID_TOP_RIGHT);

        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glPushMatrix();
        gl.glLoadIdentity();
        gl.glRotatef(90f, 0, 1, 0);
        drawScene(glDrawable, particleSet);
        routines.drawPanels(glDrawable, View.VIEWPORT_ID_TOP_RIGHT);
        gl.glPopMatrix();

        routines.applyOrthoViewportSetup(glDrawable,
            View.VIEWPORT_ID_BOTTOM_LEFT);

        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glPushMatrix();
        gl.glLoadIdentity();
        gl.glRotatef(-90f, 1, 0, 0);
        drawScene(glDrawable, particleSet);
        routines.drawPanels(glDrawable, View.VIEWPORT_ID_BOTTOM_LEFT);
        gl.glPopMatrix();

        routines.applyOrthoViewportSetup(glDrawable,
            View.VIEWPORT_ID_BOTTOM_RIGHT);

        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glPushMatrix();
        gl.glLoadIdentity();
        routines.processDataPanel.draw(glDrawable, true);
        gl.glPopMatrix();
    }

    // Perspective viewing or particle viewing mode
    else {
        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glPushMatrix();

        routines.applyPerspectiveViewportSetup(glDrawable);

        drawScene(glDrawable, particleSet);
        routines.processDataPanel.draw(glDrawable, false);
        gl.glPopMatrix();
    }

    gl.glFlush();

    // The animation thread does not allow the swing thread to
    // react properly to user triggered events. The sleep function
    // works around this problem
    try{ Thread.sleep(5); } catch(InterruptedException e){}
}

// Paint the scene
private void drawScene(GLDrawable glDrawable, Set<Particle> particles) {

    GL gl = glDrawable.getGL();

    if (routines.detailLevel == Routines.DETAIL_LEVEL_HIGH) {

        int viewMode = routines.view.getViewMode();

        if (viewMode == View.DEFAULT_PERSPECTIVE ||
            viewMode == View.PARTICLE_PERSPECTIVE) {

            if (viewMode == View.PARTICLE_PERSPECTIVE) {
                routines.view.updateCameraPosition();
            }

            gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT |
                GL.GL_STENCIL_BUFFER_BIT);

            routines.light.draw(gl);
            drawParticles(glDrawable, particles);

            gl.glDisable(GL.GL_DEPTH_TEST);
            gl.glColorMask(false, false, false, false);

            // Draw 1 into the stencil buffer.
            gl.glEnable(GL.GL_STENCIL_TEST);
            gl.glStencilOp(GL.GL_REPLACE, GL.GL_REPLACE, GL.GL_REPLACE);
            gl.glStencilFunc(GL.GL_ALWAYS, 1, 0xffffffff);

            box.drawReflecting(gl, glDrawable.getGLU());

            // Re-enable update of color and depth.
            gl.glColorMask(true, true, true, true);
            gl.glEnable(GL.GL_DEPTH_TEST);
            gl.glDepthMask(false);

            // draw if stencil == 1
            gl.glStencilFunc(GL.GL_EQUAL, 1, 0xffffffff);
            gl.glStencilOp(GL.GL_KEEP, GL.GL_KEEP, GL.GL_KEEP);

            // scaling messes up the normals, this is a precausion
            gl.glEnable(GL.GL_NORMALIZE);

            float[][] sides =
                {{1.0f, 1.0f, -1.0f}, {-1.0f, 1.0f, 1.0f},
                    {1.0f, -1.0f, 1.0f}, {-1.0f, -1.0f, 1.0f},
                    {1.0f, 1.0f, -1.0f}, {-1.0f, 1.0f, -1.0f},
                    {-1.0f, -1.0f, -1.0f}};

            for (int i = 0; i < 3; i++) {
                gl.glPushMatrix();
                for (Particle particle : particles) {
                    gl.glPushMatrix();
                    gl.glScalef(sides[i][0], sides[i][1], sides[i][2]);
                    routines.light.draw(gl);
                    ((Drawable)particle).transform(glDrawable);
                    displayList.call(gl, (DisplayListElement)particle);
                    gl.glPopMatrix();
                }
                gl.glPopMatrix();
            }

            gl.glDisable(GL.GL_STENCIL_TEST);
            gl.glDisable(GL.GL_NORMALIZE);

            gl.glEnable(GL.GL_BLEND);
            gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);

            routines.light.draw(gl);
            box.drawReflecting(gl, glDrawable.getGLU());
            box.drawOutline(gl, glDrawable.getGLU());

            gl.glDisable(GL.GL_BLEND);

            gl.glDepthMask(true);
            gl.glDisable(GL.GL_DEPTH_TEST);
            gl.glColorMask(false, false, false, false);

            // Draw 1 into the stencil buffer.
            gl.glEnable(GL.GL_STENCIL_TEST);
            gl.glStencilOp(GL.GL_REPLACE, GL.GL_REPLACE, GL.GL_REPLACE);
            gl.glStencilFunc(GL.GL_ALWAYS, 1, 0xffffffff);

            routines.light.draw(gl);
            gl.glFrontFace(GL.GL_CW);
            box.drawReflecting(gl, glDrawable.getGLU());
            gl.glFrontFace(GL.GL_CW);

            // Re-enable update of color and depth.
            gl.glColorMask(true, true, true, true);

            // draw if stencil == 1
            gl.glStencilFunc(GL.GL_EQUAL, 1, 0xffffffff);
            gl.glStencilOp(GL.GL_KEEP, GL.GL_KEEP, GL.GL_KEEP);

            // scaling messes up the normals, this is a precausion
            gl.glEnable(GL.GL_NORMALIZE);
            gl.glDisable(GL.GL_STENCIL_TEST);
            gl.glDisable(GL.GL_NORMALIZE);
            gl.glEnable(GL.GL_BLEND);
            gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);

            gl.glFrontFace(GL.GL_CW);
            routines.light.draw(gl);
            box.drawReflecting(gl, glDrawable.getGLU());
            gl.glFrontFace(GL.GL_CCW);

            gl.glEnable(GL.GL_DEPTH_TEST);
            gl.glDisable(GL.GL_BLEND);

        } // Perspective or Particle View, else : Orthographic
        else {
            gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
            drawBox(gl);
            drawParticles(glDrawable, particles);
        }
    }

    else if (routines.detailLevel == Routines.DETAIL_LEVEL_MEDIUM) {
        gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
        gl.glEnable(GL.GL_DEPTH_TEST);

        drawBox(gl);
        drawParticles(glDrawable, particles);
    }

    else if (routines.detailLevel == Routines.DETAIL_LEVEL_LOW) {
        gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
        gl.glEnable(GL.GL_DEPTH_TEST);

        box.drawBox(gl);

        for (Particle particle : particles) {
            gl.glPushMatrix();
            ((Drawable)particle).transform(glDrawable);
            ((Sphere)particle).displayMethod(gl, glDrawable.getGLU());
            gl.glPopMatrix();
        }
    }
}

// Paint the reaction vessel walls
private void drawBox(GL gl) {

    gl.glPushMatrix();
    routines.light.draw(gl);
    displayList.call(gl, box);
    gl.glPopMatrix();

}

// Paint the particles
private void drawParticles(GLDrawable glDrawable, Set<Particle> particles) {

    GL gl = glDrawable.getGL();

    for (Particle particle : particles) {
        gl.glPushMatrix();
        routines.light.draw(gl);
        ((Drawable)particle).transform(glDrawable);
        displayList.call(gl, (DisplayListElement)particle);
        gl.glPopMatrix();
    }

}

/** GLEvent listener interface method implementation. */
public void reshape(
    GLDrawable glDrawable, int x, int y, int width, int height
) {

    routines.setupBoundingSphere(width, height);
    routines.reshape(glDrawable, x, y, width, height);

    this.screenWidth = width;
    this.screenHeight = height;

    viewChanged = true;
}


/** GLEvent listener interface method. */
public void displayChanged(
    GLDrawable gLDrawable, boolean modeChanged, boolean deviceChanged
) {}

/**
 * Increase simulation speed
 *
 * @return float : new simulation speed [0..1]
 */
public float increaseSpeed() {
    return simulation.increaseSpeed();
}

/**
 * Decrease simulation speed
 *
 * @return float : new simulation speed [0..1]
 */
public float decreaseSpeed() {
    return simulation.decreaseSpeed();
}

/**
 * Starts the animation thread. This is needed only once per session. Once the
 * animation thread is run, it will run until the application closes.
 *
 * @param start : true, start the animation thread
 */
private void startAnimationThread(boolean start) {
    routines.startAnimationThread(start);
}

/**
 * Start or stop the animation. Repeated calls are harmless.
 *
 * @param isRunning : run simulation, else stop
 */
public void setRunning(boolean isRunning) {

    if (isRunning) {
        simulation.updateTime();
        if (initialRun) {
            simulation.resetTime();
            initialRun = false;
        }
    }
    else
        simulation.stopTime();

    startAnimationThread(isRunning);
    routines.isAnimatorRunning = isRunning;

}

/**
 * Returns simulation running state
 *
 * @return true : simulation is running, else not.
 */
public boolean isRunning() {
    return simulation.isRunning();
}

/** Repaint gl */
public void repaint() {
//  routines.repaint();
}

/**
 * Set and initialize experiment: get plug-and-play simulation steps, create gl
 * data structures and particles and distribute them.
 *
 * @param experiment : Experiment
 */
public void setExperiment(Experiment experiment) {
    doNotUpdate = true;
    getSimulation().setExperiment(experiment);
    setRunning(false);

    box = new Box(experiment.getWidth(), experiment.getHeight(),
        experiment.getDepth());
    box.setColor(experiment.getColor());
    routines.setBox(box);

    // Get plug-and-play simulation steps
    // except Distributor because it is used only once
    reflector = simulation.createWallReflector();
    detector = simulation.createCollisionDetector();
    merger = simulation.createReactionMerger();
    transformer = simulation.createReactionTransformer();
    response = simulation.createCollisionResponse();
    decayer = simulation.createParticleDecayer();

    // Create these particles and add them to the simulation
    for (ParticleClass pc : experiment.getParticleClasses().values())
        for (int i = 0; i < pc.getInitialCount(); i++)
            simulation.addParticle(simulation.createParticle(pc));

    Distributor distributor = simulation.createParticleDistributor();
    try {
        Particle[] particles = simulation.getParticles();
        distributor.distribute(particles, simulation);
    }
    catch (Exception e) {
        StackTraceElement[] ste = e.getStackTrace();
        Util.showErrorMessage(
            "Problem distributing the particles", ""
            + e.getMessage() + ".\n"
            + "Please change the settings of the experiment.\n\n"
            + "-----------\n"
            + "Details: The particle distributor (of "
            + distributor.getClass() + ")\n"
            + "couldn't distribute the particles and threw:\n\n"
            + e + "\n" + "    at " + ste[0]
        );
        throw new SimulationPNPException(e);
    }

    simulation.getLogger().writeProperty("ExperimentInitialized",
        experiment.getDescription());

    doNotUpdate = false;
}

/**
 * Returns the experiment
 *
 * @return Experiment
 */
public Experiment getExperiment() {
    return simulation.getExperiment();
}

/**
 * Sets the viewing mode
 *
 * @param viewMode :  DEFAULT_PERSPECTIVE | ORTHOGRAPHIC | PARTICLE_PERSPECTIVE
 */
public void setViewMode(int viewMode) {
    // the null experiment doesn't have any particles
    if (simulation.getTotalParticleCount() == 0) return;

    if (viewMode == View.PARTICLE_PERSPECTIVE) {
        Particle[] particles = simulation.getParticles();

        routines.view.setParticle(particles[random.nextInt(particles.length)]);
    }
    routines.view.setViewMode(viewMode);
    viewChanged = true;
}

/**
 * Returns the viewing mode
 *
 * @return PARTICLE_PERSPECTIVE
 */
public int getViewMode() {

    if (routines.view != null) {
        return routines.view.getViewMode();
    }
    else return 0;
}

/**
 * Returns the speed factor.
 *
 * @return float : speed factor [0..1]
 */
public float getSpeedFactor() {
    return simulation.getSlowingFactor();
}

/**
 * Returns the detail level
 *
 * @return int : detail level
 */
public int getDetailLevel() {
    return routines.detailLevel;
}

}



