/*
*                        s a c e s
*      Simple Artificial Chemistry Experiment System
*
* A semester project Summer 2005 and a diploma project 2006
*           Berne University of Applied Sciences
*   HTI School of Engineering and Information Technology
*             Anthony Aguillon, Daniel Noelpp
*
*/
package saces.sim;

import static saces.Util.*;
import saces.exp.ParticleClass;

import java.awt.Color;
import java.util.HashMap;
import java.util.Map;


/**
 * A class representing the parameters of a particle in the simulation. It is
 * abstract. A class extending it should provide the implementation in form of
 * OpenGL graphics as does {@linkplain saces.gl.Sphere}.
 * <p/>
 * This class does not correspond to Java coding guidelines very well. There are
 * many public fields and some interdependency between them. This is neccessary
 * because of performance considerations. {@link #overlaps(Particle)}, for
 * example, is one of the busiest method of the whole application and must run
 * as efficient as possible.
 */
public abstract class Particle {

/** Partition particle is inside. Public field because of efficiency. */
public Partition partition;

/**
 * Absolute speed squared. Public field because of efficiency. It is calculated
 * by the update() method of the Positioner. Whenever someone changes the
 * velocity of the particle, speedSp must be updated, too.
 */
public float speedSq;

/** Particle class of particle. Public field because of efficiency. Final. */
public final ParticleClass particleClass;

/** Radius of particle. Same as particleClass.getRadius(). Final. */
public final float radius;

/** Mass of particle. Same as particleClass.getMass(). Final. */
public final float mass;

/** Position of particle. Public field because of efficiency. Final. */
public final float[] position = new float[3];

/** Velocity of particle. Public field because of efficiency. Final. */
public final float[] velocity = new float[3];

/**
 * Is initialized with the particle class energy and can be used independently
 * of mass and velocity. This is interpreted for particles which store energy
 * which is not translated into kinetic energy, i.e producer-consumer dynamics,
 * etc. It is not final, thus can be modified by reactions.
 * <p/>
 * An idea, how this would be useful is the scenario with merge reaction. The
 * energy excess of inelastic collision could be added to this field and be
 * converted to kinetic energy the next collision. This idea has not been
 * implemented in the current release of 'saces'.
 */
public float storedEnergy;

/**
 * Helper public field for collision detection. A marker object marks particle
 * alredy checked. Profiling has shown, that using the particle properties to
 * mark checked particles was too slow. This field is used by {@linkplain
 * saces.pnp.DetectorPartitionized}.
 */
public Object alreadyCollided;

private final Map<String, Object> properties = new HashMap<String, Object>();

/** Create a particle and initialize public final fields. */
public Particle(ParticleClass particleClass) {
    this.particleClass = particleClass;
    this.radius = particleClass.getRadius();
    this.mass = particleClass.getMass();
    this.storedEnergy = particleClass.getEnergy();
}

/** Color of particle. Same as particleClass.getColor(). Convenience. */
public Color getColor() { return particleClass.getColor(); }

/** Get a particle property. */
public Object get(String key) { return properties.get(key); }

/** Save a particle property. */
public void put(String key, Object value) { properties.put(key, value); }

/** Do the particle touch each other or do they even overlap each other? */
public boolean overlaps(Particle particle) {
    // This method is a hot spot.
    float dx = position[X] - particle.position[X];
    float dy = position[Y] - particle.position[Y];
    float dz = position[Z] - particle.position[Z];
    float d = radius + particle.radius;
    return dx * dx + dy * dy + dz * dz < d * d;
}

/**
 * Print a compact representation of the particle. If called by {@link
 * saces.sim.Collision}, prints an even shorter representation.
 */
public String toString() {
    // a shorter representation if used by collision objects
    String callingClass = new Throwable().getStackTrace()[3].getClassName();
    if (callingClass.equals("saces.sim.Collision")) {
        return String.format("%s@%x(speedSq=%5.1f)",
            particleClass.getName(), hashCode(), speedSq);
    }
    else {
        float p[] = position;
        float v[] = velocity;
        return String.format(fmt, particleClass.getName(), hashCode(),
            p[X], p[Y], p[Z], v[X], v[Y], v[Z]);
    }
}

// Format string for toString()
private static final String fmt = String.format(
    "%%s@%%x<p=(%1$s %1$s %1$s) v=(%2$s %2$s %2$s)>", "%4.1f", "% 5.1f");

}