/*
 *                        s a c e s
 *      Simple Artificial Chemistry Experiment System
 *
 * A semester project Summer 2005 and a diploma project 2006
 *           Berne University of Applied Sciences
 *   HTI School of Engineering and Information Technology
 *             Anthony Aguillon, Daniel Noelpp
 *
 */
package saces.sim;

import saces.Util;
import static saces.Util.*;
import saces.exp.Experiment;
import saces.exp.ParticleClass;
import saces.file.BinaryLog;
import saces.pnp.Decayer;
import saces.pnp.Detector;
import saces.pnp.Distributor;
import saces.pnp.DistributorMaxwellBoltzmann;
import saces.pnp.Merger;
import saces.pnp.Reflector;
import saces.pnp.Response;
import saces.pnp.Transformer;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;


/**
 * The fields part of a simulation with all parameters of an experiment which
 * change during the simulation, like current pressure, pressure, the number of
 * particles (total or per particle class), etc.
 */
public class Simulation {

/** For efficiency some fields are cached here. */
public static class Cache {

    /** The bounds of the experiment box (a cache!) */
    public float[] bounds = new float[3];
}

// NOTE: if you add another member variable, please (re)setExperiment it
// in setExperiment().

public Cache cache;

private Experiment experiment = Experiment.NULL;

private long timeOfLastUpdate;
private long timeOfLastMeasurement;
private long measureInterval;

/** Total simulation running time.*/
private long runningTime;
/** Timestamp of simulation halt.*/
private long timeStartOfPause;
/** Total time of simulation halt.*/
private long totalTimePaused;
/** Has the simulation been restarted (from a halt).*/
private boolean restarted;

private int frameCount;
private int frameRate;
private float slowingFactor;
private float histogramMaxSpeed;
private int histogramStepCount;
private float volume;
private float boltzmannConstant;
private float idealGasConstant;


private final Map<String, Object> properties = new HashMap<String, Object>();
private int[] countPerParticleClass;

// The particled
private Set<Particle> particles = new HashSet<Particle>();

// The partitions
private Partition[][][] partitions;

private BinaryLog log = BinaryLog.NULL;

public static final float SLOWING_FACTOR_STEPS = 0.05f;
public static final float SIMULATION_MAX_SLOWING_FACTOR = 1.0f;
public static final float SIMULATION_MIN_SLOWING_FACTOR = 0.05f;

private static final String[] REACTION_COUNT_KEYS = {
    Merger.MERGE_COUNT, Transformer.TRANSFORM_COUNT, Decayer.DECAY_COUNT,
};

/** Create a simulation instance. */
public Simulation() {
    setExperiment(Experiment.NULL);
}

public Experiment getExperiment() {
    return experiment;
}

// Don't forget to (re)setExperiment ALL member variables here!

/** Set the experiment of the simulation. */
public synchronized void setExperiment(Experiment experiment) {
    String logfileName = experiment.getString("BinaryLog", "saces.binlog");
    log.close();
    log = experiment == Experiment.NULL ? BinaryLog.NULL
        : BinaryLog.create(logfileName);

    cache = new Cache();
    cache.bounds[X] = experiment.getWidth();
    cache.bounds[Y] = experiment.getHeight();
    cache.bounds[Z] = experiment.getDepth();

    this.experiment = experiment;

    timeOfLastUpdate = 0;
    timeOfLastMeasurement = 0;
    measureInterval = experiment.getInt("MeasureInterval", 2000);

    resetTime();

    frameCount = 0;
    frameRate = -1;
    slowingFactor = 1.0f;
    histogramMaxSpeed = experiment.getFloat("HistogramMax", 0f);
    histogramStepCount = experiment.getInt("HistogramSize", 100);
    boltzmannConstant = experiment.getFloat(
        DistributorMaxwellBoltzmann.KEY_BOLTZMANN_CONSTANT,
        DistributorMaxwellBoltzmann.DEFAULT_BOLTZMANN_CONSTANT);
    idealGasConstant = experiment.getFloat("IdealGasConstant", 1f);
    volume = cache.bounds[X] * cache.bounds[Y] * cache.bounds[Z];

    properties.clear();
    countPerParticleClass = new int[experiment.getParticleClassesCount()];
    particles.clear();
    partitions = null;

    for (String KEY : REACTION_COUNT_KEYS) putProperty(KEY, 0);
}

public BinaryLog getLogger() { return log; }

public long currentTimeMillis() {
    return System.currentTimeMillis();
}

/** Updates the simulation's total running time
 *  considering simulation halt and restart deltas*/
public void updateTime() {

    runningTime += Math.round((float)(currentTimeMillis() - timeOfLastUpdate)
        * slowingFactor);

    timeOfLastUpdate = currentTimeMillis();

    if (restarted) {
        totalTimePaused += Math.round(
            (float)(timeOfLastUpdate - timeStartOfPause)
                * slowingFactor);
        timeStartOfPause = 0;
        restarted = false;
    }
}

/** Resets the simulation's running time*/
public void resetTime() {
    runningTime = 0;
    timeStartOfPause = 0;
    totalTimePaused = 0;
    restarted = false;
}

public long getRunningTime() {
    return (runningTime - totalTimePaused) -
      java.util.TimeZone.getDefault().getOffset(currentTimeMillis());
}

/** Halts the simulation time*/
public void stopTime() {
    timeStartOfPause = timeOfLastUpdate;
    restarted = true;
}

public boolean isRunning() { return timeOfLastUpdate != 0; }

public long getTimeOfLastUpdate() { return timeOfLastUpdate; }

public float getSlowingFactor() {
    return slowingFactor;
}

public synchronized void clear() {
    particles.clear();
    partitions = null;
}

public float increaseSpeed() {

    if (slowingFactor + SLOWING_FACTOR_STEPS <= SIMULATION_MAX_SLOWING_FACTOR) {

        slowingFactor += SLOWING_FACTOR_STEPS;
    }
    else {
        slowingFactor = SIMULATION_MAX_SLOWING_FACTOR;
    }

    log.writeProperty("SlowerFactor", slowingFactor);
    return slowingFactor;

}

public float decreaseSpeed() {

    if (slowingFactor - SLOWING_FACTOR_STEPS >= SIMULATION_MIN_SLOWING_FACTOR) {
        slowingFactor -= SLOWING_FACTOR_STEPS;
    }
    else {
        slowingFactor = SIMULATION_MIN_SLOWING_FACTOR;
    }

    log.writeProperty("SlowerFactor", slowingFactor);
    return slowingFactor;
}

public int getFrameRate() { return frameRate; }

public Object getProperty(String key) { return properties.get(key); }

public void putProperty(String key, Object value) {
    properties.put(key, value);
}

/** Do measurements at most once per second. */
public void measure() {

    // ------
    // -- Calculate frame rate and decide whether really do the measurements
    // -- (once every "measureInterval" milliseconds)

    frameCount++;
    frameRate = -1;

    long currentTime = currentTimeMillis();
    if (currentTime - timeOfLastMeasurement < measureInterval) return;

    timeOfLastMeasurement = currentTime;
    frameRate = frameCount;
    frameCount = 0;

    // ------
    // -- Write the counts of the particle classes
    log.writeParticleClassesRecord(countPerParticleClass);

    // ------
    // -- Calculate max and average speed and kinetic energy
    // -- Prepare mass center and total momentum

    float min = Float.MAX_VALUE;
    float avg = 0;
    float max = 0;
    float kinetic = 0;
    final float[] totalMomentum = new float[3];
    final float[] massCenter = new float[3];
    final int n = particles.size();

    for (Particle p : particles) {
        float speed = (float)Math.sqrt(p.speedSq);
        if (min > speed) min = speed;
        if (max < speed) max = speed;
        avg += speed;

        kinetic += p.mass * p.speedSq;

        totalMomentum[X] += p.mass * p.velocity[X];
        totalMomentum[Y] += p.mass * p.velocity[Y];
        totalMomentum[Z] += p.mass * p.velocity[Z];

        massCenter[X] += p.mass * p.position[X];
        massCenter[Y] += p.mass * p.position[Y];
        massCenter[Z] += p.mass * p.position[Z];
    }

    avg /= n;
    kinetic /= 2;

    // ------
    // -- Calculate a histogram of the particle speeds and write out

    final int[] speedHistogram = new int[histogramStepCount + 1];
    float top = histogramMaxSpeed == 0 ? max : histogramMaxSpeed;
    for (Particle p : particles) {

        // Calculate sqrt a 2nd time is inefficient but the alternative is to
        // have a big float array of sqrt values.
        float speed = (float)Math.sqrt(p.speedSq);

        int speedIndex = Math.round(histogramStepCount * speed / top);
        if (speedIndex < 0) speedIndex = 0;
        if (speedIndex > histogramStepCount) speedIndex = histogramStepCount;
        speedHistogram[speedIndex]++;
    }
    log.writeHistogram(min, avg, max, top, speedHistogram);

    // ------
    // Calculate total bound energy, mass and momentum and the mass center

    float bound = 0;
    float mass = 0;
    for (ParticleClass pc : experiment.getParticleClasses().values()) {
        int count = countPerParticleClass[pc.getIndex()];
        mass += pc.getMass() * count;
        bound += pc.getEnergy() * count;
    }

    massCenter[X] /= mass;
    massCenter[Y] /= mass;
    massCenter[Z] /= mass;

    float avgKinetic = kinetic / n;
    float temperature = 2 / 3f * avgKinetic / boltzmannConstant;
    float pressure = n * idealGasConstant * temperature / volume;

    log.writeProperty("ParticleCount", n);
    log.writeProperty("TotalKineticEnergy", kinetic);
    log.writeProperty("AverageKineticEnergy", avgKinetic);
    log.writeProperty("Temperature", temperature);
    log.writeProperty("Pressure", pressure);
    log.writeProperty("TotalBoundEnergy", bound);
    log.writeProperty("TotalEnergy", kinetic + bound);
    log.writeProperty("TotalMass", mass);
//    log.writeProperty("TotalMomentum[X]", totalMomentum[X]);
//    log.writeProperty("TotalMomentum[Y]", totalMomentum[Y]);
//    log.writeProperty("TotalMomentum[Z]", totalMomentum[Z]);
//    log.writeProperty("MassCenter[X]", massCenter[X]);
//    log.writeProperty("MassCenter[Y]", massCenter[Y]);
//    log.writeProperty("MassCenter[Z]", massCenter[Z]);

    // write reaction counts to binary log and reset counts
    for (String KEY : REACTION_COUNT_KEYS) {
        log.writeProperty(KEY, getProperty(KEY));
        putProperty(KEY, 0);
    }

    log.flush();
}

public void saveSnapshot() {
    log.writeSnapshot(getParticles());
}

public Particle[] getParticles() {
    return particles.toArray(new Particle[particles.size()]);
}

public Set<Particle> getParticleSet() {
    return particles;
}

public synchronized Partition[][][] getPartitions() {
    return partitions;
}

public synchronized void setPartitions(Partition[][][] partitions) {
    this.partitions = partitions;
}

/** Create a plug-and-play instance of a Reflector. */
public Reflector createWallReflector() {
    return createPlugAndPlayStep(Reflector.class);
}

/** Create a plug-and-play instance of a Response. */
public Response createCollisionResponse() {
    return createPlugAndPlayStep(Response.class);
}

/** Create a plug-and-play instance of a Decayer. */
public Decayer createParticleDecayer() {
    return createPlugAndPlayStep(Decayer.class);
}


/** Create a plug-and-play instance of a Distributor. */
public Distributor createParticleDistributor() {
    return createPlugAndPlayStep(Distributor.class);
}

/** Create a plug-and-play instance of a Detector. */
public Detector createCollisionDetector() {
    return createPlugAndPlayStep(Detector.class);
}

public Transformer createReactionTransformer() {
    return createPlugAndPlayStep(Transformer.class);
}

public Merger createReactionMerger() {
    return createPlugAndPlayStep(Merger.class);
}

/**
 * Create a particle (which class exactly is defined in the experiment
 * properties). The class must contain a constructor with the exact parameter
 * types Simulation, ParticleClass, or it wont' be loaded by this factory
 * method.
 *
 * @param pc Particle class of that particle (but not a Java class)
 * @return particle instance
 */
public Particle createParticle(ParticleClass pc) {
    String className = getExperiment().getString("Particle",
        "saces.gl.Sphere");

    Throwable e;
    try {
        if (className == null) throw new NullPointerException(
            "no property for key Particle");

        Class implClass = Class.forName(className);
        Constructor constructor = implClass.getDeclaredConstructor(
            Simulation.class, ParticleClass.class);
        Object instance = constructor.newInstance(this, pc);

        // An unchecked cast should additionally be verified
        if (!Particle.class.isInstance(instance))
            throw new ClassCastException(
                instance.getClass() + " instead of Particle");

        return Particle.class.cast(instance);
    }
    catch (ClassNotFoundException cnfe) { e = cnfe; }
    catch (IllegalAccessException iae) { e = iae; }
    catch (InstantiationException ie) { e = ie; }
    catch (ClassCastException cce) { e = cce; }
    catch (NullPointerException npe) { e = npe; }
    catch (NoSuchMethodException nsne) { e = nsne; }
    catch (InvocationTargetException ite) { e = ite; }

    Util.showErrorMessage("Problem creating particles", ""
        + "To create a particle, the Java class:\n"
        + className + "\n"
        + "is to be used, but there was a problem.\n\n"
        + "The exception throwed by the Java Virtual Machine is:\n"
        + e + ".\n\n"
        + "You need to edit either a) the 'Particle' property or\n"
        + "b) define/code the implementation class differently\n"
        + "(e.g. don't forget the constructor with two\n"
        + "arguments of type Simulation and ParticleClass)\n"
        + "or c) set up the Java classpath correctly and restart."
    );
    throw new SimulationPNPException(e);
}

public int getTotalParticleCount() {
    return particles.size();
}

public void removeParticle(Particle p) {
    boolean ok1 = particles.remove(p);
    if (!ok1) {
        System.out.println(missing(p));
        return; // to avoid negative particle numbers
    }
    countPerParticleClass[p.particleClass.getIndex()]--;

    // Dear Anthony, you cannot comment out these lines. If they are missing
    // the simulation goes crashing sooner or later because the collision
    // detection will report collisions with particles already removed
    // from the simulation, like ghost particles.
    if (partitions != null) {
        if (p.partition == null) {
            System.out.println(p + " doesn't have partition");
        }
        else if (p.partition.particles == null) {
            System.out.println("partition of " + p
                + " doesn't have particle list");
        }
        else {
            boolean ok2 = p.partition.particles.remove(p);
            if (!ok2) System.out.println(missing(p, p.partition.particles));
        }
    }
}

public void removeParticles(Particle p1, Particle p2) {
    removeParticle(p1);
    removeParticle(p2);
}

public void addParticle(Particle p) {
    countPerParticleClass[p.particleClass.getIndex()]++;
    particles.add(p);
}

public void addParticles(Particle p1, Particle p2) {
    particles.add(p1);
    particles.add(p2);
    countPerParticleClass[p1.particleClass.getIndex()]++;
    countPerParticleClass[p2.particleClass.getIndex()]++;
}

public boolean contains(Particle p) {
    return particles.contains(p);
}

public String missing(Particle p) {
    return missing(p, particles);
}

public static String missing(Particle p, Collection<Particle> particles) {
    return "missing " + p + "," + particles;
}

/**
 * Create a plug-and-play instance with reflection. This is a generic factory
 * method.
 *
 * @param clazz The super class or interface the instance implements or extends
 */
private <T> T createPlugAndPlayStep(Class<T> clazz) {
    String key = clazz.getName();
    key = key.substring(key.lastIndexOf('.') + 1); // strip away package
    String className = getExperiment().getString(key, defaultStep(key));

    Throwable e;
    try {
        if (className == null) throw new NullPointerException(
            "no property for key " + key);

        Object instance = Class.forName(className).newInstance();

// An unchecked cast should additionally be verified
        if (clazz.isInstance(instance)) return clazz.cast(instance);
        throw new ClassCastException(
            instance.getClass() + " instead of " + clazz);
    }
    catch (ClassNotFoundException cnfe) { e = cnfe; }
    catch (IllegalAccessException iae) { e = iae; }
    catch (InstantiationException ie) { e = ie; }
    catch (ClassCastException cce) { e = cce; }
    catch (NullPointerException npe) { e = npe; }
    catch (LinkageError le) { e = le; }

//    catch (NoSuchMethodException nsne) { e = nsne; }
//    catch (InvocationTargetException ite) { e = ite; }

    Util.showErrorMessage("Simulation Plug-and-Play Problem", ""
        + "A plug-and-play step for the simulation\n"
        + "could not be loaded. This is the step:\n"
        + key + ".\n\n"
        + "The exception thrown by the Java Virtual Machine is:\n"
        + e + ".\n\n"
        + "You need to edit either a) the property for the step or\n"
        + "b) define/code the plug-and-play step class differently\n"
        + "or c) set up the Java classpath correctly and restart."
    );
    throw new SimulationPNPException(e);
}

private String defaultStep(String key) {
    String[] defaultSteps = {
        "Distributor", "saces.pnp.DistributorNone",
        "Reflector", "saces.pnp.ReflectorSimple",
        "Detector", "saces.pnp.DetectorNone",
        "Merger", "saces.pnp.MergerSimple",
        "Transformer", "saces.pnp.TransformerNone",
        "Response", "saces.pnp.ResponseNone",
        "Decayer", "saces.pnp.DecayerSimple",
    };

    for (int i = 0; i < defaultSteps.length; i += 2)
        if (key.equals(defaultSteps[i])) return defaultSteps[i + 1];

    return null;
}

}